#!/bin/perl
#############################################################
#	Read output files from LANL's tree-building codes and 
# reformat them for use with divt. This script produces
# input suited to either the FREEZE_ or TETHER_LEAF_AGES
# version of divt.
#############################################################

##############################################################
#	The lines below are best-suited to a command line 
# interface on a unix box.  If you are using a Mac, you
# may perefer to edit the lines below and set the parameters
# directly.
###############################################################

	if( $#ARGV != 2 ) {
		print "usage: LANL2divt.pl outgroup_name tau width\n" ;
		print "where:\n" ;
		print "\toutgroup_name is the name associated with the outgroup node;\n" ;
		print "\ttau (in years) is the parameter of the exponential part of \n" ;
		print "\tthe prior on leaf ages.\n" ;
		print "\twidth is the uncertainty about the age of the leaf nodes arising from\n" ;
		print "\t      the precison with which the date of sampling was recorded.\n\n" ;
		
		exit( 1 ) ;
	}

	my( $outgroup ) = shift( @ARGV ) ;
	my( $leaf_prior_tau ) = shift( @ARGV ) ;
	my( $leaf_prior_width ) = shift( @ARGV ) ;

###############################################################
#	Hard-coded inputs used during development or on a Mac
#

# 	my( $infile ) = "gag.outfile" ;
# 	push @ARGV, $infile ;	# <> will read $infile after this
# 	
# 	my( $outgroup ) = "CONSENSUS" ;
# 	my( $leaf_prior_tau ) = "3.45" ;
# 	my( $leaf_prior_width ) = "1.0" ;

###############################################################
#	Stuff having to do with the formatting of the output
###############################################################
	
	my( $comment ) = "Your message here." ;
	my( $deriv_format, $br_format ) = ("%16.8e", "%16.8e" ) ;
	my( $vals_per_line ) = 4 ;

###############################################################
#	Skip down to the tree and read it, then read the derivs
#
	my( $LOOKAHEAD ) ;	# used to buffer lines from reads with <>

	skip2tree() ;
	my( @tm_branch_list ) = read_tree() ;
	# list_branches( @tm_branch_list ) ;

	my( $tm_N_branches ) = $#tm_branch_list ; 	# Tanmoy's branch nums start with 1
	my( $grad_ref, $Hess_ref) = build_mats( $tm_N_branches ) ;
	read_derivs( $grad_ref, $Hess_ref, $tm_N_branches ) ;
	# list_derivs( $grad_ref, $Hess_ref, $tm_N_branches ) ;

###############################################################
#	Process the data we've just read, renumbering the nodes
# and branches, so that we can write them out in a format  
# suited to divtime.
#
	my( $node_dt2tm_ref, $branch_dt2tm_ref, 
		 $kid_ref, $leaf_ref ) = renumber(	$outgroup,
		 									@tm_branch_list ) ;
		 
	my( $length_ref ) = dt_lengths( $branch_dt2tm_ref, 
									@tm_branch_list ) ;

###############################################################
#	Rewrite everything in a format suitable for divt.
#
	# write_Newick_tree( $node_dt2tm_ref, $kid_ref, $length_ref ) ;
	# write_comment( $comment ) ;
	write_leaves( $leaf_ref, $leaf_prior_width, $leaf_prior_tau  ) ;
	write_internal_nodes( $leaf_ref, $kid_ref ) ;
	write_branch_lengths( $branch_dt2tm_ref, \@tm_branch_list ) ;
	write_derivs( $grad_ref, $Hess_ref, $branch_dt2tm_ref ) ;
	# write_leaf_priors( $leaf_ref, $leaf_prior_width, $leaf_prior_tau ) ;
#________________________________________________________________________#

sub skip2tree()
########################################################
#	Gobble lines from STDIN until one reaches one that
# contains the words "Between",  "And" and "Length".
# Read one more line, then return.
######################################################### {
{
	my( $line ) = "" ;
	
	while( $line = <> ) {
		if( $line =~ m/^\s*Between\s+And\s+Length/ )  {
			$line = <> ;	# read the line of dashes too
			
			return ;
		}
	}
}
#_____________________________________________________#

sub read_tree()
########################################################
#	Read a description of a tree from one of Tanmoy's 
# output files and convert it to an array whose 
# entries are hashes containing the keys:
#
#	{ "to_node",  "from_node", "length" }
#
# The ordering implicit in to_node and from_node 
# is the one given by Tanmoy's output.
########################################################
{
	my( $line, $to_node, $from_node, $length ) ;
	my( %tags_n_lengths, $branch_tag ) ;
	my( $last_line_okay ) ;
	my( @branches ) ;	# used to return results
	
########################################################
#	Read lines describing branches of the form
#  
#		from to length  ---other stuff---
#
# and use them to build up a description of the 
# tree.
########################################################

	while( ($LOOKAHEAD = <>) =~ m/^\s*(\d+)\s+(\S*)\s*(0\.\d+)/ ) {
	#############################################
	#	The line has the correct form: for the
	# moment, stash the length in a hash whose
	# keys are derived from $to_node and 
	# $from_node
	#
		$from_node = $1 ;
		$to_node = $2 ;
		$length = $3 ;
		
		$branch_tag = "$from_node $to_node" ;	
		$tags_n_lengths{$branch_tag} = $length ;
		
#		print "<$branch_tag>\n" ;
	}
	
########################################################
#  Read a numbered list of the branches whose entries 
# look like:
#
#	branch_num: to_node from_node
#
# one needs this to make sense of the Hessian and
# gradient.
########################################################

	while(	($LOOKAHEAD = <>) &&
			!($LOOKAHEAD =~ m/^\s*(\d+):\s+(\d+)\s*(\S+)/) ) {}
			
	do {
	####################################################
	#	Parse the line and assemble the branch tag.
	# N.B. we assume, perhaps incorrectly, that
	# from_node and to_node will appear in the same 
	# order as in the previous section
	#
		$line = $LOOKAHEAD ;
		if( !($line =~ m/^\s*(\d+):\s+(\d+)\s*(\S+)/) ) {
			$last_line_okay = 0 ;
		}
		else {
			$last_line_okay = 1 ;
		
			$branch_num = $1 ;
			$from_node = $2 ;
			$to_node = $3 ;
			
			$branch_tag = "$from_node $to_node" ;
			
		####################################################
		#	Store the data in a local hash and save a
		# reference to it.
		#
			my( %branch_hash ) = () ;	# created afresh each time 
			
			$branch_hash{ "to_node" } = $to_node ;
			$branch_hash{ "from_node" } = $from_node ;
			if( defined( $tags_n_lengths{ $branch_tag } ) ) {
				$branch_hash{ "length" } = $tags_n_lengths{ $branch_tag } ;
				$branches[$branch_num] = \%branch_hash ;
			}
			else {
				die "Tag <$branch_tag> is undefined.\n" ;
			}
		}
			
	} while( $last_line_okay && ($LOOKAHEAD = <>) ) ;
		
########################################################
#  Return the array-of-hashes
########################################################

	return @branches ;
}
#_____________________________________________________#

sub list_branches( @lb_tree )
########################################################
#	Print to_node, from_node and length for each branch
#
{
	my( @lb_tree ) = @_ ;
	my( $branch_num, $br ) ;
	
	foreach $branch_num (1..$#lb_tree ) {
		$br = $lb_tree[$branch_num] ;  # br is short for branch_ref
		print	sprintf( "%3d:  ", $branch_num ) .
				sprintf( "%3d  ", $br->{from_node} ) .
				sprintf( "%-14s    ", $br->{to_node} ) .
				"$br->{length}\n" ;
	} 

}
#_____________________________________________________#

sub build_mats( $N_branches )
########################################################
#	Build a square matrix to hold the Hessian and
# a vector to hold the gradient.
#
{
	my( $N_branches, @etc ) = @_ ;
	my( $row_num, @gradient ) ;
	my( @Hessain ) ;
	
	foreach $row_num ( 1..$N_branches ) {
		my( @row ) ;	# created afresh each time 
		$Hessian[$row_num] = \@row ;
	}
	
	return( \@gradient, \@Hessain ) ;
}
#_____________________________________________________#

sub read_derivs( $grad_ref, $Hess_ref, $N_branches )
########################################################
#	Read derivatives until the Hessain is full
#
{
	my( $grad_ref, $Hess_ref, $N_branches, @etc ) = @_ ;
	my( $row, $col, $line, $Hess_val, $grad_val ) ;
	my( $line_was_okay ) ;
	my( $Hess_vals_set, $grad_vals_set ) = ( 0, 0 ) ;
	
	do {
		$line = $LOOKAHEAD ;
		if( $line =~ m/^\s*(\d*)\s*(\d*):\s*([\d-+\.e]*)\s*([\d-+\.e]*)/ ) {
			$line_was_okay = 1 ;
			
			( $row, $col, $Hess_val, $grad_val ) = ( $1, $2, $3, $4 ) ;
		
			if( !defined( $$Hess_ref[$row][$col] ) ) {
				$$Hess_ref[$row][$col] = $Hess_val ;
				$$Hess_ref[$col][$row] = $Hess_val ;
				if( $row != $col )	{ $Hess_vals_set += 2 ; }
				else				{ $Hess_vals_set += 1 ; } 
			}
			
			if( !defined( $$grad_ref[$row] ) ) {
				$$grad_ref[$row] = $grad_val ;
				$grad_vals_set++ ; 
			}
		}
		else {
			$line_was_okay = 0 ;
		}
		
	} while( $line_was_okay && ($LOOKAHEAD = <>) ) ;

			
	die "Wrong number of gradient vals\n" unless 
						($grad_vals_set == $N_branches) ;
		
	my( $Hess_vals_expected ) = $N_branches * $N_branches ;
	if( $Hess_vals_set != $Hess_vals_expected ) {
		print "Expected $Hess_vals_expected Hessian values, " ;
		print "but set $Hess_vals_set\n" ;
		foreach $row (1..$N_branches) {
			foreach $col (1..$row) {
				if( !defined( $$Hess_ref[$row][$col] ) ) {
					die "Hessain[$row][$col] is undefined.\n" ;
				}
			}
		}
		
		die "Couldn't identify missing value?!?\n" ;
	}
}
#_____________________________________________________#

sub list_derivs( $grad_ref, $Hess_ref, $N_branches )
#####################################################
#	Print out the components of the gradient and
# the Hessian - this is mainly useful for debugging
#
{
	my( $grad_ref, $Hess_ref, $N_branches, @etc ) = @_ ;
	my( $j, $k, $vals_on_line, $line ) ;
	
#####################################################
#	First print the gradient . . .
#
	print "\n\n# Gradient:\n" ; 
	
	$line = "" ;
	foreach $j ( 1..$N_branches) {
		die "Gradient[$j] is undefined!\n" unless defined($$grad_ref[$j]) ;
		
		$line .= sprintf( $deriv_format, $$grad_ref[$j] ) ;
		if( ++$vals_on_line == $vals_per_line ) {
			print "$line \n" ;
			
			$line = "" ;
			$vals_on_line = 0 ;
		}
	}
	
	if( $vals_on_line ) {
		print "$line \n" ;
		
		$line = "\t" ;
		$vals_on_line = 0 ;
	}

#####################################################
#	. . . then the rows of the Hessian.
#
	print "\n" ;
	
	foreach $j (1 .. $N_branches ) {
		print "\n# Hessian row $j:\n" ; 
	
		$line = "" ;
		foreach $k ( 1..$N_branches) {
			die "Hessain[$j][$k] is undefined!\n" unless defined($$Hess_ref[$j][$k]) ;
			
			$line .= sprintf( $deriv_format, $$Hess_ref[$j][$k] ) ;
			if( ++$vals_on_line == $vals_per_line ) {
				print "$line \n" ;
				
				$line = "" ;
				$vals_on_line = 0 ;
			}
		}
		
		if( $vals_on_line ) {
			print "$line \n" ;
			
			$line = "\t" ;
			$vals_on_line = 0 ;
		}
	}
}
#_____________________________________________________#

sub renumber( $outgroup, @tm_branch_list )
#####################################################
#	Assign divtime-style numbers to the branches
# and nodes.
#
{
	my( $outgroup, @tm_branch_list ) = @_ ;
	
######################################################
#	Run through the list of branches noting how many 
# times each node occurs as an endpoint of a branch:
# leaf nodes appear only once.
#
	my( %nodes_n_counts, %nodes_n_peers ) ;
	my( $branch_num, $br, $to_node, $from_node ) ;
	foreach $branch_num (1..$#tm_branch_list) {
		$br = $tm_branch_list[$branch_num] ; 	# Tanmoy's array starts at 1
		
		$to_node = $br->{to_node} ;
		$from_node = $br->{from_node} ;
		
		$nodes_n_counts{ $to_node }++ ;
		$nodes_n_peers{ $to_node } .= "$from_node " ;
		
		
		$nodes_n_counts{ "$from_node" }++ ;
		$nodes_n_peers{ $from_node } .= "$to_node " ;
		
		# print "$to_node $from_node\n" ;  # development stuff
	}
	
	die "Outgroup not present!?!\n" unless 
				defined($nodes_n_counts{$outgroup}) ;
	
#######################################################
#	Now go through the list of nodes classifying
# each as a leaf node, an internal node, or the
# outgroup and counting each type.
#
	my( @leaves ) ;
	my( $N_leaves, $N_internal )  = ( 0, 0 ) ;
	foreach $node (sort keys %nodes_n_counts ) {
		# print "Node <$node> has incidence $nodes_n_counts{$node}\n" ;
		
		if( $nodes_n_counts{ $node } == 1 ) {
		###############################################
		#	It's a leaf - check whether it's the
		# outgroup and if not, count it.
		#
			if( $node ne $outgroup ) { 
				$leaves[$N_leaves++] = $node ; 
			}
		}
		elsif( $nodes_n_counts{ $node } == 3 ) {
			$N_internal++ ;
		}
		else {
			die "Node <$node> has incidence $nodes_n_counts{$node}!?\n" ;
		}
	}
	
	die "Node counts ($N_internal, $N_leaves) are wrong.\n" unless
		($N_internal == ($N_leaves - 1)) ;
	
########################################################
#	Assign numbers to the leaf nodes and the outgroup
#
	my( %nodes_n_nums, @node_dt2tm ) ;
	
	my( $leaf ) ;
	my( $leaf_num ) = 0 ;	# divtime starts numbering from zero
	foreach $leaf (@leaves) {
		$node_dt2tm[$leaf_num] = $leaf ;
		$nodes_n_nums{ $leaf } = $leaf_num++ ;
	}
	
	$nodes_n_nums{ $outgroup } = $N_leaves + $N_internal ;
	$node_dt2tm[$N_leaves + $N_internal] = $outgroup ;
	
########################################################
#	Development rubbish
#
# 	print "\nN_leaves = $N_leaves, N_internal = $N_internal\n" ;
# 	foreach $leaf (@leaves) {
# 		print "$nodes_n_nums{ $leaf }: $leaf\n" ;
# 	}

########################################################
#	Assign numbers to the internal nodes in such
# a way that a node always has a larger number than
# either of its children (use outgroup to root).
#
	my( $next_node_num ) = $N_leaves ;	# first unused number
	my( %to_process ) = %nodes_n_peers ;
	my( @parents_n_kids ) ;
	
	my( @node_list ) ;
	while( (@node_list = keys( %to_process )) ) {
	##########################################
	#	Look through the list of nodes 
	# that are still to be processed.
	#
		my( $crnt_node ) ;
		foreach $crnt_node ( @node_list ) {
		############################################
		#	Delete the entries for those nodes
		# that already have numbers. Otherwise,
		# examine the list of nodes to which the
		# current one is connected (its "peers").
		# If two of these are numbered we are
		# in a position to number the current
		# node too.
		#
			if( defined( $nodes_n_nums{ $crnt_node } ) ) {
				delete $to_process{ $crnt_node} ;
			}
			else {
				my( @peers ) = 
					split( " ", $to_process{ $crnt_node} ) ;
				
				my( $kids, $N_numbered_peers ) ;
				foreach $peer (@peers) {
					if( defined( $nodes_n_nums{ $peer } ) 
								&& ($peer ne $outgroup) ) {
					#########################################
					#	The peer is an already-numbered
					# child node: take note of it.
					#
						$N_numbered_peers++ ;
						$kids .= sprintf( "%3d ", 
									$nodes_n_nums{ $peer } ) ;
					}
				}
				
			################################################
			#	If the current node has two numbered peers,
			# it's children, we are fix it's number too.
			#
				if( $N_numbered_peers == 2 ) {	
					$node_dt2tm[$next_node_num] = $crnt_node ;
					$parents_n_kids[$next_node_num] = $kids ;
					$nodes_n_nums{ $crnt_node} = $next_node_num++ ;
				}				
			}
			
		} # End of loop that numbers nodes
				
	} # End of loop over un-numbered nodes
	
#############################################################
#	Construct a map that relates divtime-style branch 
# numbers to Tanmoy's: one needs this to sort out the
# derivatives.
#
	my( @branch_dt2tm ) ;
	my( $child, $parent) ;
	my( $dt_to, $dt_from ) ;
	foreach $branch_num (1..$#tm_branch_list) {
	##########################################################
	#	By design, child nodes always have lower indices
	# than their parents
	#
		$br = $tm_branch_list[$branch_num] ;
		
		$dt_to = $nodes_n_nums{ $br->{to_node} } ;		
		$dt_from = $nodes_n_nums{ $br->{from_node} } ;
		
		if( $dt_to < $dt_from ) {	$child = $dt_to ; }
		else {						$child = $dt_from ; }
		
		$branch_dt2tm[$child] = $branch_num ;
	}

########################################################
#	Yet more development rubbish
#

# 	print "\n" ;
# 	foreach $node (sort keys %nodes_n_nums) {
# 		print "$node: $nodes_n_nums{$node}\n" ;
# 	}
# 	
# 	my( @dt_kid2parent ) ;
# 	foreach $branch_num (1..$#tm_branch_list) {
# 	##########################################################
# 	#	By design, child nodes always have lower indices
# 	# than their parents
# 	#
# 		$br = $tm_branch_list[$branch_num] ;
# 		
# 		$dt_to = $nodes_n_nums{ $br->{to_node} } ;		
# 		$dt_from = $nodes_n_nums{ $br->{from_node} } ;
# 		
# 		if( $dt_to < $dt_from ) {
# 			$child = $dt_to ; $parent = $dt_from ;
# 		}
# 		else {
# 			$child = $dt_from ; $parent = $dt_to ;
# 		}
# 
# 		$dt_kid2parent[$child] = $parent ;
# 	}
# 
# 	print "\n" ;
# 	my( $dt_branch_num ) ;
# 	foreach $dt_branch_num (0..$#branch_dt2tm) {
# 	#########################################################
# 	#	By divtime convention, branches have the same index 
# 	# as the child node to which they are connected.
# 	#
# 		print	"$dt_branch_num: " .
# 				"$dt_kid2parent[$dt_branch_num] -> " .
# 				"$dt_branch_num  =?=  " .
# 				"$branch_dt2tm[$dt_branch_num]: " .
# 				"$node_dt2tm[$dt_kid2parent[$dt_branch_num]] & " .
# 				"$node_dt2tm[$dt_branch_num]\n" ;
# 	}
		
#############################################################
# Return references to all the arrays of interest
#
# 	my( $node_dt2tm_ref, $branch_dt2tm_ref, $kid_ref,
# 		 $tip_ref ) = renumber( $outgroup, @tm_branch_list ) ;
#
	return( \@node_dt2tm, \@branch_dt2tm,
				\@parents_n_kids, \@leaves,  ) ;
}
#____________________________________________________________#

sub dt_lengths( $branch_dt2tm_ref, @tm_branches )
############################################################
#	Build a table of branch lengths as a function of
# divtime-style branch number.
#
{
	my( $branch_dt2tm_ref, @tm_branches ) = @_ ;
	
	my( $br, $dt_num, $tm_num, @lengths ) ;
	foreach $dt_num (0..($#$branch_dt2tm_ref-1)) {
		$tm_num = $$branch_dt2tm_ref[$dt_num] ;
		
		$br = $tm_branches[$tm_num] ;
		$lengths[$dt_num] = $br->{length} ;
		
		# print "\t $dt_num ($tm_num): $lengths[$dt_num]\n" ;
	}
	
	return( \@lengths ) ;
}
#____________________________________________________________#

sub write_Newick_tree( $n_ref, $k_ref, $l_ref )
##########################################################
#	Write out a rooted tree in Newick format. Most
# of the work is done by the recursive function
# print_descendants().
#
{
	my( $node_ref, $kid_ref, $len_ref ) = @_ ;

########################
#	Debugging stuff
#

#	$global_depth = 0 ;
#	
# 	foreach $j (0..$root) {
# 		print "\t$j: $$kid_ref[$j]\n" ;
# 	}
	
	my( $root ) = $#$n_ref ;
	print_descendants( $root, $node_ref, $kid_ref, $len_ref ) ;
	print " ;\n" ;
}
#____________________________________________________________#

sub print_descendants( $root, $n_ref, $k_ref, $l_ref )
#########################################################
#	Print all the descendants of this root node in
# Newick format
#
{
	my( $root, $node_ref, $kid_ref, $len_ref ) = @_ ;
	
	# $global_depth++ ;
	# print "Recursion depth $global_depth: root is $root; kids are <$$kid_ref[$root]>\n" ;
	
	my( @kids ) = split( " ", $$kid_ref[$root] ) ;
	if( @kids ) {
		print "(" ;
		print_descendants( $kids[0], $node_ref, $kid_ref, $len_ref ) ;
		print ": $$len_ref[$kids[0]], " ;
		print_descendants( $kids[1], $node_ref, $kid_ref, $len_ref ) ;
		print ": $$len_ref[$kids[1]])" ;
	}
	else {
		print "$$node_ref[$root]" ;
	}
	
	# $global_depth-- ;
}
#____________________________________________________________#

sub write_leaves( $leaf_ref ) 
##############################################################
#	Write out leaf nodes and numbers
#
{
	my( $leaf_ref, $width, $tau ) = @_ ;
	
	print "#####################################################\n" ;
	print "#	Leaf names with constraints on their ages: \n" 		;
	print "# name  left-edge-o-window  window_width tau\n"		;
	print "#####################################################\n"	;

	my( $leaf_num, $year ) ;
	foreach $leaf_num ( 0 .. $#$leaf_ref ) {
		$leaf_name = $$leaf_ref[$leaf_num] ;
		if( $leaf_name =~ m/\s*\S(\d{2})\S*/ ) {
			$year = $1 ;
			
			print	"Leaf:  " . sprintf( "%3d", $leaf_num ) .
					sprintf( "%20s", $leaf_name ) .
					"  19$year  $width  $tau\n" ;
		}
		else {
			die "Can't recognize year of measurement for <$leaf_name>.\n" ;
		}
	}
}
#____________________________________________________________#

sub write_internal_nodes( $leaf_ref, $kid_ref ) 
##############################################################
#	Write out leaf nodes and numbers
#
{
	my( $leaf_ref, $kid_ref, @etc ) = @_ ;
	
	print "##################################################### \n" ;
	print "# Internal nodes in the format:\n" ;
	print "#\tInternal: child1 child2 parent constraint\n" ;
	print "##################################################### \n" ;
	
	my( $node_num ) ;
	foreach $node_num ( ($#$leaf_ref+1)..$#$kid_ref ) {
		print "Internal: $$kid_ref[$node_num] " . 
				sprintf( "%3d", $node_num ) . "  0.0 \n" ;
	}
}
#_____________________________________________________________#

sub write_branch_lengths( $branch_dt2tm_ref, $branch_list_ref )
###############################################################
#	Write out the list of branch lengths.
#
{
	my( $branch_dt2tm_ref, $branch_list_ref ) = @_ ;
	my( $dt_j, $tm_j, $line, $vals_on_line ) ;
	
###############################################################
#	Write all the branch lengths: this format it easier to 
# set up the Taylor approx forthe log-likelihood.
#
	print "##################################################### \n" ;
	print "# Branch lengths: a branch has the same number as\n" ;
	print "# the child node to which it is connected.\n" ;
	print "##################################################### \n" ;
	
	$line = "" ;
	my( $last_branch ) = $#$branch_dt2tm_ref - 1 ;	# highest numbered 
													# branch goes to outgroup
	foreach $dt_j ( 0..$last_branch ) {
		$tm_j = $$branch_dt2tm_ref[$dt_j] ;
		
		my( $br ) = $$branch_list_ref[$tm_j] ;
		
		die "Branch length $dt_j ($tm_j) is undefined!\n" unless 
									defined($$branch_list_ref[$tm_j] ) ;
		
		$line .= sprintf( $br_format, $br->{length} ) ;
		if( ++$vals_on_line == $vals_per_line ) {
			print "$line \n" ;
			
			$line = "" ;
			$vals_on_line = 0 ;
		}
	}
	
	if( $vals_on_line ) {
		print "$line \n" ;
		
		$line = "\t" ;
		$vals_on_line = 0 ;
	}
}
#_____________________________________________________________#

sub write_derivs( $grad_ref, $Hess_ref, $branch_dt2tm_ref )
#############################################################
#	Print out the components of the gradient and the
# Hessian - this is essentially the same code as
# list_derivs(), but this one numbers things according
# to divtime conventions and suppresses data from that
# branch whose tip is the outgroup.
#
{
	my( $grad_ref, $Hess_ref, $branch_dt2tm_ref, @etc ) = @_ ;
	my( $dt_j, $tm_j, $k, $vals_on_line, $line ) ;
	
#####################################################
#	First print the gradient . . .
#
	print "##################################################### \n" ;
	print "# Components of the gradient and Hessian listed \n" ;
	print "# in order of branch number.\n" ;
	print "##################################################### \n" ;
	
	$line = "" ;	
	my( $last_branch ) = $#$branch_dt2tm_ref - 1 ;	# highest numbered 
													# branch goes to outgroup
	foreach $dt_j ( 0..$last_branch ) {
		$tm_j = $$branch_dt2tm_ref[$dt_j] ;
		
		die "Gradient[$dt_j ($tm_j)] is undefined!\n" unless 
									defined($$grad_ref[$tm_j]) ;
		
		$line .= sprintf( $deriv_format, -$$grad_ref[$tm_j] ) ;
		if( ++$vals_on_line == $vals_per_line ) {
			print "$line \n" ;
			
			$line = "" ;
			$vals_on_line = 0 ;
		}
	}
	
	if( $vals_on_line ) {
		print "$line \n" ;	# print the partial line of remaining vals
		$vals_on_line = 0 ;
	}

#####################################################
#	. . . then the rows of the Hessian.
#	
	foreach $dt_j ( 0..$last_branch ) {
		$tm_j = $$branch_dt2tm_ref[$dt_j] ;
		print "# Hessian row $dt_j:\n" ; 
	
		$line = "" ;
		foreach $dt_k ( 0..$dt_j ) {
			$tm_k = $$branch_dt2tm_ref[$dt_k] ;
			
			die "Hessain[$dt_j ($tm_j)][$dt_k ($tm_k)] is undefined!\n" unless 
				defined($$Hess_ref[$tm_j][$tm_k]) ;
			
			$line .= sprintf( $deriv_format, -$$Hess_ref[$tm_j][$tm_k] ) ;
			if( ++$vals_on_line == $vals_per_line ) {
				print "$line \n" ;
				
				$line = "" ;
				$vals_on_line = 0 ;
			}
		}
		
		if( $vals_on_line ) {
			print "$line \n" ; # print the partial line of remaining vals
			$vals_on_line = 0 ;
		}
	}
}
#_____________________________________________________________#

sub write_leaf_priors( $leaf_ref, $width, $tau )
#####################################################
#	Specify the data for the priors on the leaves
#
{
	my( $leaf_ref, $width, $tau ) = @_ ;
	
	print "#####################################################\n" ;
	print "#	Constrain the dates of the tips: \n" 				;
	print "# node_num  left-edge-o-window  window_width tau\n"		;
	print "#####################################################\n"	;

	my( $leaf_num, $year ) ;
	foreach $leaf_num ( 0 .. $#$leaf_ref ) {
		$leaf_name = $$leaf_ref[$leaf_num] ;
		if( $leaf_name =~ m/\s*\S(\d{2})\S*/ ) {
			$year = $1 ;
			
			print	sprintf( "%3s", $leaf_num ) .
					"  19$year  $width  $tau\n" ;
		}
		else {
			die "Can't recognize year of measurement for <$leaf_name>.\n" ;
		}
	}
}
# ____________________________________________________

sub write_comment( $comment_str )
####################################################
#	Note the name of the input file
#
{
	my( $comment_str, @etc ) = @_ ;
	
	print "# $comment_str\n" ;
}
