/**************************************************************
*	The functions below are used to compute a prior on
* the ages of the internal nodes. It is constant over sets of 
* ages that have the same leaf ages and root age and is
* nonzero only for those sets of ages in which parents are
* at least as old as their kids.
*
* mrm, LANL, All Saints Day 2000
*/
# include <stdio.h>	
# include <stdlib.h>	/* calloc(), free()	*/
# include <string.h>	/* memcpy()			*/

# define DEBUG_LEVEL 0
# include "tune_debug.h"

# include "tether_tree.h"	/* Defines N_KIDS	*/
# include "div_switches.h"	/* We need to know if the tips are frozen	*/
# include "age_prior.h" 

/****************************************************
*	Share some data with tether_tree.c
*/
extern int		ntips, nnodes, **tonode ;
extern double	*constraint ;

/****************************************************
*	Prototype for a function private to this file.
*/
static int age_prior_allocate_mem(void);

/************************************************
*	Variables private to this file.
*/
static double		*coef_block = NULL ;
static double		**prior_denom = NULL ;
/* ______________________________________________________________________ */

double age_prior_ratio( double *crnt_ages, double *prop_ages )
/*************************************************************
*	Use the machinery defined below to compute a ratio of
* prior densities for the internal nodes. This calcualtion
* does not include any priors on the ages of the leaves or
* root node and assumes that only a single node moves.
*/
{
	int		j ;
	double	*coefs, crnt_age, prop_age ;
	double	crnt_rho_denom, prop_rho_denom, ratio ;
	
	/* If debugging, check initialization */
	assert( prior_denom != NULL ) ;
	
/***************************************************************
*	Check every internal node to see whether it has moved
* and, if so, multiply in a suitable ratio of priors.
*/
	ratio = 1.0 ;	/* to begin with */
	for( j = ntips ; j < nnodes ; j++ ) {
	/**************************************************************
	*	Check whether node j has moved and skip straight to 
	* the bottom of the loop if it hasn't.
	*/
		if( crnt_ages[j] == prop_ages[j] )
			continue ;
			
	/**************************************************************
	*	The node has moved: make some convenient local nicknames 
	* for all the varaibles in play . . .
	*/
		crnt_age = crnt_ages[j] ;
		prop_age = prop_ages[j] ;

		coefs = prior_denom[j] ;

	/**************************************************************
	*	. . . then calculate the relevant ratio and multiply
	* it into the result.
	*/
		crnt_rho_denom = 
			coefs[0] + crnt_age * ( coefs[1] + crnt_age * coefs[2] ) ;
			
		prop_rho_denom = 
			coefs[0] + prop_age * ( coefs[1] + prop_age * coefs[2] ) ;
			
		assert( crnt_rho_denom >= 0.0 ) ;
		assert( prop_rho_denom >= 0.0 ) ;
		
					/* prop_prior / crnt_prior */
		ratio *= crnt_rho_denom / prop_rho_denom ;
	}
	
	return( ratio ) ;
}
/* ______________________________________________________________________ */

double single_move_age_prior_ratio( double		*crnt_ages,
									double		*prop_ages,
									int			moving_node )
/*************************************************************
*	Use the machinery defined below to compute a ratio of
* prior densities for the internal nodes. This calcualtion
* does not include any priors on the ages of the leaves or
* root node and assumes that only a single node moves.
*/
{
	double	*coefs, crnt_age, prop_age ;
	double	crnt_rho_denom, prop_rho_denom ;
	
	/* If debugging, check initialization and moving_node arg */
	assert( prior_denom != NULL ) ;
	assert( (moving_node >= 0) && (moving_node < nnodes) ) ;
	
/**************************************************************
*	. . . make some convenient local nicknames for all the
* varaibles in play . . .
*/
	coefs = prior_denom[moving_node] ;
	crnt_age = crnt_ages[moving_node] ;
	prop_age = prop_ages[moving_node] ;

/**************************************************************
*	Calculate everything and return.
*/
	crnt_rho_denom = 
		coefs[0] + crnt_age * ( coefs[1] + crnt_age * coefs[2] ) ;
		
	prop_rho_denom = 
		coefs[0] + prop_age * ( coefs[1] + prop_age * coefs[2] ) ;
		
	assert( crnt_rho_denom >= 0.0 ) ;
	assert( prop_rho_denom >= 0.0 ) ;
	
	/* prop_prior / crnt_prior */
	return( crnt_rho_denom / prop_rho_denom ) ;
}
/* _____________________________________________________________ */

int init_age_prior( void )
/**********************************************************
*	Build the array of polynomials that appear in the
* denominator of the prior.  We treat all of them as
* quadratics (a.k.a. arrays of three doubles), even 
* though some may have coefficients for the quadratic 
* and even the linear terms.  Return zero for success
* nonzero for faiure.
*/
{
	int		j, k, child ;
	double	*coefs ;
	
	/* At compile time, set a switch that tells		*/
	/* whether leaf nodesare free to move or not.	*/
# if defined( TETHER_LEAF_AGES ) 
	static int	leaves_move = 1 ;	/* they move	*/
# else
	static int	leaves_move = 0 ;	/* they don't	*/
# endif
	
/**********************************************************
*	First, try to allocate memory.
*/
	if( age_prior_allocate_mem() )
		return( 1 ) ;
		
/**********************************************************
*	If we reach these lines, the arrays are all set up
* and we need only fill them. We begin with the polynomials
* for the tips, whose ages do not actually come into the
* prior.
*/
	for( j = 0 ; j < ntips ; j++ ) {
		coefs = prior_denom[j] ;
		
		coefs[2] = coefs[1] = 0.0 ;	/* no linear or quadratic terms */
		coefs[0] = 1.0 ;	/* (arbitrary) nonzero constant term	*/
	}
	
/***********************************************************
*	Now for the internal nodes. Their contributions to the
* prior depend on how many moveable child nodes they have.
*/
	for( j = ntips ; j < nnodes ; j++ ) {
	/********************************************************
	*	The results depend slightly on whether the leaf
	* nodes are frozen (skightly more fussy) or free to
	* move. 
	*/
		coefs = prior_denom[j] ;
		
		coefs[0] = 1.0 ;	/* start with a constant poly */
		coefs[1] = coefs[2] = 0.0 ;

		/* Multiply in a linear factor for each moveable child */
		for( k = 1 ; k <= tonode[j][N_KIDS] ; k++ ) { 
		/***************************************************
		*	Recall that the indices of child nodes are 
		* listed in 
		*
		*	{ tonode[j][1], . . . tonode[j][N_KIDS] }
		*/ 
			child = tonode[j][k] ;
			
			if( leaves_move || (child >= ntips) ) {
			/***********************************************
			*	The child node is free to move and hence 
			* contributes a factor of 
			*
			*	(t_p - constraint[child])
			* 
			* to the denominator of the prior. Here t_p
			* is the age of the parent and constraint[child]
			* is the constraint (may be zero) on the age of 
			* the child node.
			*/
				coefs[2] = coefs[1] - constraint[child] * coefs[2] ;
				coefs[1] = coefs[0] - constraint[child] * coefs[1] ;
				coefs[0] = -constraint[child] * coefs[0] ;
			}
			
		} 	/* End of loop over child nodes */
			
	}	/* End of loop over internal nodes */
	
	return( 0 ) ;	/* If we return from here, all went well */	
}
/* _____________________________________________________________ */

int age_prior_allocate_mem( void )
/**********************************************************
*	Try to set up this file's private arrays, returning
* zero for success, nonzero for failure.
*/
{
	int			j ;
	double		*cpt ;

/***********************************************************
*	First try to allocate all the memory we'll need.
*/
	prior_denom = (double **) calloc( ((size_t) nnodes), sizeof(double *) ) ;
	if( prior_denom == NULL ) {
		fprintf( stderr, 
			"init_age_prior(): can't allocate array of pointers-to-coefficients.\n" ) ;
			
		return( 1 ) ;
	}

	coef_block = (double *) calloc( 3 *((size_t) nnodes), sizeof(double) ) ;
	if( coef_block == NULL ) {
		fprintf( stderr, 
			"init_age_prior(): can't allocate array of coefficients.\n" ) ;
			
		return( 2 ) ;
	}

/*************************************************************
*	If we reach these lines, all the arrays we're allocated
* successfully, so now we just fill up the array of 
*/
	cpt = coef_block ;
	for( j=0 ; j < nnodes ; j++ ) {
		prior_denom[j] = cpt ;
		cpt += 3 ;
	}
	
	return( 0 ) ;	/* Success */
}
/* _____________________________________________________________ */

void age_prior_cleanup( void )
/*************************************************************
*	Release any memory allocated by age_prior_allocate_mem().
*/
{
	if( prior_denom != NULL ) {
		free( prior_denom ) ;
		prior_denom = NULL ;
	}

	if( coef_block != NULL ) {
		free( coef_block ) ;
		coef_block = NULL ;
	}
}
