/**************************************************************
*	The functions below generate initial conditions for 
* the divtime programe. Generally speaking, the inputs 
* to this suite of functions are a tree topology, a set
* of branch lengths and the parameters of the priors.
* The outputs are a set ages for the nodes and rates for
* for the branches.
*
* mrm, UMIST, 3 December 1999
*/
# include <stdio.h>
# include <string.h>
# include <stdlib.h>
# include <math.h>

# define DEBUG_LEVEL 0
# include "tune_debug.h"

# include "random.h"
# include "arrayutl.h"
# include "dt_startup.h"
# include "tether_tree.h"

/**************************************************
*	Variables shared with tehter_divtime.c
*/
extern int		nnodes, ntips, **tonode ;
extern double	*mean ;		/* branch lengths */
extern double	mean_root_age, mean_nu, mean_root_rate ; 
extern double	timegamr, timegamlam ;	/* prior on root age */
extern double	rategamr, rategamlam ;	/* prior on root rate */

/*************************************************
*	Parameters used in the group of functions
* that smooth rate variations.
*/
# define MAX_CYCLES		100		/* in maximize_local_loglike() */

/**************************************************
*	A standard macro
*/
# define max(A,B) ((A) > (B) ? (A) : (B))

/**************************************************
*	Some limits on sizes, used in random_start()
* and least_sq_start().
*/
# define TINYLEN 		0.000001
# define TINYTIME 		0.0000001

/***************************************************
*	Prototypes for functions private to this file.
*/
static double root_loglike(double t_root);
static double local_loglike(double t_parent);
static double maximize_root_loglike(double *time);
static double maximize_local_loglike(double *len, double *t, double *v);
static double golden_maximizer(double *x, double (*f)(double x));

/***************************************************
*	These variables, private to this file, are
* used in the calculations of local log-likelihoods.
*/

/* lll -> local log likelihood */
static double	lll_len[3], lll_t[4], lll_v[2] ;

/* rll -> root log likelihood */
static double	*rll_len = lll_len, *rll_t = lll_t ;
/* ______________________________________________________________ */

void rates_from_tips(	double *mean_rate, double *rate_sd, 
						double time[], double branch_length[] )
/************************************************************
*	Do a survey of the leaf nodes and use a molecular
* clock hypothesis to estimate a rate for each pair that
* share a parent.
*/
{
	int		leaf, parent, sib, n_rates ;
	double	local_rate, time_diff, length_diff ;
	double	sum, sum_sq ;
	
	FILE	*rate_fp ;
	
/************************************************************
*	Open a file to receive the rates we're about to 
* estimate.
*/
	rate_fp = fopen( "tip_rates", "w" ) ;
	
/************************************************************
*	Run through the tips checking whether their parent's
* other child is a tip.
*/
	n_rates = 0 ;
	sum = sum_sq = 0 ;
	for( leaf = 0 ; leaf < ntips ; leaf++ ) {
	/********************************************************
	*	Find the indices of the tip's parent and sib nodes.
	*/
		parent = tonode[leaf][PARENT] ;
		if( tonode[parent][1] == leaf )
			sib = tonode[parent][2] ;
		else /* tonode[parent][2] == leaf */
			sib = tonode[parent][1] ;
			
		if( sib < leaf ) {	/* avoids double-counting */
		/****************************************************
		*	Work out the difference in ages between the tips
		* and the difference between in the lengths of the
		* branches that reach back to the root.
		*/
			length_diff = branch_length[leaf] - branch_length[sib] ;
			time_diff = time[leaf] - time[sib] ;
			
		/*****************************************************
		*	If the two tips are not from the same date, 
		* work out a local rate for the {leaf, sib, parent}
		* subtree.
		*/
			if( fabs(time_diff) > fabs(TINY_FRAC * time[leaf]) ) {
				local_rate = length_diff / time_diff ;
			
			/*****************************************************
			*	If the local rate appears sensible, report it.
			*/
				if( local_rate > 0 ) {
					sum += local_rate ;
					sum_sq += local_rate * local_rate ;
					n_rates++ ;
					
					if( rate_fp != NULL ) 
						fprintf( rate_fp, "%8f\n", local_rate ) ;
						
				}	/* End of case where rate looks sensible */
				
			}	/* End of case where tips differ in age */
			
		} /* End of case that we try to compute a rate */
	} 

/***************************************************
*	Tidy up and return.
*/
	if( rate_fp != NULL )
		fclose( rate_fp ) ;
		
	if( n_rates > 1 ) {
		*mean_rate = sum / n_rates ;
		*rate_sd = sqrt( (sum_sq - (sum * sum / n_rates)) / (n_rates - 1) ) ;
	}
}
/*____________________________________________________________*/

void random_start(  double *time, double *rate  )
/**************************************************
*	Generate a random initial condition, ignoring 
* both the priors and the branch lengths.
* This is probably not a good idea.
*/ 
{	
	int		ktrack ;

/****************************************************
*	Generate an age and rate for the root, then
* for the rest of the tree.
*/
	printf("starting at semi-random guess\n");
	time[nnodes-1] = exptial(1.0/ mean_root_age);
	rate[nbranches] = exptial(1.0/mean_nu);
	for(ktrack=nbranches-1; ktrack >= 0; ktrack--)
	{
		rate[ktrack] = 0.5+std_uniform()*1.5;
	}
	for(ktrack=nnodes-2; ktrack >= ntips; ktrack--)
	{
		if(mean[ktrack] > TINYLEN)
		{
			time[ktrack] = time[tonode[ktrack][PARENT]] - 
			(mean[ktrack]/rate[ktrack]);
		}
		else 
		{
			time[ktrack] = (1.0-TINYLEN)*time[tonode[ktrack][PARENT]];
		}
		if(time[ktrack] < TINYTIME)
		{
			time[ktrack] = (1.0-TINYTIME)*time[tonode[ktrack][PARENT]];
		}
	} 
}
/*____________________________________________________________*/

void mol_clock_start(  double *time, double *rate  )
/**************************************************************
*	Use the branch lengths and a molecular clock assumption
* to generate an initial condition. This is Jeff Thorne's
* original approach.
*/
{	
	int		ktrack, panode ;
	double	maxdis, *dist ;
	
/*******************************************************
*	Work out the distance from the root node to the
* most distant leaf. If some branches are shorter
* than SHORT_BRANCH, just act as though they were
* that long.
*/
	printf("Starting from a hopefully good guess\n");
	maxdis = 0.0;
	dist = dvec(ntips-1);
	for(ktrack = 0; ktrack < ntips; ktrack++)
	{
		panode = tonode[ktrack][PARENT];
		dist[ktrack] = max( SHORT_BRANCH, mean[ktrack] );
		while(tonode[panode][PARENT] != -1)
		{
			dist[ktrack] += max( SHORT_BRANCH, mean[panode] );
			panode = tonode[panode][PARENT];
		}
		if(maxdis < dist[ktrack])
		{
			maxdis = dist[ktrack];
		}
	}

/****************************************************
*	Set the "age" of the deepest node to be the 
* total branch distance to its most distant child,
* then work back toward the tips setting the ages
* of the other nodes in accord with their remove 
* from the root.  This is akin to guessing ages
* consistent with a constant-rate molecular clock.
*/
	time[nnodes-1] = maxdis;
	for(ktrack = nnodes-2; ktrack >= ntips; ktrack--)
	{
		time[ktrack] = time[tonode[ktrack][PARENT]] 
						- max(mean[ktrack], SHORT_BRANCH);
	}

/*****************************************************
*	In light of the times we've just chosen, set
* rates so that the branch lengths come out as
* specified in the input.
*/
	for(ktrack =0; ktrack < nbranches; ktrack++)
	{
		rate[ktrack] =max( SHORT_BRANCH, mean[ktrack])/(
			time[tonode[ktrack][PARENT]]- time[ktrack]);
	}
	
/*****************************************************
*	The last entry in the rate array, the one that
* looks as though it is a rate associated with the
* root node, is really the parameter \nu.
*/ 
	rate[nbranches] = mean_nu ; 
	freedvec( dist ) ;
}
/*____________________________________________________________*/

void least_sq_start(  double *time, double *rate  )
{	
	int		ktrack, panode, child_a, child_b ;
	double	*dist ;
	double	sum_d, sum_t, sum_dt, sum_tsq ;
	double	denom, est_rate, min_age, duration ;

/*******************************************************
*	Work out the distance from the root node to all
* the leaves. 
*/
	printf( "Estimating ages and rate at root.\n" ) ;
		
	dist = dvec(ntips-1) ;
	for( ktrack = 0 ; ktrack < ntips ; ktrack++ ) {
		panode = tonode[ktrack][PARENT] ;
		dist[ktrack] = mean[ktrack] ;	
		while(tonode[panode][PARENT] != -1 ) {
			dist[ktrack] += mean[panode] ;
			panode = tonode[panode][PARENT];
		}
	}

/*****************************************************
*	Compute various useful sums.
*/
	sum_t = sum_tsq = 0.0 ;
	sum_d = sum_dt = 0.0 ;
	for( ktrack = 0 ; ktrack < ntips ; ktrack++ ) {
		sum_d += dist[ktrack] ;
		sum_dt += dist[ktrack] * time[ktrack] ;
		
		sum_t += time[ktrack] ;
		sum_tsq += time[ktrack] * time[ktrack] ;
	}
	
/*****************************************************
*	Compute a best-fit constant rate, r, and and
* age for the root, t_r, by minimizing the sum 
*
* 	(r * (t_r - t_j) - dist[j])^2 
*
* as j varies over the tips.
*/
	denom = sum_tsq - ((sum_t * sum_t) / ntips) ;
	if( denom < ntips * sum_tsq * DBL_EPSILON ) {
		fprintf( stderr, 
			"Tips all appear to have same age; can't estimate rate.\n" ) ;
		exit( 2 ) ;
	}
	else {
		est_rate = (((sum_d * sum_t) / ntips) - sum_dt) / denom ;
		time[nnodes-1] = ((est_rate * sum_t) + sum_d) / (ntips * est_rate) ;
	}
	
	printf( "\tRoot_age = %.2f, estimated rate = %.6f\n", 
								time[nnodes-1], est_rate ) ;
						
/*****************************************************
*	Use the rate and root-age we've just estimated 
* to set ages for all the internal nodes.
*/
	for( ktrack = (nnodes - 2) ; ktrack >= ntips ; ktrack-- ) {
		time[ktrack] = time[tonode[ktrack][PARENT]] 
								- (mean[ktrack] / est_rate) ;
	}
	
/*****************************************************
*	Ensure that all nodes are at least as old as
* their children.
*/
	for( ktrack = ntips ; ktrack < nnodes ; ktrack++ ) {
		child_a = tonode[ktrack][1] ;
		child_b = tonode[ktrack][2] ;
		min_age = max( time[child_a] + 0.1 * (mean[child_a]  / est_rate) + TINYLEN, 
						time[child_b] + 0.1 * (mean[child_b] / est_rate) + TINYLEN) ;
						
		if( time[ktrack] < min_age ) {
			time[ktrack] = min_age ;
		}
	}

/*****************************************************
*	In light of the times we've just chosen, set
* rates so that the branch lengths come out as
* specified in the input.
*/
	for( ktrack =0 ; ktrack < nbranches ; ktrack++ ) {
		duration = time[tonode[ktrack][PARENT]] - time[ktrack] ;
		
		if( (mean[ktrack] < TINYLEN) || 
			(duration <= 0.1 * (mean[ktrack] / est_rate) ) ) {
			rate[ktrack] = est_rate ;
		}
		else {
			rate[ktrack] = mean[ktrack] / duration ;
		}
	}
		
/*****************************************************
*	The last entry in the rate array, the one that
* looks as though it is a rate associated with the
* root node, is really the parameter \nu. J. Thorne
* initialized this to (1 / mean_nu), but I can't see
* why.
*/ 
	rate[nbranches] = mean_nu ; 
	freedvec( dist ) ;
}
/* ______________________________________________________ */

void tip_rate_heterogeneity( double *rate )
/**********************************************************
*	Survey the tips and, for each pair that share a
* common parent, estimate a rate. Combine these to get
* a measure of rate heterogeneity at the tips.
*/
{	
	int		ktrack ;
	double	sum_x, sum_xsq ;

/***********************************************************
*	Print estimates of the rate heterogeneity at the tips.
*/
	sum_x = sum_xsq = 0.0 ;
	for( ktrack=0 ; ktrack < ntips ; ktrack++ ) {
		sum_x += log( rate[ktrack] ) ;
		sum_xsq += log( rate[ktrack] ) * log( rate[ktrack] ) ;
	}
	
	sum_x /= ntips ;
	sum_xsq = sqrt( (sum_xsq - (sum_x * sum_x * ntips)) /(ntips - 1)) ;
	fprintf( stderr, "\nLog of rates at tips:\n\t" ) ;
	fprintf( stderr, "%.8f +/- %.8f [%.6f, %.6f %.6f]\n\n",
		sum_x, sum_xsq, exp( sum_x - sum_xsq ),
		exp( sum_x ), exp( sum_x + sum_xsq ) 
	) ;
	
	sum_x = sum_xsq = 0.0 ;
	for( ktrack=0 ; ktrack < ntips ; ktrack++ ) {
		sum_x += rate[ktrack] ;
		sum_xsq += rate[ktrack] * rate[ktrack] ;
	}
	
	sum_x /= ntips ;
	sum_xsq = sqrt( (sum_xsq - (sum_x * sum_x * ntips)) /(ntips - 1)) ;
	fprintf( stderr, "Rates at tips:  " ) ;
	fprintf( stderr, "%.8f +/- %.8f\n\n", sum_x, sum_xsq ) ;
}
/* ______________________________________________________ */

void smooth_rates( double *time, double *rate )
/**********************************************************
*	Pass through the tree adjusting the ages of the
* internal nodes, but preserving the branch lengths.
* The idea is to use this as part of the least_sq 
* initializer in such a way as to make the rate 
* heterogeneity more evenly distributed and less severe.
*/
{
	int		j, child, child_a, child_b ;
	int		crnt_depth, n_to_smooth, n_rounds ;
	int		*depth ;
	
	double	tiny_length, var[2], len[3], t[4] ;
		
/***********************************************************
*	Work out the depth (max number of branches to tip)
* of each of the nodes.
*/
	depth = ivec( nnodes ) ;
	for( j=0 ; j < ntips ; j++ )
		depth[j] = 0 ;
		
	for( j = ntips ; j < nnodes ; j++ )
		depth[j] = -1 ; /* signifies that depth is undefined */
	
	n_to_smooth = nnodes - ntips ;
	while( n_to_smooth ) {	
		for( j = ntips ; j < nnodes ; j++ ) {
			child_a = tonode[j][1] ;
			child_b = tonode[j][2] ;
			if( (depth[child_a] >= 0) && (depth[child_b] >= 0) ) {
				n_to_smooth-- ;
				if( depth[child_a] > depth[child_b] )
					depth[j] = depth[child_a] + 1 ;
				else
					depth[j] = depth[child_b] + 1 ;
			}
		}
	}
	
/**********************************************************
*	Do several rounds of smoothing.
*/
	for( n_rounds = 0 ; n_rounds < depth[nnodes-1] ; n_rounds++ ) {
	/************************************************************
	*	Work backward in depth, shifting nodes to try to
	* smooth out the variations in rate.
	*/
		crnt_depth = 1 ;
		n_to_smooth = nnodes - ntips - 1 ;	/* root is special	*/
		while( n_to_smooth ) {
		/****************************************************
		*	Have a go at shifting all the nodes at the 
		* current depth.
		*/ 
			for( j = ntips ; j < nnodes ; j++ ) {
				if( depth[j] == crnt_depth ) {
				/***************************************************
				*	Get all the times, rates and branch lengths
				* that will be in play here.
				*/
					child_a = tonode[j][1] ;
					child_b = tonode[j][2] ;
					
					t[0] = time[child_a],	len[0] = mean[child_a] ;
					t[1] = time[child_b],	len[1] = mean[child_b] ;
					t[2] = time[j], 		len[2] = mean[j] ;
					t[3] = time[tonode[j][PARENT]] ;	// grandparent
					
					var[0] = 0.5 * mean_nu * (t[3] - t[0]) ;
					var[1] = 0.5 * mean_nu * (t[3] - t[1]) ;
					
					tiny_length = TINY_FRAC * 
								(len[0] + len[1] + len[2]) / 3.0 ;
					if (tiny_length < TINYLEN) tiny_length = TINYLEN;
					
				/****************************************************
				*	Minimize the conditional-rate contribution 
				* arising from this { grandparent, parent, kids }
				* group. We do this numerically, as a function of
				* the age of the parent node, time_p.
				*/
					time[j] = maximize_local_loglike( len, t, var ) ;
					
				/*****************************************************
				*	Set the rates so as to preserve the branch
				* lengths, but use the newly-computed time.
				*/	
					n_to_smooth-- ;			
					if( mean[child_a] < tiny_length ) 
						rate[child_a] = rate[j] ;
					else
						rate[child_a] = mean[child_a] / (time[j] - time[child_a]) ;
						
					if( mean[child_b] < tiny_length ) 
						rate[child_b] = rate[j] ;
					else
						rate[child_b] = mean[child_b] / (time[j] - time[child_b]) ;
						
					if( mean[j] < tiny_length ) 
						rate[j] = rate[tonode[j][PARENT]] ;
					else
						rate[j] = mean[j] / (time[tonode[j][PARENT]] - time[j]) ;
					
				}	/* updates age of a single node */
				
			}	/* loop updates age of nodes at crnt_depth */
			
			crnt_depth++ ;
			
		}	/* loop over non-root internal nodes */
		
	/*****************************************************
	*	Smoothing step for the root
	*/
		time[nnodes-1] = maximize_root_loglike( time ) ;
		
		tiny_length = TINY_FRAC * ( mean[tonode[nnodes-1][1]] +
									mean[tonode[nnodes-1][2]] ) ;
		
		for( j=1 ; j <= tonode[nnodes-1][N_KIDS] ; j++ ) {
			child = tonode[nnodes-1][j] ;
			if( mean[child] < tiny_length ) 
				rate[child] = mean_root_rate ;
			else 
				rate[child] = mean[child] / 
					(time[nnodes-1] - time[child]) ;
			
		}	/* loop of root's kids */
		
	} /* loop over rounds of smoothing */
	
	freeivec( depth ) ;
}	
/* ______________________________________________________ */

double local_loglike( double t_parent )
/*********************************************************
*	Compute the contributions to the log-likelihood
* of a tree due to the 4-member subtree under discussion
*/
{
	double		result, dr, r[3] ;

/*******************************************
*	Use the times and the branch lengths 
* to work out the rates.
*/
	lll_t[2] = t_parent ;
	
	r[0] = lll_len[0] / (lll_t[2] - lll_t[0]) ;
	r[1] = lll_len[1] / (lll_t[2] - lll_t[1]) ;
	r[2] = lll_len[2] / (lll_t[3] - lll_t[2]) ;
	
	if( r[0] == 0.0 )	r[0] = r[2] ;
	if( r[1] == 0.0 ) 	r[1] = r[2] ;
	if( r[2] == 0.0 )	r[2] = sqrt( r[0] * r[1] ) ;
	
/*******************************************
*	Compute the contributions to the
* log-likelihood.
*/
	dr = log( r[0] ) - log( r[2] ) + (0.5 * lll_v[0]) ;
	result = -(dr * dr) / lll_v[0] ;
	
	dr = log( r[1] ) - log( r[2] ) + (0.5 * lll_v[1]) ;
	result += -(dr * dr) / lll_v[1] ; 
	
	return( 0.5 * result ) ;
}
/* _______________________________________________________ */

double root_loglike( double t_root )
/*********************************************************
*	Compute the contributions to the log-likelihood
* of a tree due to the root and its children.
*/
{
	double		result, r[2], dr, var ;

/*******************************************
*	Use the times and the branch lengths 
* to work out the rates.
*/
	rll_t[2] = t_root ;
	
	r[0] = rll_len[0] / (rll_t[2] - rll_t[0]) ;
	r[1] = rll_len[1] / (rll_t[2] - rll_t[1]) ;
	
	if( r[0] == 0.0 )	r[0] = mean_root_rate ;
	if( r[1] == 0.0 ) 	r[1] = mean_root_rate ;
	
/*******************************************
*	Compute the contributions to the
* log-likelihood.
*/
	/* prior on rate on child branch zero */
	result = (rategamr - 1.0) * log(r[0])
				- rategamlam * r[0] ;

	/* conditional prior on branch 1 given zero */
	var = mean_nu * (rll_t[2] - 0.5 * (rll_t[0] + rll_t[1])) ;
	dr = log( r[0] )  - log( r[1] ) + (0.5 * var) ;
	
	result += -0.5 * (dr * dr / var) ;
	
	/* prior on age of root */
	result += (timegamr - 1.0) * log(rll_t[2])
				- timegamlam * rll_t[2] ;
	
	return( result ) ;
}
/* _______________________________________________________ */

double maximize_local_loglike( double *len, double *t, double *v )
/**************************************************************
*	Do a golden-section maximization on the local 
* contributions to the log-likelihood arising from a
* small {parent, grandparent and two kids} subtree.
* If we specify the lengths of the branches and the 
* ages of the kids and grandparent, the age of the
* parent is the only free parameter.  We return that
* value that gives rise to a maximum of log-likelihood.
*/
{
	double	t_min, t_max, alpha ;
	double	a, fa, b, fb, c, fc ;	/* bracketing triple */
	double	triple[3] ;

/***************************************************************
*	Copy the user's data into the global variables used
* in calculation of the local log likelihood.
*/
	memcpy( lll_len, len, 3 * sizeof(double) ) ;
	memcpy( lll_t, t, 4 * sizeof(double) ) ;
	memcpy( lll_v, v, 2 * sizeof(double) ) ;

/***************************************************************
* Try to work out some bracketing values. We want ,
*
*	a < b < c and f(a) < f(b)  and f(c) < f(b)
*/
	t_max = t[3] ;			/* set by age of grandparent	*/
	t_min = max( t[0], t[1] ) ; 	/* set by eldest child	*/

	b = t[2] ;	/* starting guess for parent's age */
	fb = local_loglike( b ) ;
	
	alpha = 0.01 ;
	do {
		a = t_min + (b - t_min) * alpha ;
		c = t_max - (t_max - b) * alpha ;
		
		fa = local_loglike( a ) ;
		fc = local_loglike( c ) ;
		
		alpha *= 0.1 ;
	} while( ((fa > fb) || (fc > fb)) && (alpha > DBL_EPSILON) ) ;
	
	if( alpha <= DBL_EPSILON ) {
		return( t[2] ) ;
	}
	
/*****************************************************************
*	Now that we've got a bracketing interval, go to it. 
*/
	triple[0] = a, triple[1] = b, triple[2] = c ;
	return( golden_maximizer( triple, local_loglike )  ) ;
}
/* _______________________________________________________ */

double maximize_root_loglike( double *time )
/**************************************************************
*	Do a golden-section maximization on the local 
* contributions to the log-likelihood arising from the
* root node and its two offspring.  We return that root-age
* which gives rise to a maximum of log-likelihood.
*/
{
	double	t_min, alpha ;
	double	a, fa, b, fb, c, fc ;	/* bracketing triple */
	double	triple[3] ;
	
/***************************************************************
*	Copy current data into the global variables used
* in calculation of the log likelihood near the root.
*/
	rll_t[0] = time[tonode[nnodes-1][1]] ;
	rll_t[1] = time[tonode[nnodes-1][2]] ;
	
	rll_len[0] = mean[tonode[nnodes-1][1]] ;
	rll_len[1] = mean[tonode[nnodes-1][2]] ;
	
/***************************************************************
* Try to work out some bracketing values. We want ,
*
*	a < b < c and f(a) < f(b)  and f(c) < f(b)
*/
	t_min = max( rll_t[0], rll_t[1] ) ; 	/* set by eldest child	*/

	b = time[nnodes-1] ;	/* starting guess for root's age */
	fb = root_loglike( b ) ;
	
	alpha = 0.5 ;
	do {
		a = t_min + alpha * (b - t_min) ;
		c = b / alpha ;
		
		fa = root_loglike( a ) ;
		fc = root_loglike( c ) ;
		
		alpha *= 0.717 ;
	} while( ((fa > fb) || (fc > fb)) && (alpha > DBL_EPSILON) ) ;
	
	if( alpha <= DBL_EPSILON )
		return( time[nnodes-1] ) ;
	
/*****************************************************************
*	Now that we've got a bracketing interval, go to it. 
*/
	triple[0] = a, triple[1] = b, triple[2] = c ;
	return( golden_maximizer( triple, root_loglike )  ) ;
}
/* ________________________________________________________________ */

double golden_maximizer( double *x, double (*f)(double x) )
/**************************************************************
*	Do a golden-section maximization on the function (*f)(),
* using  x[0] < x[1] < x[3] as the initial bracketing triple.
* Return the position of the maximum.
*/
{
	int		n_cycles ;		/* Used to avoid inifinite loops	*/
	
	double	tol ;			/* convergence tolerance */
	double	alpha ;			/* a constant related to the golden mean */
	double	a, b, c, d ; 	/* points in the bracketing interval */
	double	fa, fb, fc, fd ;	/* values of (*f)() */
	
/*****************************************************************
*	Set a < b < c, the bracxketing interval, with the
* user-suplied values.
*/
	a = x[0], fa = (*f)(a) ;
	b = x[1], fb = (*f)(b) ;
	c = x[2], fc = (*f)(c) ;
	
/*****************************************************************
*	Now that we've got a bracketing interval, go to it. 
*/
	tol = 10.0 * sqrt(DBL_EPSILON) * (c - a) ;	/* convergence tolerance */
	alpha = 0.5 * (3.0 - sqrt(5.0)) ;	/* (1 - gamma^{-1})	*/
		
	n_cycles = 0 ;
	while( ((c - a) > tol) && (n_cycles++ < MAX_CYCLES) ) {
	/****************************************
	*	Always subdivide the larger of
	* the two intervals (a,b) and (b,c)
	*/
		if( (b-a) > (c - b) ) {
		/******************************
		*	Take a < d < b < c 
		*/
			d = b - alpha * (b - a) ;
			fd = (*f)(d) ;
			if( fd > fb ) {	/* new triple is (a, d, b) */
				c = b, fc = fb ;
				b = d, fd = fb ; 
			}
			else {	/* new triple is (d, b, c) */
				a = d, fa = fd ;
			}
		}
		else {
		/******************************
		*	Take a < b < d < c 
		*/
			d = b + alpha * (c - b) ;
			fd = (*f)(d) ;
			if( fd > fb ) {	/* new triple is (b, d, c) */
				a = b, fa = fb ;
				b = d, fb = fd ;
			}
			else {	/* new triple is (a, b, d) */
				c = d, fc = fd ;
			}
		}
	}
	
	if( (c - a) >= tol ) 
		b = x[1] ;	/* no convergence, return original midpoint */

	return( b ) ;
}
/* ________________________________________________________________ */

double exptial( double explambda )
/**********************************************************
*	Return an exponentially distributed random value
* with the specified paramater. The mean of this variable
* is (1 / explambda).
*/
{
	double expvalue;
	
	assert( explambda > 0.0 ) ;	/* debugging paranoia */
	
	expvalue = -(1/ explambda) * log( 1 - std_uniform() ) ;
	
	return( expvalue ) ;
}
