/*********************************************************
*	A few utility functions used to read input files.
*
* mrm: LANL, 12 April 1999
*/
# include <stdio.h>
# include <string.h>
# include <stdlib.h>
# include <ctype.h>

# define DEBUG_LEVEL 2
# include "tune_debug.h"

# include "mrm_utils.h"

/*******************************************************
*	The chacters in DELIMITERS mark the beginning of
* a comment. Everything to their right is ignored.
* Also define some masks for character translation.
*/
# define DELIMITERS		"#%"
# define WHITE_SPACE	" \r\n\t"
# define PUNCTUATION		"=;"

/********************************************************
*	functions private to this file
*/
static void copy_n_punctuate(char *copy, char *orig);

/********************************************************
*	Variables used in this, and other, files.
*/
int				cmd_all_well = 1 ;
static int		n_tokens ;
static char		*my_tokens[MAX_TOKENS] ;
static char		local_line[3*LINE_BUF_SIZE] ;
/* ______________________________________________________________ */

void trim_comments( char *line )
/**********************************************************
*	Look for the leftmost occurence of a character from
* DELIMITERS and change it to the null character. This
* truncates the string, discarding anything to the right
* of the delimiter.
*
* N.B.  This function will run amok if applied to a
* non-terminated string.
*/
{
	char	*lpt ;
	
	if( line != NULL ) {
		lpt = line ;
		while( *lpt != '\0' ) {
			if(  strchr( DELIMITERS, *lpt ) != NULL ) 
				*lpt = '\0' ;
			else
				lpt++ ;
		}
	}
}
/* ______________________________________________________________ */

double get_double( FILE *fp )
{
	char	cmd_line[LINE_BUF_SIZE], *val_token ;
		
	val_token = get_next_value( cmd_line, LINE_BUF_SIZE, fp ) ;
	if( val_token != NULL )
		return( atof( val_token )  ) ;
	else
		return( 0.0 ) ;
}
/* ______________________________________________________________ */

int get_int( FILE *fp )
{
	char	cmd_line[LINE_BUF_SIZE], *val_token ;
			
	val_token = get_next_value( cmd_line, LINE_BUF_SIZE, fp ) ;
	if( val_token != NULL )
		return( atoi( val_token ) ) ;
	else
		return( 0 ) ;
}
/* _______________________________________________________ */

int get_tokens( char *tokens[], size_t max_tokens, char *line )
/************************************************************
*	Use strtok() to break a line up into space-separated 
* tokens.  Store them, up to a specified maximum number, 
* in tokens[], then return the number of tokens found.
*
* N.B. This function chops up the original line.
*/
{
	int		n_tokens ;
	char	*crnt_token ;

/***********************************
*	If debugging, check the args.
*/
	assert( tokens != NULL ) ;
	
/***********************************
*	Do the do.
*/
	n_tokens = 0 ;
	crnt_token = strtok( line, WHITE_SPACE ) ;
	while( (crnt_token != NULL) && (n_tokens < max_tokens) ) {
	/*********************************************************
	*	Store the current token, then get the next one.
	*/
		tokens[n_tokens++] = crnt_token ;
		crnt_token = strtok( NULL, WHITE_SPACE ) ;
	}
	
	return( n_tokens ) ;
}
/* ______________________________________________________ */

char *value_token( char *line )
/***********************************************************
*	If the line has any C-style comments, remove 'em.
* Then, if it has a "key = value" look about it, strip
* the "key =" away.
*/
{
	int		j, n_tokens ;
	char	*result ;

/***********************************************************
*	Make a local copy of the line and, at the same time,
* ensure that the punctuation characters appear as 
* separate, space-delimited tokens.
*/
	copy_n_punctuate( local_line, line ) ;

/***********************************************************
*	First remove comments; don't permit nesting.
*/
	remove_C_comments( local_line ) ;

/************************************************************
*	Split the line into space-separated tokens, then 
* look for "key = value" sequences and return the value.
*/
	n_tokens = line_to_tokens( local_line, my_tokens ) ;

	result = NULL ;		/* for now */
	if( n_tokens >= 3 ) {
		j = 1 ;
		while( j < (n_tokens - 1) && (result == NULL) ) {
			if( (my_tokens[j][0] == '=') && 
				(my_tokens[j][1] == '\0') )	result = my_tokens[j+1] ;
			else							j++ ;
		}
	}

	return( result ) ;
}
/* _______________________________________________________ */

int line_to_tokens( char *line, char *tokens[] )
/*******************************************************
*	Break the line up into space separated tokens
*/
{
	char	*crnt_token ;

	n_tokens = 0 ;
	crnt_token = strtok( line, WHITE_SPACE ) ;
	while( crnt_token != NULL ) {
		tokens[n_tokens++] = strip_ends( crnt_token ) ;
		crnt_token = strtok( NULL, WHITE_SPACE ) ;
	}

	return( n_tokens ) ;
}
/* ______________________________________________________ */

void remove_C_comments( char *line )
/***********************************************************
*	Remove C - style comments. This function doesn't
* do nested comments, but then neither does my compiler.
*/
{
	int		in_comment ;
	char	*lpt ;

	lpt = line ;
	in_comment = 0 ;
	while( *lpt != '\0' ) {
		if( !in_comment ) {
		/************************************************
		*	We're not yet inside a comment: check 
		* whether one is about to start, and if
		* not leave the current character unmolested.
		*/
			if( (lpt[0] == '/') && (lpt[1] == '*') )	{
			/********************************************
			*	Start of a C comment
			*/
				in_comment = 1 ;
				lpt[0] = lpt[1] = ' ' ;
				lpt += 2 ;
			}
			else if( (lpt[0] == '/') && (lpt[1] == '/') ) {
			/**********************************************
			*	A C++ comment: just chop everything
			* to the right.
			*/
				*lpt = '\0' ;
			}
			else	/* the most common case */
				lpt++ ;
		}
		else {
		/**********************************************
		*	Were inside a comment.
		*/
			if( (lpt[0] == '*') && (lpt[1] == '/') ) {
			/*******************************************
			*	We've reached the end of the comment
			*/
				in_comment = 0 ;
				lpt[0] = lpt[1] = ' ' ;
				lpt += 2 ;
			}
			else {
			/*******************************************
			*	Convert the interior of the comment
			* to white space.
			*/
				*lpt++ = ' ' ;
			}
		}
	}
}
/* ____________________________________________________ */

char  *get_next_value( char *line, int length, FILE *fp ) 
/*************************************************
*	Read 'til we encounter either a line with 
* something of interest on it or the end 
* of the file.
*/
{
	char	*val_token ;

	val_token = NULL ;
	while( (val_token == NULL) &&
			(fgets( line, length, fp ) != NULL) ) {
	/************************************************
	*	line[] may hold something interesting
	*/			
		val_token = value_token( line ) ;
	}
	
	if( feof(fp) ) {
	/********************************
	* We've run out of data
	*/
		cmd_all_well = 0 ;
	}
	else
		cmd_all_well = 1 ;

	return( val_token ) ;
}			 
/* _____________________________________________________ */

char *strip_ends( char *s )
/*******************************************************
*	Remove any leading or trailing white space from s.
*/
{
	register char	*spt, *lpt ;

	spt = s ;
	lpt = s ;

/*****************************************
*	Find the first non-space character.
*/
	while( isspace( *spt ) && (*spt != '\0') ) spt++ ;

/************************************************************************
*	Shift the characters of s backward so as to cover up leading spaces.
*/
	if( spt != lpt )  {
		while( *spt != '\0' ) 
			*lpt++ = *spt++ ;
	}
	else {
		lpt = s + strlen( s ) ;
	}
	
/************************************************
*	lpt now points to a character one position
*	beyond the end of the string;
*	run backward over the trailing white space.
*/
	lpt-- ;
	while( isspace(*lpt) && (lpt >= s) ) lpt-- ;
	
/*************************************************
*	Make sure the string is properly terminated.
*/
	lpt[1] = '\0' ;
	return( s ) ;
}
/* ______________________________________________________ */

void copy_n_punctuate( char *copy, char *orig )
/**************************************************************
*	Copy a string into a local buffer, expanding certain
* punctuation characters to ensure that they appear as
* distinct, space-delimited tokens.
*/
{
	char	*rpt, *wpt ;

	rpt = orig ;
	wpt = copy ;
	while( *rpt != '\0' ) {
	/**********************************************************
	*	Is *rpt a punctuation character?
	*/
		if( strchr( PUNCTUATION, *rpt ) != NULL ) {
		/******************************************************
		*	It is - surround it with white space in the copy
		*/
			*wpt++ = ' ' ;
			*wpt++ = *rpt++ ;
			*wpt++ = ' ' ;
		}
		else {
			*wpt++ = *rpt++ ;
		}
	}

/******************************************************
*	Make sure the copy is properly terminated.
*/
	*wpt = '\0' ;
}
/* ___________________________________________________________________ */

int read_val_list( FILE *fp, double *array, size_t vals_to_read )
/***************************************************************
*	Read a list of values into an array.  They should appear
* consecutively and be the only non-comment tokens on their 
* lines.  Return zero for success, non-zero for failure.
*/
{
	char	line[LINE_BUF_SIZE], *tokens[MAX_TOKENS] ;
	size_t	j, vals_read, n_tokens, tokens_read ;

/*************************************************************
*	Check the args and croak immediately if they look wrong.
*/
	if( fp == NULL ) {
		fprintf( stderr, "NULL file pointer in read_val_list().\n" ) ;
		return( 1 ) ;
	}
	else if( array == NULL ) {
		fprintf( stderr, "NULL array in read_val_list().\n" ) ;
		return( 2 ) ;
	}

/******************************************************************
*	Read space-separated tokens and convert 'em to floats 
* 'til all is well.
*/
	tokens_read = vals_read = 0 ;
	while( (vals_read < vals_to_read) &&
			(fgets( line, LINE_BUF_SIZE, fp ) != NULL) ) {
	/******************************************************
	*	Chop off comments and remove any leading 
	* or trailing white space, then break the line
	* into tokens.
	*/
		trim_comments( line ) ;
		strip_ends( line ) ;
		
		n_tokens = get_tokens( tokens, MAX_TOKENS, line ) ;
		
		for( j=0 ; j < n_tokens ; j++ ) 
			array[vals_read++] = atof( tokens[j] ) ;
			
		tokens_read += n_tokens ;		
	}
	
/***************************************************************
*	Check that we actually used all the tokens we read: 
* print a warning if not.
*/
	if( tokens_read > vals_read ) 
		fprintf( stderr, "Leftover tokens in read_val_list().\n" ) ;
		
	return( 0 ) ;	/* If we reach this line, all has gone well. */
}
/* _________________________________________________________ */

int dbl_cmpr( const void *a, const void *b )
/**********************************************************
*	A comparison function of the sort required by qsort.
*/
{
	double	*a_ptr, *b_ptr ;
	
	a_ptr = (double *) a ;
	b_ptr = (double *) b ;
	
	if( *a_ptr > *b_ptr )			return(  1 ) ;
	else if( *a_ptr == *b_ptr )		return(  0 ) ;
	else /* *a_ptr < *b_ptr */		return( -1 ) ;
}
