/**************************************************************
*	The functions below form a rudimentary progress monitor
* that works on Macs (under Metrowerks CodeWarrior) and under
* SunOS (tested with gcc on various Solaris boxes).
*
*	mrm: Oshkosh WI, 2 September 1999
*/
# include <stdio.h>
# include <string.h>
# include <time.h>		/* ++pac */
# include "progress_mon.h"	/* prototypes */

/*****************************************************************
*	These OS-dependent includes allow one to find the time
* used during a cycle of the progress monitor.
*/
# ifdef __MWERKS__
/*******************
*	MacOS
*******************/
	# include <time.h>
# else
/*******************
*	SunOS
*******************/
	# include <sys/resource.h>
	# include <sys/time.h>
# endif 

/****************************************************************
*	These control the way symbols denoting progress 
* are displayed.
*/
# define DEFAULT_SYMBOL		'*'

# define SYMS_PER_LINE		50
# define SYMS_PER_GROUP		5

# define SYMS_PER_CYCLE		100

/****************************************************************
*	These variables are shared throughout the program.
*/
static char		crnt_sym = DEFAULT_SYMBOL ;
static double	steps_per_sym, next_sym ;
static size_t	syms_printed, syms_on_line, syms_in_group ;

static double	start_time ;	/* in seconds */
static unsigned long	N_calls, max_calls ;
/* ________________________________________________________________ */

void init_progress_monitor( size_t n_calls, char new_sym )
/*******************************************************************
*	Set some constants so that this code will display SYMS_PER_CYCLE
* symbols during a period display_progress() gets called n_calls
* times.
*/
{
	char		*date_str ;
	time_t		now ;
	struct tm	*date ;

/********************************************
*	Set up various constants . . . 
*/
	steps_per_sym = n_calls / ((double) SYMS_PER_CYCLE) ;
	next_sym = steps_per_sym ;
	crnt_sym = new_sym ;
	
	N_calls = 0 ;
	max_calls = n_calls ;
	syms_on_line = syms_in_group = syms_printed = 0 ;
	
/************************************************
*	. . . print a message about the time.
*/
	now = time( NULL ) ;
	date = localtime( &now ) ;
	date_str = asctime( date ) ;
	date_str[strlen(date_str) - 1] = '\0' ;	/* chop terminating newline */
	
	printf( "\t<%s>\n\t", date_str ) ;
	
# ifdef __MWERKS__
/*************************************************
*  Under MacOS, make a note of a finer-grained  
* measure of the starting time.
*/
	start_time = ((double) clock()) / CLOCKS_PER_SEC ;
	
# else
/**************************************************
*	Do the same sort of thing under SunOS
*/
	{
		struct timeval		start ;
		
		gettimeofday( &start, NULL ) ;
		start_time = start.tv_sec +
						( start.tv_usec / 1000000.0 ) ;
	}
	
# endif 
}
/* ________________________________________________________________ */

double close_progress_monitor()
/***********************************************
*	Print a summary line, then return the
* number of seconds that has passed since
* init_progress_monitor() was called.
*/
{
	char			*date_str ;
	time_t			now ;
	struct tm		*date ;
	double			elapsed ;
		
/**********************************************
*	Print some stuff and return the number
* of seconds the process took, then return.
*/
	now = time( NULL ) ;
	date = localtime( &now ) ;
	date_str = asctime( date ) ;
	date_str[strlen(date_str) - 1] = '\0' ;	/* chop terminating newline */
	
	printf( "\n\t<%s>", date_str ) ;
	
/**********************************************
*	Note the time taken to complete the 
* current invocation of the progress monitor.
*/

# ifdef __MWERKS__	/* Under MacOS */

	elapsed = ((double) clock() / CLOCKS_PER_SEC) - start_time ;
	
# else  
#ifdef _Plan9_		/* ++pac */
	elapsed =  0.0;
#else
/***********************************************
*	Under SunOS, it's possible to get an
* account of the time used by the calling
* process alone.  
*/
	{
		struct rusage	resources_used ;
		
		getrusage( RUSAGE_SELF, &resources_used ) ;
		elapsed = resources_used.ru_utime.tv_sec +
					(resources_used.ru_utime.tv_usec / 1000000.0) ;
	}
# endif 
# endif /* elapsed time calculation */
		

	return( elapsed ) ;
}
/* ________________________________________________________________ */

void display_progress()
/********************************************************
*	Print blocks of the current symbol to stdout 
* to show progress.
*/
{	
	if( (++N_calls >= next_sym) && 
		(syms_printed < SYMS_PER_CYCLE) ) {
	/****************************************************
	*	Handle some formatting rubbish first
	*/
		if( syms_on_line == SYMS_PER_LINE ) {
			printf( "\n\t" ) ;
			syms_on_line = syms_in_group = 0 ;
		}
		else if( syms_in_group == SYMS_PER_GROUP ) {
			printf( " " ) ;
			syms_in_group = 0 ;
		}
		
	/*****************************************************
	*	Print the current symbol and update all
	* the counts.
	*/
		printf( "%c", crnt_sym ), fflush( stdout ) ;
		syms_printed++, syms_in_group++, syms_on_line++ ;
		next_sym += steps_per_sym ;
		if( next_sym > max_calls  )
			next_sym = max_calls ;
	}
}
