/*******************************************************************
*	The code below is derived from Jeff Thorne's divtime, 
* (VERSION 2/20/99) though it has been fairly heavily modified
* to permit the inclusion of constraints on the ages of the leaf
* nodes.  
*
* M. Muldoon:	Whalley Range, 23 August, 10-30 September,
*								31 October, 6 November
*				Oshkosh WI, 2 & 3 September 1999
*				Whalley Range, weekends in November & December 1999
*								then January and February 2000
*
* T. Battacharya:	LANL, Winter 2000, added lots of debugging code and 
*					fixed a bug in the TETHER_LEAVES sections.
*
********************************************************************
*	Original copyright notice:
* 
* (c) Copyright 1999 by Jeff Thorne
* 
* VERSION 2/20/99
* 
* As this program is very crude, please do not distribute it. Instead,
* people can obtain hopefully improved versions by contacting me (Jeff Thorne)
* (Also, please don't delete this message)
*  
* This version has the expected rate at a descendant node equal to the rate
* at the ancestral node - this is different than the previous version which
* had expected log rate at descendant equal to log rate of ancestor.
*/
# include <string.h>
# include <float.h>		/* mrm: for DBL_MAX */
# include <math.h>
# include <time.h>		/* mrm: for clock()	*/
# include <stdio.h>
# include <stdlib.h>

# define DEBUG_LEVEL 0		/* Next two lines control assertion checking	*/
# include "tune_debug.h"	/* see tune_debug.h for details					*/

# include "mersenne.h"
# include "random.h"
# include "tll.h"			/* replaces logcndmvn()	*/
# include "arrayutl.h"
# include "mrm_utils.h"		/* prototypes	*/
# include "platform.h"		/* redirect stdio on macs or unix boxen */
# include "div_switches.h"	/* compiler switches */
# include "progress_mon.h"	/* print blocks of symbols to show progress */
# include "tether_tree.h"	/* read input data	*/
# include "dt_startup.h"	/* generate initial conditions	*/
# include "age_prior.h"		/* prior on ages of internal nodes	*/

/**********************************************
*	Standard macros to find extrema.
*/
# define min2(A,B) ((A) < (B) ? (A) : (B))
# define min3(A,B,C) ((A) < (B) ? min2(A,C) : min2(B,C)) 
# define max2(A,B) ((A) > (B) ? (A) : (B))
# define max3(A,B,C) ((A) > (B) ? max2(A,C) : max2(B,C))

/**************************************************
* Choose a way to generate the initial data:
*
*	0	By choosing at random using the priors
*	1	By assuming rates close to constant
*	2	Similar to 1, but better suited to 
*		data where leaves are not measured
*		simultaneously.
*/
# define START_TYPE	2

/*********************************************
*	Used in switch_settings() below.
*/
# define SWITCH_STRLEN	128

/*********************************************
*	Prototypes for the functions below
*/
int main(MAIN_ARGS);
int rddata(char *treefilename);
void initialize(double *time, double *rate, int start_type);
void dothemcmc(double *time, double *rate, double **samples);
void inflate(double *time, double *rate, double *logr);
void changebrown(double *time, double *rate, double *logr);
void timechange(double *time, double *rate, double *logr);
void ratechange(double *time, double *rate, double *logr);
void summarize(double *first_sample, double **samples, 
						char *treefilename, char *nodefilename);
double probratio(double *proptime, double *proprate, double *proplogr, 
						double *oldtime, double *oldrate, double *oldlogr);
long seed_from_file(char *seedfile);
int params_from_file(char *param_file);
char *switch_settings(void);
/************** end of prototypes ***************************/

/************************************************************
*	Variables shared with tether_tree.c and dt_startup.c
*/
int 	nbranches, nnodes, ntips ;
int 	**tonode ;	/* tree topology: see tether_tree.h */

char 	**name ;	/* names of leaf nodes	*/
double	tip_date[MAX_TIPS], max_tip_date = 0.0 ;
double	*constraint, *mean ;	/* constraints and branch lengths */

double	mean_nu, mean_root_age, mean_root_rate ;		/* dt_startup.c	*/
double	timegamr, timegamlam, rategamr, rategamlam ;	/* these too */

Tether_Prior_Data	tethers[MAX_TIPS] ;	/* see tether_tree.h */

/********************************************************************
*	These variables are private to this file: most are parameters
* derived from the inputs, but which don't change later. 
*/
static double	browngamr, browngamlam ;

static double	tip_age[MAX_TIPS] ;			/* derived from date */

/********************************************************************
*	These variables are private to this file and are read from
* the parameter file.
*/
static int		burn_in_length ;
static int		cycles_per_sample, num_o_samples ;

static double	mean_root_date, root_age_sd ;
static double	root_rate_sd ;
static double	std_deev_nu ;

static double	node_depth_ab ;
static double	confidence ;

static double	time_K, mixing_K, root_K ;

static long		failrate[MAX_NODES], failtime[MAX_NODES], failflate ;
static long		sucrate[MAX_NODES], suctime[MAX_NODES], sucflate ;
/* ____________________________________________________________________ */

int main( MAIN_ARGS )	/* MAIN_ARGS is defined in platform.h */
{
	int		ktrack ;
	char	*my_argv[5] ;
	double	*time, *rate, **samples, *initial_sample ;

	unsigned long	initseed ;
	
/*************************************************************
*	Do any platform-dependent initialization.
*/
	MAIN_PREAMBLE
		
# ifdef PROFILE
/*************************************************************
*	Do anything needed to fire up profiling.
*/
	START_PROFILE
	
# endif /* if profiling */


# ifdef READ_COMMAND_LINE	/* The standard, non-debugging case */
/************************************************************
*	Do platform-dependent stuff to get command-line args
* and redirect stdio.
*/
	GET_COMMAND_LINE
	
# else
/************************************************************
*	Use standard files whose names are declared in 
* div_switches.h.
*/	
	argc = 5 ;
	argv = my_argv ;
	
	my_argv[0] = PROGRAM_NAME ;
	my_argv[1] = INPUT_FILE ;
	my_argv[2] = PARAM_FILE ;
	my_argv[3] = TREE_FILE ;
	my_argv[4] = NODE_FILE ;	
	
# endif /* assembling the command line */
				
/*************************************************************
*	Check the args and croak if there aren't enough.
*/
	if( argc != 5 ) {	
		fprintf( stderr, 
			"%s branch_infile param_file treefile nodefile >outfile\n",
			PROGRAM_NAME ) ;
			
		return( 1 ) ;
	}
	else {
	/***************************
	*	Read everything
	*/
		if( params_from_file( argv[2] ) ) 	return( 2 ) ;
		if( rddata( argv[1] ) )				return( 3 ) ;
	}
	
/***********************************************************
*	Print out a string that describes the choices
* made in div_switches.h
*/
	printf( "%s\n\n", switch_settings() ) ;
			
/************************************************************
*	Convert the means and standard deviations specified for 
* the gamma-distributed priors into the standard gamma
* parameters.  The gamma distrib is 
*
*	f(x) = C( r, \lambda )  * x^{r-1} \exp{ x * \lambda }
*
* with mean and variance given by
*
* 	mean = (r / \lambda)
*	var = std_deev^2 = (r / \lambda^2)
*/
	browngamr = (mean_nu * mean_nu) / (std_deev_nu * std_deev_nu) ; 
	browngamlam = mean_nu / (std_deev_nu * std_deev_nu) ;
	
	timegamr = (mean_root_age * mean_root_age) / (root_age_sd * root_age_sd) ;
	timegamlam = mean_root_age / (root_age_sd * root_age_sd) ;
	
	rategamr = (mean_root_rate * mean_root_rate) / (root_rate_sd * root_rate_sd) ;
	rategamlam = mean_root_rate / (root_rate_sd * root_rate_sd) ;
	
/*************************************************************
*	Seed the random number generator with either a previous
* seed stored in a file (useful for debugging) or from the 
* clock for non-debugging, production code (NDEBUG # defined).
*/

# ifdef NDEBUG	
	initseed = (unsigned long) clock() ;
# else 
	initseed = seed_from_file( SEED_FILE ) ;
# endif
	
	init_random( initseed ) ;
	
/*************************************************************
*	Echo the parameters to stdout
*/	
	printf("initseed = %ld\n", initseed ) ;
	
	printf("NEWK= %6.3f\n", time_K ) ;
	printf("SAMPFREQ %6d NUMSAMPS %6d BURNIN %6d\n", 
				cycles_per_sample, num_o_samples, burn_in_length ) ;
	printf("OTHK %6.3f THEK %6.3f\n", mixing_K, root_K ) ;
	printf("BROWNMEAN %6.3f BROWNSD %6.3f\n", mean_nu, std_deev_nu ) ;
	printf("browngamr %6.3f browngamlam %6.3f\n", browngamr, browngamlam) ;
	
	printf("RTTM %6.3f RTTMSD %6.3f MINAB %6.3f\n", 
			mean_root_age, root_age_sd, node_depth_ab ) ;

	printf("timegamr %6.3f timegamlam %6.3f\n", timegamr, timegamlam ) ;
	printf("RTRATE %8.5f RTRATESD %8.5f\n", mean_root_rate, root_rate_sd ) ;
	printf("rategamr %6.3f rategamlam %6.3f\n", rategamr, rategamlam ) ;

	printf("nbranches = %d, nnodes = %d, ntips = %d\n", 
										nbranches, nnodes, ntips ) ;
	
/*******************************************************************
*	Build arrays to hold the ages of the nodes and the rates on
* the branches, then initialize the arrays used to tabulate 
* success and failure of the MCMC algorithm.
*/
	time = dvec( nnodes ) ;
	rate = dvec( nnodes ) ;
	for(  ktrack =0 ;  ktrack  <= nnodes ;  ktrack ++ ) {
		sucrate[ ktrack ] = 0 ;
		suctime[ ktrack ] = 0 ;
		failrate[ ktrack ] = 0 ;
		failtime[ ktrack ] = 0 ;
	}
	sucflate = 0 ;
	failflate = 0 ;
	
/*****************************************************************
*	If we are applying the slice-volume prior (see age_prior.c)
* attempt the necessary initialization here.
*/
# if defined(SLICE_VOL_PRIOR)
	
	if( init_age_prior() )
		return ( 4 ) ;	/* return non-zero for failure */
		
# endif

/*******************************************************************
*	Print the dates associated with the leaves.
*/
	printf( "Constraints on Leaves\n" ) ;
	for( ktrack=0 ; ktrack < ntips ; ktrack++ ) {
	
# if defined( TETHER_LEAF_AGES )
		printf("(%3d) %8.2f ",
			ktrack, (max_tip_date - constraint[ktrack]) ) ;
			
# else	/* FREEZE_LEAF_AGES: constraint on leaf is enforced on parent node */

	printf( "(%3d) %8.2f ",
			ktrack, (max_tip_date - constraint[tonode[ktrack][PARENT]]) ) ;

# endif
			
		if( (ktrack+1)%5 == 0 ) { printf("\n") ; }
	}
	
	printf( "\n\n" ) ;
	
/*******************************************************************
*	Print out the node constraints. If there aren't any, they
* will all appear as zero.
*/
	printf("Constraints on internal nodes\n");	
	for( ktrack = ntips ; ktrack < nnodes ; ktrack++ ) {
		printf("(%3d) %8.2f ",
			ktrack, (max_tip_date - constraint[ktrack]) ) ;
			
		if( (ktrack-ntips+1)%5 == 0 )	printf("\n") ; 
	}
	
	printf("\n\n");
	

/******************************************************************
*	Make a starting guess and record it in initial_sample,
* the plunge into the main work.
*/
	samples = dmat( num_o_samples, 2*nnodes+5 ) ;
	initial_sample = dvec( 2*nnodes+5 ) ;
	
	printf("Before initialize\n");
	initialize(time, rate, START_TYPE ) ;
	for( ktrack = 0 ; ktrack < nnodes ; ktrack++ ) {
		initial_sample[ktrack] 			= time[ktrack] ;
		initial_sample[ktrack + nnodes] = rate[ktrack] ;
	}
	
	dothemcmc( time, rate, samples) ;
	
/*************************************************
*	Write the results to various files.
*/
	summarize( initial_sample, samples, argv[3], argv[4] ) ;

/*************************************************
*	Tidy up and return.
*/
	freedmat(samples,num_o_samples,2*nnodes+5);
	freedvec(time);
	freedvec(constraint);
	freedvec(rate);
	freedvec(mean);
	freecmat(name,MAX_TIPS, NAME_LEN);
	freeimat(tonode, MAX_NODES, (MAX_KIDS + 2));
	
# ifdef PROFILE
/*************************************************************
*	Do anything needed to conclude profiling.
*/
	FINISH_PROFILE
	
# endif /* if profiling */

	
	return( 0 ) ; 	/* Zero means success */
}
/* __________________________________________________________________ */

int rddata( char *treefilename )
/*************************************************************
*	Read the tree topology and the derivs from the named
* file.
*/
{
	int		j ;
	FILE	*treefile	;
 
/************************************************************
*	Try to open the input file and die if unsuccessful
*/
	treefile = fopen( treefilename, "r" ) ;
	if( treefile == NULL ) {
		fprintf( stderr, "Can't open the tree file %s.\n", treefilename ) ;
		return( 1 ) ;
	}
	
/**************************************************************
*	Read the tree topology, constraints and, if they're
* present, the priors on the tips.
*/
	if( read_tree( treefile ) != 0 ) {
		fprintf( stderr, "Trouble reading the tree.\n" ) ;
		return( 2 ) ;
	}
	
/**************************************************************
*	Read the details of a quadratic approximation to the
* log-likelihood as a function of branch length.
*/
	if( file_init_tree_loglike( nbranches, treefile )  ) {
		fprintf( stderr, "Trouble reading derivs.\n" ) ;
		return( 3 ) ;
	}
	else {
	/*************************************************
	*	Get the branch lengths from the tree_loglike
	* machinery.
	*/
		tll_copy_max_pos( mean ) ;
	}
	
/*************************************************************
*	Convert all dates and constraints into ages.
*/	
	for( j=0 ;  j < ntips ; j++ ) 
		tip_age[j] = max_tip_date - tip_date[j] ;
		
	for( j = ntips ; j < nnodes ; j++ ) {
		if( constraint[j] != 0.0 ) {
			constraint[j] = max_tip_date - constraint[j] ;
		}
	}
	
	mean_root_age = max_tip_date - mean_root_date ;
		
# if defined( TETHER_LEAF_AGES )
/*********************************************************
*	Record the parameters of the prior on the age of
* the j-th leaf. It's a convolution between an 
* exponential, exp( -lambda * t ), that decays as one
* goes into the past, and a square window,
*
* 		t_0 <= t <= t_0 + window.
* 
* Up to (unimportant) normalizing factors this is
*
* f(t) = 	1.0 - exp( - lambda * (t - t_0) ) 
*			if t is in the interval above
*
*			[1.0 - exp( - lambda * window )]
*			* exp( - lambda * (t - t_0 - window) )
*			if t >= (t_0 + (window))
*/
	for( j=0 ; j < ntips ; j++ ) {
		tethers[j].win_min = tip_age[j] - (0.5 * tethers[j].win_width) ;
		tethers[j].win_max = tip_age[j] + (0.5 * tethers[j].win_width) ;
		
		/* constant for the decaying exponential part */
		tethers[j].max_val = 1.0 - 
			exp( -(tethers[j].lambda * tethers[j].win_width) ) ;
	}
	
# endif	/* TETHER_LEAF_AGES */

/****************************************************************
* 	The following block propagates constraints from the
* nodes on which they are imposed backward through their 
* parents and so on to the root node.
****************************************************************/

/****************************************************************
*	Install constraints on the tips (TETHER_LEAF_AGES) or on
* their parents (FREEZE_LEAF_AGES).
*/		
	for( j=0 ; j < ntips ; j++ ) {
	/***********************************************
	*	Impose the constraint.
	*/ 
				
# if defined( FREEZE_LEAF_AGES )

		constraint[j] = 0.0 ;	/* Tip is frozen, so true constraint is on parent */ 
		if( constraint[tonode[j][PARENT]] < tip_age[j] )
			constraint[tonode[j][PARENT]] = tip_age[j] ;
				
# else /* TETHER_LEAF_AGES */

		constraint[j] = tethers[j].win_min ;		/* Tips can move too.	*/
# endif 

	}	/* End of loop that imposes user-supplied constraints on leaves */

/***************************************************************
*	Propagate constraints backward toward the root. Those
* on the tips, as well as any user-imposed constraints,
* will already be in place.
*/		
	for( j = 0 ; j < (nnodes - 1) ; j++ ) {
		constraint[tonode[j][PARENT]] = max2(
			constraint[tonode[j][PARENT]], constraint[j]);
	}

	fclose(treefile);
	
	return( 0 ) ;	/* If we're here, success! */
}
/* _______________________________________________________ */

void initialize( double *time, double *rate, int start_type )
{	
	int		badcon, j, ktrack ;

/********************************************************
*	Install the ages of the leaves as specified by the
* user.  Even if the ages of the leaves are applied 
* as priors or inequality constraints rather than 
* equalities, this approach ought to produce a 
* reasonable starting guess.
*/ 
	for( ktrack=0 ; ktrack < ntips ; ktrack++ ) 
		time[ktrack] = tip_age[ktrack] ;
		
/********************************************************
*	Try to generate a reasonable initial condition
*/
	switch( start_type ) {
	/******************************************************************
	*	Three options are available. The first (start_type = 0)
	* is a purely random start  with parameter values drawn from
	* the appropriate prior distribution. The second (start_type = 1) 
	* method, originally developed by Jeff Thorne, converts distances 
	* to the root node (that is, summed branch lengths) to estimated ages. 
	* If all tips were measured at t=0, then this is akin to assuming a 
	* molecular clock. Finally there is a scheme (start_type = 2) that 
	* estimates a single constant overall rate and the age of the root 
	* node. It only makes sense when the tips were measured at distinct 
	* times.
	*/
		case 0 :
			random_start( time, rate ) ;			
			break ;
	
		case 1 :	/* J. Thorne's original constant-rate approx. */
			mol_clock_start( time, rate ) ;
			break ;
			
		case 2 :	/* mrm's least-squares fit to a constant rate */
			least_sq_start( time, rate ) ;
			smooth_rates( time, rate ) ;
			break ;
						
		default:
			fprintf( stderr, "Eh? Unrecognized start-type in initialize().\n" ) ;
			exit( 1 ) ;
		
	} /* End of switch{} that generates initial guesses */
	
/********************************************************************
*	This section deals with discrepancies between the times
* chosen above and any user-imposed constraints. The idea is
* to sweep through the tree looking for nodes that aren't as old 
* as they are constrained to be. Each time we find such a node 
* we apply a uniform increase in time scale to the internal nodes, 
* pushing the offending node deeper into the past.
*
*	We adopt the same strategy even in the cases when there are
* constraints on the tips: any tip-age data that we've placed in 
* time[] should go unmolested by the expansion of time scale. 
*/
	do { 
	/****************************************************************
	*	Look for nodes that are younger than their constrained age.
	*/
		badcon = 0 ;	/* assume all is well */
		for(ktrack=ntips; ktrack <= nnodes-1; ktrack++) {	
			for( j=0 ; (j < tonode[ktrack][N_KIDS]) && !badcon ; j++ ) {
				if( time[ktrack] <= time[tonode[ktrack][j+1]] ) 
					badcon = 3 ;
			}
		}
		
	/***************************************************************
	*	If any problem nodes were found, expand the time scale
	* by a factor of EXPAND_TIME and readjust rates accordingly.
	*/
		if( badcon > 1 ) {
		/**************************************************************
		*	Expand the time scale for all internal nodes. 
		*/
			for(ktrack = ntips ; ktrack <= nnodes-1 ; ktrack++ ) 
			{
				time[ktrack] *= EXPAND_TIME ;
			}
			
		/**************************************************************
		*	The ages of the leaf nodes, which may have been fixed
		* by the user, won't have changed so we must recalculate
		* the rates to make sure they're consistent with the branch
		* lengths.
		*/
			for(ktrack = 0; ktrack < nbranches; ktrack++) 
			{
				rate[ktrack] =max2( SHORT_BRANCH, mean[ktrack])/(
					time[tonode[ktrack][PARENT]]- time[ktrack]);
			}

		/*************************************************************
		*	We have changed all the times, so must ensure that we 
		* check all the constraints one more time.
		*/
			badcon = 1 ;
			
		}	/* end of code that handles a failed constraint */
		
	} while(badcon != 0) ;
}
/* __________________________________________________________________ */

void dothemcmc( double *time, double *rate, double **samples )
/**************************************************************
*	This is the main function: it performs repeated cycles
* of the Markov Chain Monte Carlo algorithm, updating all the
* rates and node ages each time.
*/
{
	int 	whichsamp, ktrack, ktrack2 ;
	FILE 	*sampfile = NULL ;
	double	time_in_secs ;
	double	*logr ; /* save time by precalculating log of rates */

	void timechange();
	void ratechange();
	void changebrown();
	void inflate();

	logr = dvec(nnodes);
	for(ktrack = 0; ktrack < nbranches; ktrack++)
	{
		logr[ktrack] = log(rate[ktrack]);
	}

	printf( "\nStarting burn-in:\n" ) ;
	init_progress_monitor( burn_in_length, '#' ) ;
	for(ktrack = 0; ktrack <= burn_in_length + (num_o_samples-1)*cycles_per_sample; ktrack++)
	{
		if( (ktrack >= burn_in_length) && ((ktrack - burn_in_length)%cycles_per_sample == 0 ))
		{
			if( ktrack == burn_in_length ) {
			/*************************************************************
			*	When the burn-in period is complete, set up the
			* progress monitor to display progress through the main
			* part of the computation.
			*/
				time_in_secs = close_progress_monitor() ;
				printf( "\nFinished in %.4f seconds, %.6f per MCMC cycle.\n", 
				time_in_secs, (time_in_secs / burn_in_length) ) ;
				
				printf( "\nCollecting samples:\n" ) ;
				init_progress_monitor( 
					((size_t) (num_o_samples-1))*cycles_per_sample, '*' ) ;
			}
			
			if( sampfile == NULL ) {
				sampfile = fopen(SAMP_FILE,"a") ;
				if( sampfile == NULL ) {
					fprintf( stderr, "Can't open the sample file %s.\n", SAMP_FILE ) ;
					exit( 1 ) ;
				}
			}
			
			for(ktrack2=0; ktrack2<nnodes; ktrack2++)
			{
			    if ( sampfile != NULL ) 
			    	fprintf(sampfile, "%10.5f ", time[ktrack2]);
			    	
				whichsamp = (ktrack-burn_in_length)/cycles_per_sample;
				samples[whichsamp][ktrack2] = time[ktrack2];
				samples[whichsamp][ktrack2+nnodes] = rate[ktrack2];
			}
			
			if ( sampfile != NULL ) {
				for(ktrack2=0; ktrack2<nnodes; ktrack2++)
					{
						fprintf(sampfile, "%10.5f ", rate[ktrack2]);
					}

				fprintf(sampfile,"\n"), fflush(sampfile) ;
			}
			/* fclose(sampfile) ;	save time by leaving the sample file open */
			
			# if (defined(__MWERKS__) && defined(GOOD_MACOS_CITIZEN))
			/******************************************************
			*	On the Mac, make a minimal gesture toward
			* cooperative multi-tasking.
			*/
				SIOUXHandleOneEvent( NULL ) ;
		
			# endif
		}
		
		timechange(time,rate,logr);
		ratechange(time,rate,logr);
		changebrown(time,rate,logr);
		inflate(time,rate,logr);
		
		display_progress() ;
	}
		
/***********************************************************
*	Report the average time for an MCMC cycle and return.
*/
	time_in_secs = close_progress_monitor() ;
	printf( "\nFinished in %.4f seconds, %.6f per MCMC cycle.\n\n", time_in_secs,
		(time_in_secs / ((num_o_samples-1)*cycles_per_sample)) ) ;

	freedvec(logr);
}
/* _____________________________________________________________________ */

void inflate( double *time, double *rate, double *logr )
/***************************************************************
*	This function implements the "mixing" step described 
* at the end of Thorne et al.'s appendix on their algorithm. 
* One applies a uniform rescaling 
*
*	M = exp( H_4 (u - 0.5) )
*
* to the ages of all the internal nodes. Here u is uniform
* in [0,1) and H_4 is the parameter whose default value,
* mixing_K, is set at the top of the file.  In the
* original code, the rates also got adjusted accordingly, 
*
*	T_j -> M T_j;  R_j -> (R_j / M)
*
* so the branch lengths L_j = (T_j * R_j) remain the same.
* This step is slightly more complicated when the ages of some 
* nodes are constrained.
*/
{
	int		maxcurnode, maxnewnode, ktrack ; 
	double	ratio, randnum, invfac, *newtime,flatefac, *newrate ;
	double	jterm, maxcurratio, maxnewratio, curdummy, newdummy, lcur, lnew ;
	double	minflatefac, loginvfac, *newlogr, Jacobian_factor ;
	
	static double	*logcon, *thecon, invothk ;
	static int		hereyet = 0, numcons, *nodecon ;
	
/**************************************************************
* 	On the first pass through inflate(), count the primary
* constraints (numcons). A constraint is primary if it has
* been imposed directly by the user.  The alternative is that
* the constraint has been propagated backward from a 
* constrained child node.
* 
* This loop tabulates the primary constraints, along with
* their logs and the indices of the nodes to which they they
* apply. Finally, it sets a flag, hereyet, to indicate these
* tasks have been finished once and for all. 
*/
	if( hereyet == 0 ) {
	/*********************************************************
	*	Allocate memory for the results we're about to
	* find, then set some constants.
	*/
		logcon = dvec(nnodes), assert( logcon != NULL ) ;
		thecon = dvec(nnodes), assert( thecon != NULL ) ;
		nodecon = ivec(nnodes), assert( nodecon != NULL ) ;

		numcons = 0 ;
		invothk = (1.0 / mixing_K) ;
		
		hereyet = 1 ;	/* been here, done this */

	/**************************************************************
	*	We need to keep track of the number of times able to
	* change under the rescaling. In the FREEZE_LEAF_AGES 
	* case it will be (nbranches - ntips), while in the
	* TETHER_LEAF_AGES all will move.
	*/

# ifdef TETHER_LEAF_AGES 
	/************************************************************
	*	There may be primary constraints on the leaf nodes. 
	*/
		for( ktrack = 0 ; ktrack < ntips ; ktrack++ ) {
		/********************************************************
		*	Leaves may carry primary constraints.
		*/
			if( constraint[ktrack] > 0.0 ) {
				numcons++ ;
				nodecon[numcons] = ktrack ;
				thecon[numcons] = constraint[ktrack] ;
				logcon[numcons] = log( constraint[ktrack] ) ;
			}
		}
		
# endif 	/* TETHER_LEAF_AGES */

	/************************************************************
	*	Look for primary constraints on the internal nodes.
	*/
		for( ktrack = ntips ; ktrack < nnodes ; ktrack++ ) 	
		{
			if(constraint[ktrack] != constraint[tonode[ktrack][1]] &&
				constraint[ktrack] != constraint[tonode[ktrack][2]])
			{
				numcons++;
				thecon[numcons] = constraint[ktrack];
				nodecon[numcons] = ktrack;
				logcon[numcons] = log(constraint[ktrack]);
			}			
		}
		
	}	/* end of initialization */
	
/***************************************************************
*	mrm: the code below this line gets executed on every 
* time inflate() is invoked.
****************************************************************/

	if( numcons == 0 ) {
		jterm = 1.0;
	}
	else {
	/***********************************************************
	*	Find that constraint which is most nearly violated.
	* Examine the ratios (constrained_age/age) and find the
	* biggest.  All these ratios should be < 1.0.
	*/
		maxcurratio = thecon[1]/time[nodecon[1]] ;	
		maxcurnode = 1 ;
		for(ktrack = 2 ; ktrack <= numcons ; ktrack++ )
		{
			curdummy = thecon[ktrack]/time[nodecon[ktrack]];
			if(curdummy > maxcurratio)
			{
				maxcurratio = curdummy;
				maxcurnode = ktrack;
			}
		}
		
	/***********************************************************
	* 	The idea here is to prevent the time scale from  
	* contracting so sharply that some of the constraints 
	* are violated. All constraints are of the form "A node   
	* is at least a certain age" so that only contraction of 
	* the time scale presents a potential problem.  
	*
	*	Say T is the current age of a node and T_c is
	* the constraint on its age. The factor by which we
	* rescale time satisfies
	*
	*		M >= exp( mixing_K * (0.0 - 0.5) )
	* or	log(M) >=  -0.5 * mixing_K
	*
	* But the constraint requires
	*
	*		MT >= T_c
	* or	log( M ) >= log( T_c / T ) 
	*
	* Hence must exercise caution if
	*
	*		-0.5 * mixing_K < log( T_c / T )
	* or	mixing_K > 2.0 * (log(T) - log(T_c))
	*
	*/
		lcur = log(time[nodecon[maxcurnode]]);
		if(mixing_K <= 2.0*(lcur-logcon[maxcurnode]))
		/*******************************************************
		*	All is well, no need to worry about excessive 
		* shrinking of the time scale.
		*/
		{
			jterm = 1.0 ;
		}
		else {
		/*******************************************************
		*	Excessive shrinkage is possible: here we find that
		* part of the range of the uniform random variable
		* 0 <= u < 1 that gives rise to acceptable rescalings.
		* The smallest acceptable value of u satisfies 
		* 
		*		(u_{min} - 0.5) * mixing_K = log( T_c / T )
		* or u_{min} = 0.5 + log( T_c / T )*(1 / mixing_K)
		*
		* The length of the intrval of acceptable u's is 
		* thus (1.0 - u_{min})
		*
		*	= 1.0 - [0.5 + log( T_c / T ) * (1 / mixing_K)]
		*	= 0.5 - log( T_c / T ) * (1 / mixing_K)
		*	= 0.5 - log( T / T_c ) * (1 / mixing_K)
		*/
			jterm = 0.5 + (lcur-logcon[maxcurnode])*invothk;
		}
	}
	
/*****************************************************************
*	Allocate memory to hold the new times, rates and logs-o-rates
*/
	newtime = dvec(nnodes);
	newrate = dvec(nnodes);
	newlogr = dvec(nnodes);

/*****************************************************************
*	In light of the discussion above, compute the minimal time
* rescaling compatible with the constraints. If there are none, 
* this factor is zero.
*/
	if(numcons==0)
	{
		minflatefac = 0.0;
	}
	else
	{
		minflatefac = maxcurratio;
	}
	
/*****************************************************************
*	Now select random rescaling (inflation) factors until
* an acceptably large one is found.
*/
	do {
		flatefac = exp(mixing_K*(std_uniform()-0.5));
	} while(flatefac < minflatefac) ;
	invfac = 1.0/flatefac;
	loginvfac = log(invfac);
	
/**********************************************************
*	Having found a suitable inflation factor, apply it
* to the times and rates. 
*
* 	Note that although we rescale the age of the root 
* node along with all the others, we leave the associated
* rate unchanged. That's because it isn't really a rate
* at all; it's the brownian motion parameter \nu.
**********************************************************/

/********************************************************
*	If the tips are frozen, hold their ages constant,
* but otherwise inflate their ages by the same factor
* as all the other nodes.
*/

# if defined( FREEZE_LEAF_AGES )
	for(ktrack = 0 ; ktrack < ntips ; ktrack++)
	{
		newtime[ktrack] = time[ktrack];
	}

	for(ktrack=ntips; ktrack < nbranches; ktrack++)

# else /* TETHER_LEAF_AGES */

	for(ktrack=0; ktrack < nbranches; ktrack++)

# endif
	{
		newtime[ktrack] = flatefac * time[ktrack] ;
	}
	
/*********************************************************
*	Allow the age of the root to change, but do not 
* change the parameter nu, which is stored in 
* rate[nbranches]. 
*/
	newrate[nbranches] = rate[nbranches];	/* preserves nu */
	newtime[nnodes-1] = flatefac*time[nnodes-1];

/*********************************************************
*	Now adjust the remaining rates in such a way as
* to keep the branch lengths constant.
*/	
	{ 
		double	blen_ratio, blen, new_blen ;
		
		Jacobian_factor = 1.0 ;
		for(ktrack=0; ktrack < nbranches; ktrack++)
		{
		
		# ifndef NDEBUG /* this double negative means "if debugging" */
		
			if( newtime[tonode[ktrack][PARENT]] < newtime[ktrack] ) {
				printf( "Yikes! Parent %d younger than child %d: ",
							tonode[ktrack][PARENT], ktrack ) ;
				printf( "%.2f < %.2f!? (diff = %.4e)!\n",
					newtime[tonode[ktrack][PARENT]], newtime[ktrack],
					fabs(newtime[tonode[ktrack][PARENT]] - newtime[ktrack]) ) ;
			}
			
		# endif 

			assert( time[tonode[ktrack][PARENT]] >= time[ktrack] ) ;				
			assert( newtime[tonode[ktrack][PARENT]] >= newtime[ktrack] ) ;
			
			blen = time[tonode[ktrack][PARENT]] - time[ktrack] ;
			new_blen = newtime[tonode[ktrack][PARENT]] - newtime[ktrack] ;
			if( new_blen > (TINY_FRAC * time[ktrack]) ) {
				blen_ratio =  blen / new_blen ;
			}
			else if( blen < (TINY_FRAC * time[ktrack]) ) {
			/*******************************************************
			*	Both new and old (temporal) branch lengths 
			* are essentially zero.
			*/
				blen_ratio = 1.0 ;
			}
			else {
			/*******************************************************
			*	Only the new rate is truly tiny.
			*/
				blen_ratio = DBL_MAX ;
			}
						
			newrate[ktrack] = blen_ratio * rate[ktrack] ;
			newlogr[ktrack] = log( blen_ratio ) + logr[ktrack] ;
			
			Jacobian_factor *= blen_ratio ;
		}
	}
	
/****************************************************************
*	Survey the new times and again find that constraint which
* is most nearly violated. One needs this for the ratio of
* proposal densities.
*/
	if(numcons != 0)
	{
		maxnewratio = thecon[1]/newtime[nodecon[1]];
		maxnewnode = 1;
		for(ktrack = 2; ktrack <= numcons; ktrack++)
		{  
			newdummy = thecon[ktrack]/newtime[nodecon[ktrack]];
			if(newdummy > maxnewratio)
			{
				maxnewratio = newdummy;
				maxnewnode = ktrack;
			}
		} 
		
		lnew = log(newtime[nodecon[maxnewnode]]);
		if(mixing_K > 2.0*(lnew - logcon[maxnewnode])) {
		/********************************************************
		*	Reasoning as above.
		*/
			jterm /= 0.5+(lnew - logcon[maxnewnode])*invothk;
		}
	}
	
/****************************************************************
*	Decide whether to accept or reject the rescaling. Here
* one wants to impose a detailed balance condition:
*
*	P( making change ) = P( reversing change ) 
*
* or
* 
*	P( inflating current tree by flatefac ) * P(current tree) = 
*	P( inflating resulting tree by invfac ) * P(resulting tree) 
*
* The first factors on each side of the = are both of the form
*
*	P( proposing inflation factor ) * P( accepting proposal )
*
* which leads to
*
*	P( accepting flatefac for current tree ) =
*		{
*			[P(resulting tree) / P(current tree)] *
*			[P(proposing invfac in resulting tree) / P(proposing flatefac in current tree)]
*		}
*		* P( accepting invfac for resulting tree )
* 
* We can compute everything here save the P(accepting . . .) factors
* and we choose can these to make the equation true: if the stuff in curly braces
* is less than 1, we choose
* 
*	P( accepting flatefac for current tree ) = { stuff in curly braces }
*	P( accepting invfac for resulting tree ) = 1.0 
*
* otherwise 
* 
*	P( accepting flatefac for current tree ) = 1.0
*	P( accepting invfac for resulting tree ) = 1.0 / { stuff in curly braces }
*
* The ratio includes a factor from the Jacobian of the (diagonal, linear) map 
* that does the inflation. In the TETHER_LEAF_AGES case the Jacobian includes
* (nnodes) factors of flatefac from the ages of the nodes and (nnodes - 1) 
* factors of invfac = (1 / flatefac) from the rates for a product of flatefac.
* In the FREEZE_LEAF_AGES case, there are only (nnodes - ntips) factors of
* flatefac and the factors due to the changing rates are more complicated,
* so are computed and multiplied into Jacobian factor above.
*/

# if defined( FREEZE_LEAF_AGES )	/* factors due to changing rates computed above */
	Jacobian_factor *= exp( (nnodes - ntips) * log( flatefac ) ) ;
# else
	Jacobian_factor = flatefac ;	/* changed age of root, but no corresponding rate */
# endif 

	ratio =  Jacobian_factor * jterm
		* probratio(newtime,newrate,newlogr,time,rate, logr) ;
	/* fprintf(stderr,"DEBUG2: ratio = %f\n",ratio); */
		
	randnum = std_uniform() ;
	if(randnum < ratio) {
		for( ktrack = 0 ; ktrack < nnodes ; ktrack++ ) { 
			rate[ktrack] = newrate[ktrack] ;
			time[ktrack] = newtime[ktrack] ;
		}
		for( ktrack = 0 ; ktrack < nbranches ; ktrack++ ) {
			logr[ktrack] = newlogr[ktrack] ;
		}

		sucflate++ ;
	}	
	else {
		failflate++ ;
	}

	freedvec(newrate) ;
	freedvec(newlogr) ;
	freedvec(newtime) ;
}
/* _________________________________________________________ */

void changebrown( double *time, double *rate, double *logr )
/****************************************************************
*	Do an MCMC step that proposes a change in the brownian 
* motion constant - the parameter called \nu in Thorne et al.
*
* N.B. This value is stored in rate[nbranches].
*/
{	
	double	ratio, randnum, *newrate;
	size_t	bytes_to_copy ;

/****************************************************************
*	Duplicate the list of rates so as to make a call to
* probratio().
*/
	newrate = dvec( nnodes ), assert( newrate != NULL ) ;
	bytes_to_copy = nbranches * sizeof( double ) ;
	memcpy( newrate, rate, bytes_to_copy ) ;
	
/*****************************************************************
*	Propose a new (!) value for \nu and store 
* it in rate[nbranches].
*/
	newrate[nbranches] = rate[nbranches] * exp( root_K * (std_uniform() - 0.5) ) ;
	ratio = (newrate[nbranches] / rate[nbranches]) *
					probratio( time, newrate, logr, time, rate, logr ) ;
	/* fprintf(stderr,"DEBUG3: ratio = %f\n",ratio); */

/******************************************************************
*	Accept or reject the change in the usual MCMC way.
*/		
	randnum = std_uniform() ;
	if(randnum < ratio) {
		rate[nbranches] = newrate[nbranches] ;
		sucrate[nbranches]++ ;
	}	
	else {
		failrate[nbranches]++ ;
	}

	freedvec( newrate ) ;
}
/* _______________________________________________________________ */

void timechange( double *time, double *rate, double *logr )
/****************************************************************
*	This function proposes changes in the ages of the floating
* nodes. In the FREEZE_LEAF_AGES case only the internal nodes
* move, but under TETHER_LEAF_AGES, *all* nodes are free to move.
*
*	Like inflate(), the original form of this function changed 
* the ages of the nodes and made compensatory changes in rates 
* so that the branch lengths were preserved.  The calculations 
* associated with altering the branch lengths were thus 
* concentrated in a single stage of the Markov-Chain Monte-Carlo
* cycle, the function ratechange().
*
* In the current version of the code there is a switch,
* USE_LOGLIKE_ALWAYS, that changes this behaviour. If 
* USE_LOGLIKE_ALWAYS is defined in div_switches.h, then the
* log-likelihood gets consulted for time changes too, so no
* changes in branch length are required. 
*/
{
	int 	thisparent, thischild, ktrack, ktrack2 ;
	double	blen, randnum, ratio ; 
	double	*proprate, *proplogr, *proptime, *brlen ;
	double	minposs,maxtime, minchild, maxchild, numratio, denomratio, oldfrac ;
	double	newfrac, junk, oldwin, newwin ;
	double	rate_Jacobian_factors ; 
		
# if defined( USE_LOGLIKE_ALWAYS )
								
	double	*proplen ;			/* proposed branch lengths			*/
	int		pos_o_changes[3] ;	/* only needed if branch lengths	*/
								/* are allowed to change			*/
# endif

/**************************************************************
*	Allocate memory to record the current branch lengths and
* the proposed new rates and times.
*/
	proplogr = dvec(nbranches), assert( proplogr != NULL ) ;
	brlen = dvec(nbranches), assert( brlen != NULL ) ;
	proprate = dvec(nnodes), assert( proprate != NULL ) ;
	proptime = dvec(nnodes), assert( proptime != NULL ) ;
	
# if defined( USE_LOGLIKE_ALWAYS )								
	proplen = dvec(nnodes), assert( proplen != NULL ) ;
# endif

/****************************************************************
*	As a first proposal, take the current rates and times.
* Also, note the relevant branch lengths. Recall that there
* is no branch running backward from the root node, hence
* there is no associated branch length.
*/
	for( ktrack = 0 ; ktrack <= nbranches ; ktrack++ ) {
		proprate[ktrack] = rate[ktrack] ;
		proptime[ktrack] = time[ktrack] ;
		
		if( ktrack < nbranches ) {
			blen = time[tonode[ktrack][PARENT]] - time[ktrack] ;
			brlen[ktrack] = rate[ktrack] * blen ;
			proplogr[ktrack] = logr[ktrack] ;

# if defined( USE_LOGLIKE_ALWAYS )								
			proplen[ktrack] = brlen[ktrack] ;
# endif

		}
	}
	
/****************************************************************
*	For each of the floating nodes, propose an alteration to 
* the age. As above, I have preserved the original code at
* the expense of some duplication.
*/

# if defined(TETHER_LEAF_AGES)
	
	for( ktrack = 0 ; ktrack < nnodes ; ktrack++ )	/* Allow leaf nodes to move */
	
# else	/* FREEZE_LEAF_AGES */

	for( ktrack = ntips ; ktrack < nnodes ; ktrack++ ) 
	
#endif 
	/************************************************************
	*	Note the parent of the node in question. Any new age
	* for this node must be less than its parent's age.
	*/
	{
		thisparent = tonode[ktrack][PARENT] ;
		
# if defined(TETHER_LEAF_AGES) 
	/*************************************************************
	*	The smallest permissible age for a leaf node is set 
	* by its constraint, while for internal nodes the minimum 
	* age may be fixed either by a constraint or by the age of 
	* a child node.
	*/
		if( ktrack < ntips ) {	/* leaf node */
			minposs = constraint[ktrack] ;
		}
		else {	/* internal node */
			maxchild = max2(time[tonode[ktrack][1]], time[tonode[ktrack][2]]);
			minchild = min2(time[tonode[ktrack][1]], time[tonode[ktrack][2]]);
			minposs = max2(maxchild, constraint[ktrack]);
		}
		
# else	/* FREEZE_LEAF_AGES */

		maxchild = max2(time[tonode[ktrack][1]], time[tonode[ktrack][2]]);
		minchild = min2(time[tonode[ktrack][1]], time[tonode[ktrack][2]]);
		minposs = max2(maxchild, constraint[ktrack]);
# endif

	/*******************************************************************
	*	The following code refers to the ages of nodes other than the
	* root. They're allowed to shift around according to a fairly
	* elaborate scheme:
	*/	
		if( ktrack < (nnodes - 1) ) {
		/****************************************************************
		*	Determine the time from eldest child (or from a constrained 
		* minimum age) to this node's parent.  Then determine the 
		* fractional position of the node's current age within that 
		* allowed interval. Finally, note that current age splits 
		* the allowed interval into two pieces and record the size 
		* of the smaller of the two.
		*/
			maxtime = time[thisparent] - minposs;
			oldfrac = (time[ktrack]-minposs)/maxtime;
			oldwin = min2(oldfrac,1.0-oldfrac);
						
		/****************************************************************
		*	Toss a biased coin (probs time_K and (1.0 - time_K)) to decide
		* what to do next.  If the time_K side comes up, choose a new
		* fractional position uniformly anywhere within the whole 
		* interval. Otherwise, choose a new fractional position uniformly 
		* within an interval centered on the current position and 
		* extending to the nearest endpoint. The idea is to choose new 
		* times from a distribution that is simple to sample and
		* is supported over the whole range of allowed values, but weakly 
		* concentrated around the current value. 
		*/
			junk = std_uniform();
			if(junk < time_K)
			{
				newfrac = std_uniform();
			}
			else
			{
				newfrac = oldfrac + (2.0*std_uniform() - 1.0)*oldwin;
			}
			
		/***************************************************************
		*	Either way, shift the time to the new position and work
		* out numratio, the probability of jumping from the new 
		* position back to the old one.
		*/
			proptime[ktrack] = minposs + maxtime*newfrac;
			
			newwin = min2(newfrac, 1.0-newfrac);
			if( fabs(newfrac - oldfrac) < newwin ) {
			/************************************************************
			*	The old fractional position is such that it could be
			* recovered by either of the two mechanisms.
			*/
				numratio = time_K + (1.0 - time_K) / (2.0 * newwin) ;
			}
			else {
			/*************************************************************
			*	The original fractional position is so far from the new
			* one that only the time_K outcome of the coin toss could 
			* bring us back to it.
			*/
				numratio = time_K;
			}
			
		/*****************************************************************
		*	Work out the other factor that comes into the detailed
		* balance calculation, the probability of jumping from the
		* original fractional position to the new one.
		*/
			if( fabs(newfrac - oldfrac) < oldwin ) {
			/*****************************************************************
			*	The new and old positions are near enough either mechanism 
			* may have acted.
			*/
				denomratio = time_K + (1.0 - time_K) / (2.0 * oldwin) ;
			}
			else {
			/***************************************************************
			*	The two are so far apart that only the time_K outcome of
			* the coin toss could have produced this result.
			*/
				denomratio = time_K;
			}
		
# if !defined( USE_LOGLIKE_ALWAYS ) 
		/*****************************************************************
		*	Propose a rate change that preserves the length of the
		* branch that ends at the ktrack node. 
		*/
			assert( proptime[thisparent] >= proptime[ktrack] ) ;
			
			rate_Jacobian_factors = 1.0 ;
			blen = proptime[thisparent] - proptime[ktrack] ;
			if( blen > (TINY_FRAC * proptime[thisparent]) ) {
				proprate[ktrack] = brlen[ktrack] / blen ;
				proplogr[ktrack] = log( proprate[ktrack] ) ;
				
				rate_Jacobian_factors *= 
						(time[thisparent] - time[ktrack]) /
							(proptime[thisparent] - proptime[ktrack]) ;
			}
			
		/*****************************************************************
		*	The loop below runs over child nodes.  Node ktrack has
		* tonode[ktrack][N_KIDS] of them and their indices are stored
		* as
		*
		*	{ tonode[ktrack][1], tonode[ktrack][2] . . .
		*						tonode[ktrack][tonode[ktrack][N_KIDS]] }.
		*
		* The leaf nodes have no children (tonode[leaf][N_KIDS] == 0),
		* so this code does nothing when (ktrack < ntips).
		*/
			for(ktrack2=1; ktrack2 <= tonode[ktrack][N_KIDS]; ktrack2++)
			{
				thischild = tonode[ktrack][ktrack2] ;
				
				blen = proptime[ktrack] - proptime[thischild] ;
				assert( blen >= 0.0 ) ;
				
				if( blen > (TINY_FRAC * proptime[ktrack]) ) {
					proprate[thischild] = brlen[thischild] / blen ;
					proplogr[thischild] = log( proprate[thischild] ) ;
					
					rate_Jacobian_factors *= 
						(time[ktrack] - time[thischild]) /
							(proptime[ktrack] - proptime[thischild]) ;
				}	
			}
			
# else /* USE_LOGLIKE_ALWAYS is defined */
		/*****************************************************************
		*	Tabulate the proposed changes in branch lengths
		*/
			blen = proptime[thisparent] - proptime[ktrack] ;
			if( blen < (TINY_FRAC * proptime[ktrack]) ) {
				proplen[ktrack] = 0.0 ;
				proptime[ktrack] = proptime[thisparent] ;
			}
			else
				proplen[ktrack] = rate[ktrack] * blen ;
				
			assert( proplen[ktrack] >= 0.0 ) ;
			
			for( ktrack2=1 ; ktrack2 <= tonode[ktrack][N_KIDS] ; ktrack2++ ) {
				thischild = tonode[ktrack][ktrack2] ;
				
				blen = proptime[ktrack] - proptime[thischild] ;
				if( fabs(blen) < (TINY_FRAC * proptime[thischild]) ) {
					proplen[thischild] = 0.0 ;
					proptime[thischild] = proptime[ktrack] ;
				}
				
				proplen[thischild] = blen * rate[thischild] ;
				assert( proplen[thischild] >= 0.0 ) ;
			}
			
		/*****************************************************************
		*	No rates change, hence no Jacobian factors come into the
		* ratio of probabilities.
		*/
			rate_Jacobian_factors = 1.0 ;
			
# endif /* USE_LOGLIKE_ALWAYS */
			
		/*****************************************************************
		*	Finally, the first part of probabilistic acceptance step.  
		* Work out the ratio of probabilites: some ratios of times having
		* to do with Jacobian factors arising from changes to the rates
		* are hidden in rate_Jacobian_factors.
		*/
			ratio = probratio(proptime,proprate,proplogr,time,rate,logr) *
					(numratio / denomratio) * rate_Jacobian_factors ;
			/* fprintf(stderr,"DEBUG4: ratio = %f\n",ratio); */

		}
		else {
		/*****************************************************************
		*	The node under consideration is the root. Try to change its
		* age by a random multiplicative factor.
		*/
			proptime[ktrack] = minposs + (time[ktrack]-minposs)
				*exp(root_K*(std_uniform()-0.5));
				
# if !defined( USE_LOGLIKE_ALWAYS ) 
		/******************************************************************
		*	Adjust the rates on the root's child nodes: we assume it
		* has some.
		*/
			rate_Jacobian_factors = 1.0 ;
			for(ktrack2=1; ktrack2 <= tonode[ktrack][N_KIDS]; ktrack2++)
			{
				thischild = tonode[ktrack][ktrack2] ;
				
				blen = proptime[ktrack] - proptime[thischild] ;
				
				if( blen > (TINY_FRAC * proptime[ktrack]) ) {
					proprate[thischild] = brlen[thischild] / blen ;
					proplogr[thischild] = log( proprate[thischild] ) ;
					
					rate_Jacobian_factors *= 
						(time[ktrack] - time[thischild]) /
							(proptime[ktrack] - proptime[thischild]) ;
				}
			}
			
# else /* USE_LOGLIKE_ALWAYS *is* defined . . . */
		/*****************************************************************
		*	The lengths of the branches issuing from the root will change.
		*/
			for( ktrack2=1 ; ktrack2 <= tonode[ktrack][N_KIDS] ; ktrack2++ ) {
				thischild = tonode[ktrack][ktrack2] ;
				
				blen = proptime[ktrack] - proptime[thischild] ;
				if( fabs(blen) < (TINY_FRAC * proptime[thischild]) )
					proplen[thischild] = 0.0 ;
				else
					proplen[thischild] = blen * rate[thischild] ;
					
				assert( proplen[thischild] >= 0.0 ) ;
			}

		/*****************************************************************
		*	No rates change, hence no Jacobian factors come into the
		* ratio of probabilities.
		*/
			rate_Jacobian_factors = 1.0 ;
			
# endif /* USE_LOGLIKE_ALWAYS */
			
		/*****************************************************************
		*	Again, work out the ratio of probabilites.
		*/		
			ratio = rate_Jacobian_factors *
				(proptime[ktrack] - minposs)/(time[ktrack] - minposs) ;
				
			ratio *= probratio(proptime,proprate,proplogr,time,rate,logr) ;
			/* fprintf(stderr,"DEBUG5: ratio = %f\n",ratio); */
		}
		
# if defined( USE_LOGLIKE_ALWAYS )
	/**********************************************************************
	*	Include the probability ratio arising from the likelihood of
	* a tree with the given branch lengths.
	*/
		if( ktrack < ntips ) {
		/******************************************************************
		*	We're moving a leaf node, so only one branch length changes.
		*/
			if (brlen[ktrack] != proplen[ktrack]) {
				double debug;
				ratio *= exp( debug = delta_tree_loglike( brlen, proplen[ktrack], ktrack ) ) ;
				/* fprintf(stderr,"DEBUG: %d %f -> %f (%f) -> (%f)\n", 
				   ktrack,brlen[ktrack], proplen[ktrack], debug, ratio); */
		  }
		}
		else if( ktrack < (nnodes - 1) ) {
			double debug;
		/******************************************************************
		*	We're moving an internal node, so three lengths change.
		*/
			pos_o_changes[0] = ktrack ;
			pos_o_changes[1] = tonode[ktrack][1] ;
			pos_o_changes[2] = tonode[ktrack][2] ;
			
			ratio *= exp( debug = vect_delta_tree_loglike(brlen, proplen, pos_o_changes, 3) ) ;
			/* fprintf(stderr,"DEBUG: %d %f -> %f, %d %f -> %f (%f) -> (%f)\n", 
				pos_o_changes[1],brlen[pos_o_changes[1]], proplen[pos_o_changes[1]], 
				pos_o_changes[2],brlen[pos_o_changes[2]], proplen[pos_o_changes[2]], 
				debug, ratio); */
		}
		else {
		/******************************************************************
		*	We're moving the root, so only two lengths change.
		*/
			pos_o_changes[0] = tonode[ktrack][1] ;
			pos_o_changes[1] = tonode[ktrack][2] ;
			
			ratio *= exp( vect_delta_tree_loglike(brlen, proplen, pos_o_changes, 2) ) ;
		}			
# endif		/* USE_LOGLIKE_ALWAYS */
		
	/**********************************************************************
	*	Generate a number between zero and one and accept the change
	* accordingly.
	*/
		/* fprintf(stderr,"DEBUG6: ratio = %f\n",ratio); */
		randnum = std_uniform();
		if(randnum < ratio) {
		/******************************************************************
		*	Accept the change: copy data from the arrays of proposals
		* into the arrays that reflect the current state.
		*/
			suctime[ktrack]++;

		/*************************************************************
		*	Make changes for the ktrack node itself: if the node
		* is other than the root, we change the rate for the branch
		* that reaches back to ktrack's parent so that the branch 
		* length remains the same.
		*/
			time[ktrack] = proptime[ktrack] ;

			if(ktrack < nnodes-1) {
				rate[ktrack] = proprate[ktrack] ;
				logr[ktrack] = proplogr[ktrack] ;

# if defined( USE_LOGLIKE_ALWAYS )								
				brlen[ktrack] = proplen[ktrack] ;
# endif	           
			}	
		
		/**************************************************************
		*	Do rate adjustments on the branches that reach from 
		* ktrack to its child nodes. Once again, we need take no 
		* special care about the leaf nodes, as they have 
		* tonode[leaf][N_KIDS] == 0.
		*/
			for(ktrack2 = 1; ktrack2 <= tonode[ktrack][N_KIDS]; ktrack2++)
			{
				thischild = tonode[ktrack][ktrack2];
				rate[thischild] = proprate[thischild];
				logr[thischild] = proplogr[thischild];

# if defined( USE_LOGLIKE_ALWAYS )								
				brlen[thischild] = proplen[thischild] ;
# endif	           
			}
		}
		else {
		/******************************************************************
		*	Reject the change: copy data from the current arrays back
		* into the array that holds proposals.
		*/
			failtime[ktrack]++ ;

		/*************************************************************
		*	Restore data for the ktrack node itself: if the node
		* is other than the root, we restore the rate for the branch
		* that reaches back to ktrack's parent too.
		*/

			proptime[ktrack] = time[ktrack] ;
			
			if(ktrack < nnodes-1) {
				proprate[ktrack] = rate[ktrack] ;
				proplogr[ktrack] = logr[ktrack] ;

# if defined( USE_LOGLIKE_ALWAYS )								
				proplen[ktrack] = brlen[ktrack] ;
# endif	           
			}

		/**************************************************************
		*	Restore data for the branches that reach from 
		* ktrack to its children (if it has any).
		*/
			for(ktrack2 = 1; ktrack2 <= tonode[ktrack][N_KIDS]; ktrack2++)
			{
	            thischild = tonode[ktrack][ktrack2] ;
	            proprate[thischild] = rate[thischild] ;
	            proplogr[thischild] = logr[thischild] ;

# if defined( USE_LOGLIKE_ALWAYS )								
				proplen[thischild] = brlen[thischild] ;
# endif	           
			}
							
		} /* end of accept/reject if{}...else{} construct */
		
	} /* end of loop that proposes changes */
	
	freedvec(proprate);
	freedvec(proplogr);
	freedvec(proptime);
	freedvec(brlen);
	
# if defined( USE_LOGLIKE_ALWAYS )								
	freedvec(proplen) ;
# endif
}
/* ____________________________________________________________________ */

void ratechange( double *time, double *rate, double *logr )
/********************************************************************
*	Propose changes in the rates on the branches.  This step changes
* the lengths of the branches, so one needs to make reference to the 
* multivariate normal approximation to the likelihood of the tree.
*/
{
	int		ktrack ;
	double	holdlen, newlen, randnum, ratio, delta_like ;
	double	*propbrlen, *proprate, *proplogr, *brlen ;

/*********************************************************************
*	Allocate temporary storage for the branch lengths, the rates
* and their logs. If assertion-checking is on (NDEBUG not defined)
* then we'll check against a NULL return.
*/
	brlen = dvec(nbranches), assert( brlen != NULL ) ;
	propbrlen = dvec(nbranches), assert( propbrlen != NULL ) ;
	proprate = dvec(nnodes), assert( proprate != NULL ) ;
	proplogr = dvec(nnodes), assert( proplogr != NULL ) ;
	
/**********************************************************************
*	Initialize the array of proposed rates to be the same as
* the current ones. At the same time, make two identical arrays
* holding the branch lengths.
*/
	for( ktrack = 0 ; ktrack <= nbranches ; ktrack++) {
	
		proprate[ktrack] = rate[ktrack] ;
		
		if( ktrack < nbranches ) {
			proplogr[ktrack] = logr[ktrack] ;
			holdlen =  time[tonode[ktrack][PARENT]] - time[ktrack] ;
			assert( holdlen >= 0.0 ) ;
			
			brlen[ktrack] = rate[ktrack] * holdlen ;
			propbrlen[ktrack] = brlen[ktrack] ;
		}
	}

/***********************************************************************
*	Now work through the list of branches, proposing to change
* each rate in turn.
*/
	for( ktrack = 0 ; ktrack < nbranches ; ktrack++ ) {
	/********************************************************************
	*	If the corresponding branch has non-zero length, propose the 
	* usual sort of random-step-in-log change for the rate and
	* calculate the consequent change in the branch length.
	*/
		if( brlen[ktrack] > 0.0 ) {
			proprate[ktrack] = exp( root_K * (std_uniform() - 0.5) ) * rate[ktrack] ;
			proplogr[ktrack] = log(proprate[ktrack]) ;
			
			newlen = proprate[ktrack] * 
						(time[tonode[ktrack][PARENT]] - time[ktrack]) ;
						
			propbrlen[ktrack] = newlen ;
		}
		else {
			newlen = 0.0 ;
		}
		
	/**********************************************************************
	*	Find the consequent change in log-likelihood of the tree.
	*/	
		delta_like = delta_tree_loglike( brlen, newlen, ktrack ) ;
			
	/***********************************************************************
	*	Finish the calculation of the ratio of probabilities; this
	* includes factors from the proposal density and from the priors.
	*/ 	
		ratio = exp( delta_like ) * (proprate[ktrack] / rate[ktrack] ) 
				* probratio( time, proprate, proplogr, time, rate, logr ) ;

		/* fprintf(stderr,"DEBUG7 %d: %f -> %f (%f) --> %f\n",
		   ktrack, brlen[ktrack], newlen, delta_like, ratio); */
	/************************************************************************
	*	Accept or reject the change in the usual MCMC way.
	*/
		randnum = std_uniform() ;

		if(randnum < ratio) {
		/****************************
		*	Accept the new rate.
		*/
			sucrate[ktrack]++;

			rate[ktrack] = proprate[ktrack] ;
			logr[ktrack] = proplogr[ktrack] ;
			brlen[ktrack] = propbrlen[ktrack] ;
		}
		else {
		/*****************************
		*	Reject the change.
		*/
			failrate[ktrack]++;

			proprate[ktrack] = rate[ktrack] ;
			proplogr[ktrack] = logr[ktrack] ;
			propbrlen[ktrack] = brlen[ktrack] ;
		}
	}
	
/***********************************************************************
*	Release the temporary storage and return.
*/
	freedvec(brlen) ;
	freedvec(propbrlen) ;
	freedvec(proprate) ;
	freedvec(proplogr) ;
}
/* ____________________________________________________________________ */

void summarize( double *first_sample, double **samples, 
				char *treefilename, char *nodefilename )
/**************************************************************************
*	Write loads of output to variously named files.
*/
{
	int ktrack, ktrack2;
	double *sumsq,*sum, *sampmean, *sampmeansq, *sumtime, *sumsqtime;
	double holdrt, holdtemp;
	double *oldbrlen, *newbrlen ;
	double *ntime, *nrate, totlen, totrate, holdparent, holdbrlen;
	double invfac, flatefac, holdtime;
	double *oldtime,*newtime,*oldrate,*newrate,theratio;
	double *actualtime, *oldlogr, *newlogr;
	FILE *ratiofile, *treefile, *nodefile;
	
	/* Used in computing confidence intervals */
	int		to_drop, conf_min_pos, conf_max_pos ;
	double	*conf_list, *conf_min, *conf_max ;
	/******************************************/

/* BEGIN DEBUG */
	printf("\nInflate: nsuc=%10ld nfail=%10ld psuc= %7.5f\n\n",
		sucflate, failflate, (1.0*sucflate)/ 
		(sucflate+failflate) ); 
	for(ktrack=0; ktrack <= nbranches; ktrack++)
	{
		if(ktrack != nbranches)
		{
			printf("RATE %3d: nsuc=%10ld nfail=%10ld psuc= %7.5f\n", 
				ktrack, sucrate[ktrack], failrate[ktrack], (1.0*sucrate[ktrack])/
				(sucrate[ktrack]+failrate[ktrack]) );
		}
		else
		{
			printf("\nBrownian: nsuc=%10ld nfail=%10ld psuc= %7.5f\n\n",
				sucrate[ktrack], failrate[ktrack], (1.0*sucrate[ktrack])/ 
				(sucrate[ktrack]+failrate[ktrack]) ); 
		} 
	}
	
# if defined( FREEZE_LEAF_AGES ) 
	for(ktrack=ntips; ktrack <= nbranches; ktrack++) 
# else
	for( ktrack=0 ; ktrack < nnodes ; ktrack++ ) /* TETHER_LEAF_AGES */
# endif
	{ 
		printf("TIME %3d: nsuc=%10ld nfail=%10ld psuc= %7.5f\n",
			ktrack, suctime[ktrack], failtime[ktrack], (1.0*suctime[ktrack])/ 
			(suctime[ktrack]+failtime[ktrack]) ); 
	}
	
	printf( "\n" ) ;
/* END DEBUG */

	oldlogr = dvec(nnodes);
	newlogr = dvec(nnodes);
	oldtime = dvec(nnodes);
	newtime = dvec(nnodes);
	oldrate = dvec(nnodes);
	newrate = dvec(nnodes);
	oldbrlen = dvec(nnodes);
	newbrlen = dvec(nnodes);

	nodefile = fopen(nodefilename, "a");
	if( nodefile == NULL ) {
		fprintf( stderr, "Can't open the node file %s.\n", nodefilename ) ;
	}
	
	ratiofile = fopen(RATIO_FILE,"w");
	if( ratiofile == NULL ) {
		fprintf( stderr, "Can't open the ratio file %s.\n", RATIO_FILE ) ;
	}
	else {
	/******************************************************
	*	Write probability ratios to the ratio file.
	*/
		for(ktrack=0; ktrack< nnodes; ktrack++)
		{
			oldtime[ktrack] = first_sample[ktrack];
			oldrate[ktrack] = first_sample[ktrack+nnodes];
			oldlogr[ktrack] = log(oldrate[ktrack]);
			
			holdtemp = first_sample[tonode[ktrack][PARENT]] - oldtime[ktrack] ;
			oldbrlen[ktrack] = holdtemp * oldrate[ktrack] ;
		}
		for(ktrack2=0; ktrack2 < num_o_samples; ktrack2++)
		{
			for(ktrack=0; ktrack< nnodes; ktrack++) 
			{ 
				newtime[ktrack] = samples[ktrack2][ktrack]; 
				newrate[ktrack] = samples[ktrack2][ktrack+nnodes]; 
				newlogr[ktrack] = log(newrate[ktrack]);
				
				holdtemp = samples[ktrack2][tonode[ktrack][PARENT]] - newtime[ktrack] ;
				newbrlen[ktrack] = holdtemp * newrate[ktrack] ;
			}
			theratio = probratio(newtime,newrate, newlogr,oldtime,oldrate,oldlogr);
			/* fprintf(stderr,"DEBUG8: theratio = %f\n",theratio); */
			fprintf(ratiofile, "%22.14e (%.6f + %.6f -> %22.14e)\n", 
				theratio, log(theratio), (tree_loglike(newbrlen) - tree_loglike(oldbrlen)),
				theratio * exp(tree_loglike(newbrlen) - tree_loglike(oldbrlen))  );
		} 
		
		fclose(ratiofile);
	}
	
	freedvec(oldtime);
	freedvec(newtime);
	freedvec(oldrate);
	freedvec(newrate);
	freedvec(oldlogr);
	freedvec(newlogr);
	
	/* More confidence interval stuff */
	conf_min = dvec( 2*nnodes + 1 ) ;
	conf_max = dvec( 2*nnodes + 1 ) ;
	conf_list = dvec( num_o_samples ) ;

	to_drop = floor( 0.5 * (1.0 - confidence) * num_o_samples ) ; 
	conf_min_pos = to_drop ;
	conf_max_pos =  num_o_samples - 1 - to_drop ;
	/**************************/
	
	ntime = dvec(nnodes);
	nrate = dvec(nnodes);
	sumtime = dvec(nnodes);
	sumsqtime = dvec(nnodes);
	sumsq = dvec(nnodes+nbranches);
	sum = dvec(nnodes+nbranches);
	sampmean = dvec(nnodes-2);
	sampmeansq = dvec(nnodes-2);
	for(ktrack=0; ktrack <= nnodes+nbranches; ktrack++)
	{
		sum[ktrack] = 0.0;
		sumsq[ktrack] = 0.0;
		if(ktrack < nnodes)
		{
			sumtime[ktrack] = 0.0;
			sumsqtime[ktrack] = 0.0;
		}

		if(ktrack < nnodes-1)
		{
			sampmean[ktrack] = 0.0;
			sampmeansq[ktrack] = 0.0;
		}
		for(ktrack2=0; ktrack2 < num_o_samples; ktrack2++)
		{
			if(ktrack < nnodes)
			{
				holdtime = samples[ktrack2][ktrack];
				sumtime[ktrack] += holdtime;
				sumsqtime[ktrack] += holdtime*holdtime;
				
				conf_list[ktrack2] = holdtime ; /* yet more confidence interval code */
			}

			if(ktrack < nnodes-1)
			{
				holdparent = samples[ktrack2][tonode[ktrack][PARENT]];
				holdrt = samples[ktrack2][nnodes+ktrack];
				holdbrlen = holdrt*(holdparent-holdtime);
				sampmean[ktrack] += holdbrlen;
				sampmeansq[ktrack] += holdbrlen*holdbrlen; 
				holdrt = samples[ktrack2][nnodes-1];
				holdtemp =  holdtime/holdrt;
				sum[ktrack] += holdtemp;
				sumsq[ktrack] += holdtemp*holdtemp;
			}
			else
			{
				holdrt = samples[ktrack2][ktrack];
				sum[ktrack] += holdrt;
				sumsq[ktrack] += holdrt*holdrt; 
				
				conf_list[ktrack2] = holdrt ; /* confidence interval code too */
			}		
		}
		
	/****************************************************
	*	 Sort the list of values so one can extract
	* confidence limits from it.
	*/
		qsort( conf_list, ((size_t) num_o_samples), sizeof(double), dbl_cmpr ) ;
		conf_min[ktrack] = conf_list[conf_min_pos] ;
		conf_max[ktrack] = conf_list[conf_max_pos] ;
	/*****************************************************/
		
		if(ktrack < nnodes-1)
		{
			if( nodefile != NULL )
				fprintf(nodefile,"p %3d", ktrack);
			/* printf("Prop. Time of Node %3d to Root = ", ktrack); */
		}
		else
		{
			if(ktrack==nnodes-1)
			{
				fprintf(nodefile, "p %3d", ktrack);
				/* printf("Root Depth = "); */
			}
			else
			{
				if( nodefile != NULL )
					fprintf(nodefile,"R %3d", ktrack-nnodes);
					
				if(ktrack < nnodes+nbranches)
				{
					printf("Rate of Branch %3d = ", ktrack-nnodes);
				}
				else
				{
					printf("\nBrownian Motion Constant = ");
				}
			}
		}
/* DEBUG */
		if(sumsq[ktrack] - sum[ktrack]*sum[ktrack]/num_o_samples < 0.0)
		{
			printf("prob in summary, can't take sqr. root of %18.12f\n",
				(sumsq[ktrack] - (sum[ktrack]*sum[ktrack]/num_o_samples))/(num_o_samples-1.0));
		}
		
		if( nodefile != NULL ) {
			fprintf(nodefile, " %9.5f %9.5f\n", sum[ktrack]/num_o_samples
				, sqrt((sumsq[ktrack] - sum[ktrack]*sum[ktrack]/num_o_samples
				)/(num_o_samples-1.0)));
		}
		
		if( ktrack >= nnodes ) {	/* commented out proportional quantities */
			printf(" %7.5f (Std. Dev. = %7.5f [%7.5f, %7.5f ] )\n", 
				sum[ktrack]/num_o_samples, 
				sqrt((sumsq[ktrack] - sum[ktrack]*sum[ktrack]/num_o_samples)/(num_o_samples-1.0)), 
				conf_min[ktrack], conf_max[ktrack] ) ;
		}
	}
	printf("\n");

# if 0
	for(ktrack=0 ; ktrack < nnodes-1; ktrack++)
	{ 
		printf("branch %3d samp mean= %8.5f (%8.5f) norm. mean = %8.5f\n", 
			ktrack, sampmean[ktrack]/num_o_samples, sqrt((sampmeansq[ktrack] -
			sampmean[ktrack]*sampmean[ktrack]/num_o_samples)/(num_o_samples-1.0)), 
			mean[ktrack]);
	}
	
# endif

	/* NOW NORMALIZE */
	for(ktrack = 0; ktrack <= nnodes+nbranches; ktrack++)
	{
		sum[ktrack] = 0.0;
		sumsq[ktrack] = 0.0;
	}
	for(ktrack = 0; ktrack < num_o_samples; ktrack++)
	{
		totrate = 0.0;
		totlen = 0.0;
		for(ktrack2=0; ktrack2 < nnodes - 1; ktrack2++)
		{
			ntime[ktrack2] = samples[ktrack][ktrack2];
			holdparent = samples[ktrack][tonode[ktrack2][PARENT]];
			nrate[ktrack2] = samples[ktrack][nnodes+ktrack2];
			totlen += holdparent - ntime[ktrack2];
			totrate += nrate[ktrack2]*(holdparent-ntime[ktrack2]);
		}
		ntime[nnodes-1] = samples[ktrack][nnodes-1];
		nrate[nnodes-1] = samples[ktrack][nnodes+nbranches];
		invfac = totrate/totlen;
		flatefac = totlen/totrate;
		for(ktrack2=0; ktrack2 <= nnodes-1; ktrack2++)
		{
			ntime[ktrack2] *= invfac;
			nrate[ktrack2] *= flatefac;
			sum[ktrack2] += ntime[ktrack2];
			sumsq[ktrack2] += ntime[ktrack2] * ntime[ktrack2];
			sum[ktrack2+nnodes] += nrate[ktrack2];
			sumsq[ktrack2+nnodes] += nrate[ktrack2]*nrate[ktrack2];
		}
	}
	for(ktrack = 0; ktrack <= nnodes+ nbranches; ktrack++)
	{
		if(ktrack <= nnodes-1)
		{
			/* printf("Norm. Time of Node %3d= ", ktrack); */
			
			if( nodefile != NULL )
				fprintf(nodefile,"t %3d", ktrack);
		}
		else
		{
			if( nodefile != NULL )
				fprintf(nodefile,"r %3d", ktrack-nnodes);
			
		/******************************************	
			if(ktrack < nnodes+nbranches)
			{
				printf("Norm. Rate of Branch %3d = ", ktrack-nnodes);
			}
			else
			{
				printf("Norm. Brownian Constant = ");
			}
		*******************************************/

			
		}
/* DEBUG */

	/********************************************
		if(sumsq[ktrack] - sum[ktrack]*sum[ktrack]/num_o_samples < 0.0)
		{
			printf("prob in summary, can't take sqr. root of %18.12f\n",
				(sumsq[ktrack] - (sum[ktrack]*sum[ktrack]/num_o_samples))/(num_o_samples-1.0));
		}
	********************************************/
	
		if( nodefile != NULL ) {
			fprintf(nodefile, " %9.5f %9.5f\n", sum[ktrack]/num_o_samples
				, sqrt((sumsq[ktrack] - sum[ktrack]*sum[ktrack]/num_o_samples
				)/(num_o_samples-1.0)));
		}
		
	/***************************************************
		printf(" %9.5f (Std. Dev. = %9.5f)\n", sum[ktrack]/num_o_samples
			, sqrt((sumsq[ktrack] - sum[ktrack]*sum[ktrack]/num_o_samples
			)/(num_o_samples-1.0)));
	*****************************************************/
	}

	for(ktrack = 0; ktrack < nnodes; ktrack++)
	{
		if( nodefile != NULL ) {
			fprintf(nodefile,"T %3d %9.5f %9.5f\n", ktrack, 
				sumtime[ktrack]/num_o_samples,  sqrt((sumsqtime[ktrack] - 
				sumtime[ktrack]* sumtime[ktrack]/num_o_samples)/(num_o_samples-1.0)));
		}
		
/*************************************************************************
*	Report the dates of the internal nodes - this only makes sense if
* max_tip_date has been defined sensibly.
*/
		{
			double	deev_sq ;
			
			printf("Actual time node %3d = %7.2f (Std. Dev. = ", 
				ktrack, max_tip_date - sumtime[ktrack]/num_o_samples) ;
				
			deev_sq = sumsqtime[ktrack] - sumtime[ktrack]*sumtime[ktrack]/num_o_samples ;
			if( deev_sq < 0.0 )
				printf("UNDEFINED; ") ;
			else
				printf("%6.2f; ", sqrt( deev_sq / (num_o_samples-1.0) ) ) ;
				
			printf( "[ %7.2f, %7.2f ] )\n", (max_tip_date - conf_max[ktrack]), 
											(max_tip_date - conf_min[ktrack]) ) ;
		}
	}
	
	actualtime = dvec(nnodes);
	for(ktrack = 0; ktrack < nnodes-1; ktrack++)
	{
		actualtime[ktrack] = (sumtime[tonode[ktrack][PARENT]] - sumtime[ktrack])
			/num_o_samples;
	}
	treefile = fopen(treefilename, "a");
	if( treefile != NULL ) {
		fprint_tree( treefile, actualtime ) ;
		fclose(treefile);
	}
	
	if( nodefile != NULL )
		fclose(nodefile);

	freedvec(actualtime);
	freedvec(sumtime);
	freedvec(sumsqtime);
	freedvec(nrate);
	freedvec(ntime);
	freedvec(sampmean);
	freedvec(sampmeansq);
	freedvec(sumsq);
	freedvec(sum);
	
	freedvec( conf_min ) ;
	freedvec( conf_max ) ;
	freedvec( conf_list ) ;
}
/* ________________________________________________________________ */

double probratio(	double *proptime, double *proprate, double *proplogr,
					double *oldtime,  double *oldrate, 	double *oldlogr  )
/************************************************************************
*  This routine calculates those contributions to the ratio
*
*	P(proposed state) / P(current state) 
*
* arising from the priors. The ratio of proposal densitites should
* be handled locally by routines that call this one, as should 
* any factors arising from the multivariate Gaussian over the
* branch lengths.
*/
{	
	double	old_nu, prop_nu ;
	double	oldvar, newvar ;
	double	theratio, oldlength, newlength, oldrdif, newrdif ;
	double	rtratio, regratio, expratio ;
	int		panode, grpanode, startnode, sibnode, ktrack ;
		
# if !defined( IGNORE_PROP_TIME ) /* The original code */
/******************************************************************
*	The code below is executed just once, at the beginning of
* each run.  It sets the depth of the nodes (number of branches
* back from the leaves) and fills the arrays a[] and b[].
*/
	static int	depth[MAX_NODES] ;	/* max number of steps to leaves */
	static int	hereyet = 0 ;		/* Flag for initialization */
	
	/* params of the node-depth prior */
	static double	a[MAX_NODES] ;
	static double	b[MAX_NODES] ;

	if( hereyet == 0 ) {
	/***************************************************************
	*	By definition, the tips are at depth zero.
	*/
		for(ktrack=0; ktrack < ntips; ktrack++)
		{
			depth[ktrack] = 0;
		}
		
	/****************************************************************
	*	The internal nodes have depths one greater than the
	* deepest of their children.
	*/
		for( ktrack=ntips; ktrack < nnodes; ktrack++ )
		{
			depth[ktrack] = 1+ max2(depth[tonode[ktrack][1]], 
				depth[tonode[ktrack][2]]);
		}
		
	/****************************************************************
	*	Set a[] and b[]: they are parameters of the priors
	* on proportional node depth.
	*/
		for(ktrack = ntips; ktrack < nnodes-1; ktrack++)
		{
			a[ktrack] = node_depth_ab * depth[ktrack] ;
			b[ktrack] = node_depth_ab ;
			a[ktrack] -= 1.0;
			b[ktrack] -= 1.0;
		}
		
		hereyet = 1;	/* Been here, done this. */
	}
	
# endif /* IGNORE_PROP_TIME or not */ 
	
/***************************************************************
*	This section looks at that contribution to probratio which
* arises from the log-normal random walk on the rates.
* Recall the model: the rate on a descendant branch is drawn
* from a log-normal distribution whose variance is proportional
* to the difference in ages of the midpoints of the branches.
* The mean of the log-normal is chosen so the descendants
* have the same expected rate as their parents.
*
* Aside, about log-normal distribs:
*
*	If x = \exp( y ) and y is distributed normally with
* mean \mu_y and variance \sigma_y^2, then
*
*	\mu_x = \exp( \mu_y + (1/2) * \sigma_y^2 )
*
* Thus if, as here, we want a family of distribs with the
* same \mu_x but various values of \sigma_y, we should choose
*
*	\mu_y = log( \mu_x ) - 0.5 * \sigma_y^2
*
* We begin by making a note of the old and new values of
* the brownian motion constant, called \nu in Thorne et al.
* and stored at the end of the array of rates.
*/	
	old_nu = oldrate[nbranches], prop_nu = proprate[nbranches] ;
	
	expratio = 0.0 ;	/* log of a ratio, starts at 0.0 = log(1.0) */
	rtratio = 1.0 ;
	regratio = 1.0 ;
	for( ktrack = 0 ; ktrack < nbranches ; ktrack++ ) {
	/************************************************************
	*	Find the parent of the ktrack-th node and, unless
	* that's the root, find the grandparent too.
	*/
		panode = tonode[ktrack][PARENT] ;
		if(panode != nnodes-1) {
		/********************************************************
		*	The ktrack-th branch is one governed by the 
		* conditional log-normal prior on the rates: work out
		* its contribution to the ratio.   Note that the
		* time of interest is
		*
		*	(t_g + t_p) / 2 - (t_c + t_p) / 2 
		*								= (t_g - t_c) / 2
		*
		* where t_c is age of the child node, t_p that of
		* its parent and t_g that of its grandparent. Then
		*
		*	\sigma_logr = \nu * (t_g - t_c) / 2
		* 
		*/
			grpanode = tonode[panode][PARENT] ;
			
			oldlength = 0.5 * (oldtime[grpanode] - oldtime[ktrack]) ;
			oldvar = old_nu * oldlength ;
			oldrdif =  oldlogr[ktrack] - oldlogr[panode] + (0.5 * oldvar) ;
			assert( oldlength >= 0.0 ) ;
			
			newlength = 0.5 * (proptime[grpanode] - proptime[ktrack]) ;
			newvar =  prop_nu * newlength ;
			newrdif =  proplogr[ktrack] - proplogr[panode] + (0.5 * newvar) ;
			assert( newlength >= 0.0 ) ;
			
			if( newvar > 0.0 ) {
			/******************************************************
			*	The first ratio here, rtratio, comes from the 
			* normalizing constants of the gaussians in log rate; 
			* ultimately we will take its square root. 
			* 
			* 	The second ratio, expratio, will eventually be multiplied
			* by -0.5 and exponentiated: it's the arg of the 
			* eponential part of the gaussian.
			*
			*	The last ratio, regratio, is, um, a regular ratio:
			* it arises from the change-of-variables formula that
			* converts the gaussian density on log( rate ) to a
			* density over the rates themselves.
			*/
				rtratio *= oldvar / newvar ; 
				expratio += ((newrdif*newrdif) / newvar)
								- ((oldrdif*oldrdif) / oldvar) ;
				regratio *= oldrate[ktrack] / proprate[ktrack] ;
			}
			else {
			/******************************************************
			*	Either prop_nu is zero, which should never happen,
			* or the proposal includes a pair of consecutive, 
			* zero-length branches. In either case, we should
			* reject the proposal.
			*/
				return( 0.0 ) ;
			}
		}
	}
	
/*******************************************************************
*	The branches issuing from the root are special. The branch
* reaching to startnode has a gamma-distributed prior on its rate
* while its sib has a log-normally distributed rate.
*/
	startnode = tonode[nnodes-1][1] ;
	sibnode = tonode[nnodes-1][2] ;

	oldlength = oldtime[nnodes-1] - 
					0.5	* (oldtime[startnode] + oldtime[sibnode]) ;			
	oldvar = old_nu * oldlength ;
	oldrdif = oldlogr[sibnode] - oldlogr[startnode] + (0.5 * oldvar) ;
	
	newlength = proptime[nnodes-1] - 
					0.5 * (proptime[startnode] + proptime[sibnode]) ;
	newvar = prop_nu * newlength ;
	
	if(newvar <= 0.0) {
	/******************************************************
	*	Either prop_nu is zero, which should never happen,
	* or the proposal gives zero length to both branches 
	* leading out of the root. We reject both cases.
	*/
		return( 0.0 ) ;
	}
	else {
	/***************************************************************
	*	This first group of contributions have to do with the
	* log-normal prior on the rate for sibnode.
	*/
		newrdif = proplogr[sibnode] - proplogr[startnode] + (0.5 * newvar) ;
		
		rtratio *= (oldvar / newvar) ;
		
		expratio += ((newrdif*newrdif) / newvar) - ((oldrdif*oldrdif) / oldvar) ; 
		regratio *= (oldrate[sibnode] / proprate[sibnode]) ;
		
	/****************************************************************
	*	Expratio has been accumulating contributions from the
	* log-normal distribs of the general form:
	*
	*		(val - mean)^2 / var
	*
	* An overall factor of -0.5 is missing, so it gets applied here.
	*/
		expratio *= (-0.5) ;
		
	/*****************************************************************
	*	This next term has to do with the gamma-distributed priors 
	* on the age of the root node, on the brownian motion constant 
	* \nu and on the rate on the branch that reaches from the root 
	* to startnode.
	*/
		expratio += /* prior on age of root node */
			(timegamr - 1.0) * log(proptime[nnodes-1] / oldtime[nnodes-1]) +
			timegamlam * (oldtime[nnodes-1] - proptime[nnodes-1]) +
				/* prior on the rate for root-to-startnode */
			(rategamr - 1.0) * log(proprate[startnode] / oldrate[startnode]) +
			rategamlam * (oldrate[startnode] - proprate[startnode]) +
				/* prior on \nu, the brownian motion constant */
			(browngamr - 1.0)*log(prop_nu / old_nu) +
			browngamlam * (old_nu - prop_nu) ;
		
# ifdef SLICE_VOL_PRIOR
	/*******************************************************************
	*	Use a prior on the ages of internal nodes such that the
	* conditional density
	*
	*		P( age of node | age of parent and ages of chidren )
	*
	* is flat.
	*/
		regratio *= age_prior_ratio( oldtime, proptime ) ;

# elif !defined( IGNORE_PROP_TIME ) 
	/* The original code, which uses a prior on proportional node depth */
	/*******************************************************************
	*	Finally, there are some contributions due to the priors
	* on proportional node depth. These apply only to internal nodes
	* other than the root.
	*/
		{
			double	oldprop, newprop ;
			
			for(ktrack = ntips; ktrack < nnodes-1; ktrack++) {
				panode = tonode[ktrack][PARENT];
				newprop = proptime[ktrack]/proptime[panode];
				oldprop = oldtime[ktrack]/oldtime[panode];
				expratio += a[ktrack]*log(newprop/oldprop) + b[ktrack]*
					log((1.0-newprop)/(1.0-oldprop));
				regratio *= oldtime[panode]/proptime[panode];
			}
		}
		
# endif /* SLICE_VOL_PRIOR and/or IGNORE_PROP_TIME */

	}
	
# if defined( TETHER_LEAF_AGES )
/**************************************************************
*	There are priors on the ages of the tips: compute
* their contribution too.
*/
{
	int		j ;
	double	tether_prior, tethering_ratio ;
	
	tethering_ratio = 1.0 ;
	for( j=0 ; j < ntips ; j++ ) {

	    double debug1, debug2;

		/* carry on straight away if the tip hasn't moved */
		if (oldtime[j] == proptime[j]) continue;

	/***********************************************************
	*	First the factors arising from the old tip ages . . .
	*/
		assert( oldtime[j] >= tethers[j].win_min ) ;
		if( oldtime[j] < tethers[j].win_max ) {
			tether_prior = 1.0 - 
				exp( -tethers[j].lambda * (oldtime[j] - tethers[j].win_min) ) ;
		}
		else {
			tether_prior = tethers[j].max_val
							* exp( -tethers[j].lambda * (oldtime[j] - tethers[j].win_max) ) ;
		}
		
		tethering_ratio /= (debug1 = tether_prior) ;
		
	/*************************************************************
	*	. . . then those from the proposed new ones.
	*/
		assert( proptime[j] >= tethers[j].win_min ) ;
		if( proptime[j] < tethers[j].win_max ) {
			tether_prior = 1.0 - exp( -tethers[j].lambda * (proptime[j] - tethers[j].win_min) ) ;
		}
		else {
			tether_prior = tethers[j].max_val
							* exp( -tethers[j].lambda * (proptime[j] - tethers[j].win_max) ) ;
		}
		
		tethering_ratio *= (debug2 = tether_prior) ;

		/* fprintf(stderr,"DEBUG %d: %f -> %f ([%f,%f],%f,%f) %f -> %f (%f)\n",
			j, oldtime[j], proptime[j], tethers[j].win_min, tethers[j].win_max,
			tethers[j].lambda, tethers[j].max_val, debug1, debug2, debug2/debug1); */

	}
	
	regratio *= tethering_ratio ;
}
# endif
	
/**************************************************************
*	Combine all the factors and return a ratio.
*/
	theratio = sqrt(rtratio) * exp(expratio) * regratio ;
	assert( theratio >= 0.0 ) ;
	return( theratio ) ;
}
/* _____________________________________________________ */

long seed_from_file( char *seedfile )
/**********************************************************
*	Get the random number seed from a named file.
*/
{
	FILE	*seed_fp ; 
	long 	initseed ;
	
/**********************************************************
*	Get the seed from the named file or, failing that,
* use a constant.
*/
	if( (seedfile != NULL) && strlen(seedfile) )
		seed_fp = fopen( seedfile, "r" ) ;
	else
		seed_fp = NULL ;
		
	if( seed_fp != NULL ){
		fscanf( seed_fp, "%ld", &initseed ) ;
		fclose( seed_fp ) ;
	}
	else {
		initseed = (unsigned long) 68587608L ;
	}
	
/*********************************
*	Return the seed.
*/	
	return( initseed ) ;
}
/* ______________________________________________________ */

int params_from_file( char *param_file )
/************************************************************
*	Read the parameters of the experiment from a file.
* Return zero for success, nonzero for failure.
*/
{
	FILE	*param_fp ;
	
	assert( param_file != NULL ) ;	/* debugging stuff */
	assert( strlen( param_file ) > 0 ) ;
	
/*************************************************************
*	Try to open the file.
*/
	param_fp = fopen( param_file, "r" ) ;
	if( param_fp == NULL ) {
		fprintf( stderr, 
			"Couldn't open the parameter file %s.\n", 
			param_file ) ;
			
		return( 1 ) ;
	}
	
/************************************************************
*	Get all the parameters
*/	
		/* Integers defining the length of the run */
	burn_in_length = get_int( param_fp ) ;
	cycles_per_sample = get_int( param_fp ) ;
	num_o_samples = get_int( param_fp ) ;
		
		/* parametrs for the Gamma-distributed priors	*/
	mean_root_date = get_double( param_fp ) ;
	root_age_sd = get_double( param_fp ) ;
	
	mean_root_rate = get_double( param_fp ) ;
	root_rate_sd = get_double( param_fp ) ;
	
	mean_nu = get_double( param_fp ) ;
	std_deev_nu = get_double( param_fp ) ;
			
		/* the parameter for the prior	*/
		/* on proportional root depth.	*/
		/* This is ignored if either of	*/
		/* SLICE_VOL_PRIOR or 			*/
		/* IGNORE_PROP_TIME is defined	*/
	node_depth_ab = get_double( param_fp ) ;
		
		/* parameters affecting some proposal densities */
	time_K = get_double( param_fp ) ;
	mixing_K = get_double( param_fp ) ;
	root_K = get_double( param_fp ) ;
	
		/* confidence band to display */
	confidence = get_double( param_fp ) ;
	assert( (confidence > 0.0) && (confidence < 1.0) ) ;

/*************************************************************
*	Do a cursory check that all is well, then return.
*/			
	if( cmd_all_well )	return( 0 ) ;
	else {
		fprintf( stderr, "Trouble reading params (too few?)." ) ;
		return( 2 ) ;
	}
}
/* ___________________________________________________________ */

char *switch_settings(void) 
/**********************************************************
*	Return a pointer to a string that describes the
* version of the code currently running.
*/
{
	static char	switch_str[SWITCH_STRLEN+1] ;
	
/**********************************************************
*	First, note the way the ages of leaf nodes 
* are handled.
*/	
	sprintf( switch_str, "\nSwitches:\n\t" ) ;
	
# if defined( TETHER_LEAF_AGES )

	strncat( switch_str, "TETHER_LEAF_AGES", SWITCH_STRLEN ) ;
	
# elif defined( FREEZE_LEAF_AGES )

	strncat( switch_str, "FREEZE_LEAF_AGES", SWITCH_STRLEN ) ;

# else
/*************************************************************
*	Die immediately: no clear policy on leaf ages is 
* in force.
*/
	fprintf( stderr, "Set either FREEZE_ or TETHER_ LEAF_AGES " ) ;
	fprintf( stderr, "(in div_switches.h) and recompile.\n" ) ;
	
	exit( 1 ) ;
	
# endif 

/*************************************************************
*	Read the command line?
*/
# if defined( READ_COMMAND_LINE )

	strncat( switch_str, ", READ_COMMAND_LINE", SWITCH_STRLEN ) ;
	
# endif 

/*************************************************************
*	Cooperative multi-tasking on the Mac?
*/
# if defined( GOOD_MACOS_CITIZEN )

	strncat( switch_str, ", GOOD_MACOS_CITIZEN", SWITCH_STRLEN ) ;
	
# endif 

/*************************************************************
*	What prior, if any, on the structure of the ages of 
* the internal nodes?
*/
# if defined( IGNORE_PROP_TIME )  

	# if defined( SLICE_VOL_PRIOR )
		strncat( switch_str, ", SLICE_VOL_PRIOR", SWITCH_STRLEN ) ;
	# else 
		strncat( switch_str, ", IGNORE_PROP_TIME", SWITCH_STRLEN ) ;
	# endif	

# else /* Prop time prior in force */
	
	# if defined( SLICE_VOL_PRIOR )
		fprintf( stderr, "Can't apply both the proportional time prior \n" ) ;
		fprintf( stderr, "and SLICE_VOL_PRIOR. Turn one of them off " ) ;
		fprintf( stderr, "(in div_switches.h) and recompile.\n" ) ;
		
		exit( 2 ) ;	
	# else 
		strncat( switch_str, ", Prop. time prior in use", SWITCH_STRLEN ) ;	
	# endif
# endif 

/*************************************************************
*	Check the log-likelihood of the branch lengths for time
* changes too?
*/
# if defined( USE_LOGLIKE_ALWAYS )

	strncat( switch_str, ", USE_LOGLIKE_ALWAYS", SWITCH_STRLEN ) ;
	
# endif 


/**************************
*	Debugging?
*/
# if !defined( NDEBUG )	/* a double negative meaning "if debugging" */

	strncat( switch_str, ", Debugging", SWITCH_STRLEN ) ;
	
# endif 

	return( switch_str ) ;
}
