/**************************************************************
*	The functions below read and write trees for a modified
* version of Jeff Thorne's divtime program. 
*
* N.B. They handle only binary (i.e. rooted) trees.
*
* mrm: Whalley Range, 22 August, 1999
*/
# include <ctype.h>
# include <stdio.h>
# include <string.h>
# include <stdlib.h>

# define DEBUG_LEVEL 0
# include "tune_debug.h"

# include "div_switches.h"
# include "tether_tree.h"
# include "mrm_utils.h"
# include "arrayutl.h"

/***************************************************************
*	Prototypes for the functions private to this file.
*/
static void fprint_subtree(FILE *outfile, int parent, double *lengths);
static int init_buffers(void);
static int record_leaf(char *tokens[]);
static int record_internal(char *tokens[]);
static void str_tolower(char *str);

/***************************************************************
*	Variables private to this file. Most are just nicknames
* for those declared in tether_divtime.c
*/
static double	*branch_len = NULL ;
static char		**node_name = NULL ;
/* _______________________________________________________________ */

int read_tree( FILE *infile )
/******************************************************************
*	Read a tree. The format is:
*
*	Leaf: Name date window tau
*	Leaf: Name date window tau
*		.
*		.
*		.
*	Internal: child_1 child_2  parent constraint
*
* Comments starting with a # character may begin anywhere on a
* line and continue to its end.  Comments and blank lines
* are ignored.  Return zero for success, non-zero for failure.
*
*/
{
	int		n_tokens, verdict ;
	char	line[LINE_BUF_SIZE], *tokens[MAX_TOKENS] ;
	char	finished ;
	
/********************************************************
*	Initialize all the buffers to be bigger than we 
* could ever use.
*/		
	if( init_buffers() ) 
		return( 1 ) ;
		
/********************************************************
*	Enter a loop that reads first the leaves, then
* the internal nodes.
*/
	verdict = 0 ;	/* indicates all is well */
	finished = 0 ;	/* not yet anyway */
	while( !finished && (verdict == 0) &&
			(fgets( line, LINE_BUF_SIZE, infile ) != NULL) ) {
	/******************************************************
	*	Chop off comments and remove any leading 
	* or trailing white space, then break the line
	* into tokens.
	*/
		trim_comments( line ) ;
		strip_ends( line ) ;
		
		n_tokens = get_tokens( tokens, MAX_TOKENS, line ) ;
		
	/*******************************************************
	*	If the line looks right, record it as a Leaf or
	* an Internal node.
	*/	
		if( n_tokens > 0 ) {
			str_tolower( tokens[0] ) ; /* be case-insensitive */
			if( strcmp( "leaf:", tokens[0] ) == 0 ) {
				verdict = record_leaf( tokens ) ;
			}
			else if( strcmp( "internal:", tokens[0] ) == 0 ) {
			/*******************************************************
			*	Stop when we seem to have read enough data.
			*/
				verdict = record_internal( tokens ) ;	
				if( (nnodes == (2*ntips - 1)) || 
					(nbranches == (nnodes-1)) 	) {
				/*************************************
				*	We're finished - set a flag.
				*/
					finished = 1 ;
				}
	
			}
		}		
	}

/******************************************************
*	Check that we actually found the end of the tree,
* then return.
*/
	if( !finished && !verdict ) {
		fprintf( stderr, "Too few internal nodes.\n" ) ;
		verdict = 1 ;
	}
	
	if( check_tree( tonode, nnodes, QUIET ) ) {
		fprintf( stderr, "Something seems wrong with the tree.\n" ) ;
		verdict = 2 ;
	}
	
	return( verdict ) ;
}
/* ________________________________________________________________ */

void fprint_tree( FILE *outfile, double *lengths )
/****************************************************************
*	Print this file's tree in Newick format on the named file.
*/
{	
	assert( outfile != NULL ) ;
			
/***********************************************
*	Print the outermost brackets, then use
* recursion to do the rest.
*/
	fprintf( outfile, "(" ) ;
	fprint_subtree( outfile, tonode[nnodes-1][1], lengths ) ;
	fprintf( outfile, ", " ) ;
	fprint_subtree( outfile, tonode[nnodes-1][2], lengths ) ;
	fprintf( outfile, ") %s ;\n", sprint_name( nnodes - 1 ) ) ;	
}
/* _________________________________________________________________ */

void fprint_subtree( FILE *outfile, int parent, double *lengths )
/******************************************************
*	Print the subtree that starts with parent
*/
{
	int		kid, kid_node_num ;
	
	assert( parent < nnodes ) ;	/* debugging stuff */

/*******************************************************
*	This function is meant to be used recursively,
* so there are two basic cases: the named parent
* is childless, hence a leaf node, in which case
* the recursion stops, or it has kids, in which
* case the recursion goes on.
*/
	if( tonode[parent][N_KIDS] == 0 ) {
	/*************************************************
	*	We've reached a leaf node - print the name
	*/
		if( lengths != NULL ) {
			fprintf( outfile, "%s: %9.6f", 
				sprint_name( parent ), lengths[parent] ) ;
		}
		else {
			fprintf( outfile, "%s", sprint_name( parent ) ) ;
		}	
	}	
	else {
	/*************************************************
	*	Print the subtrees for the children enclosed
	* in parentheses.
	*/
		for( kid = 1 ; kid <= tonode[parent][N_KIDS] ; kid++ ) {
		/*****************************************************
		*	First print either a left-paren or a comma,
		* depending on whether any other kids have yet
		* been read.
		*/
			if( kid == 1 )	fprintf( outfile, "(" ) ;
			else			fprintf( outfile, ", " ) ;
		
		/*****************************************************
		*	Now print the kid's sub-tree.
		*/
			kid_node_num = tonode[parent][kid] ;
			fprint_subtree( outfile, kid_node_num, lengths ) ;
		}
		
		if( lengths != NULL ) {
			fprintf(outfile, ") %s: %9.6f", 
				sprint_name(parent), lengths[parent] ) ;
		}
		else {	
			fprintf(outfile, ") %s", sprint_name(parent) ) ;
		}	
	}
}
/* _______________________________________________________________ */

char *sprint_name( int whichnode )
/*********************************************
*	Returns a string containing the name 
* of the specified node
*/
{
	static char 	name_buf[NAME_LEN+1] ;
	register char	*bpt, *npt ;

	bpt = name_buf, npt = name[whichnode] ;
	while( (*npt != '\n') && (*npt != '\0') ) {
		*bpt++ = *npt++ ;
	}
	
	*bpt = '\0' ;
	return( name_buf ) ;
}
/* _______________________________________________________________ */

int check_tree( int **tree_matrix, int N_nodes, int verboseQ )
/**************************************************************
* Check for all sorts of problems in a tree, (orphaned nodes,
* danglng sub-trees, internal leaves . . . ), printing output 
* as directed by verboseQ: if YES, print all sorts of rubbish,
* if NO, print only error messages.
*
* Tree_matrix[][] is assumed to have the same layout as
* tonode[][], described in tether_tree.h.
*/
{
	int		node, N_kids, kid_num, kid, panode, verdict ;
	
	if( verboseQ == VERBOSE ) {
		printf( "Tree topology\n" ) ;
		printf( "Node num (n_kids): child1 child2 parent\n" ) ;
	}
	
	verdict = 0 ;	/* all appears well */
	for( node = 0 ; (node < N_nodes) && !verdict ; node++ ) {
	/**********************************************************
	*	First print out a description of the node and its
	* children, then check for bastardy.
	*/
		if( verboseQ == VERBOSE ) {
			printf( "%3d (%1d kids): ", node, tree_matrix[node][N_KIDS] ) ;
		
			if( tree_matrix[node][N_KIDS] ) 
				printf( "%3d  %2d  ", tree_matrix[node][1], tree_matrix[node][2] ) ;
			else
				printf( " *   *  " ) ;
				
			if( tree_matrix[node][PARENT] >= 0 )	
				printf( "%3d\n", tree_matrix[node][PARENT] ) ;
			else
				printf( " *\n" ) ;
		}
		
	/************************************************************
	*	Check for improper or missing connections
	*/	
		panode = tree_matrix[node][PARENT] ;
		if( (panode == -1) && (node != (N_nodes - 1)) ) {
		/********************************************************
		*	Only the root node, which is node number (N_nodes-1),
		* should be without a valid parent.
		*/
			verdict = 1 ;
			printf( "Node %d has no parent; pretends to be root.\n", node ) ; 
		}
		else if( panode >= N_nodes ) {
		/********************************************************
		*	No node corresponds to panode - its index is too big.	
		*/
			verdict = 2 ;
			printf( "Node %d has an imaginary parent, %d.\n", node, panode ) ;
		}
		else if(	(node != (N_nodes - 1)) && /* skip the root */
					(tree_matrix[panode][1] != node) && 
					(tree_matrix[panode][2] != node) 	) {
		/*******************************************************************
		*	The node's parent does not recognize it.
		*/ 
			verdict = 3 ;
			printf( "Node %d is a bastard. (%d is dad?)\n", node, panode ) ;
		}
		
	/**************************************************************************
	*	Check that any child nodes point back to their parent.
	*/
		N_kids = tree_matrix[node][N_KIDS] ;
		if( (N_kids != 2) && (N_kids != 0) ) {
			verdict = 4 ;
			printf( "Node %d claims to have %d kids?!?\n", node, N_kids ) ;
		}
		else for( kid_num = 1 ; kid_num <= N_kids ; kid_num++ ) {
			kid = tree_matrix[node][kid_num] ;
			if( (kid < 0) || (kid >= N_nodes) ) {
				verdict = 5 ;
				printf( "Node %d claims imaginary node %d as a child?!?\n", 
						node, kid ) ;
			}
			else if( tree_matrix[kid][PARENT] != node ) {
			/*******************************************************************
			*	Child node does not acknowledge its parent.
			*/
				verdict = 6 ;
				printf( "Node %d is an ungrateful child of %d.\n", kid, node ) ;
			}
		}
	}
	
	if( verboseQ == VERBOSE )
		printf( "\n\n" ) ;
		
	return( verdict ) ;		
}
/* ________________________________________________________________ */

int init_buffers()
/*******************************************************
*	Build big fixed arrays and be done with it.
*/
{
	nnodes = nbranches = ntips = 0 ;
	
	branch_len = mean = dvec( MAX_NODES ) ;
	constraint = dvec( MAX_NODES ) ;
	node_name = name = cmat( MAX_NODES, NAME_LEN ) ;
	tonode = imat( MAX_NODES, (MAX_KIDS + 2) ) ;
	
	if( (tonode == NULL) || (node_name == NULL) || 
		(branch_len == NULL) || (constraint == NULL) ) 
		return( 1 ) ;
	else
		return( 0 ) ;
}
/* _________________________________________________________________ */

int record_leaf( char *tokens[] )
/****************************************************************
*	Copy data into the appropriate arrays.
*/
{
	int			tip_num ;
	double		date ;
	
/************************************************
*	Check for overflow
*/
	if( ntips == MAX_TIPS ) {
		fprintf( stderr, "Too many leaf nodes.\n" ) ;
		return( 1 ) ;	/* failure */
	}

/************************************************
*	First record the cannonical data.
*/
	tonode[ntips][N_KIDS] = 0 ;
	tonode[ntips][PARENT] = -1 ;	/* value should be replaced */
	
	tip_num = atoi( tokens[1] ) ;
	sprintf( node_name[tip_num], "%d->%s", tip_num, tokens[2] ) ;
	date = tip_date[tip_num] = atof( tokens[3] ) ;
	
# ifdef TETHER_LEAF_AGES
/************************************************
*	Record the parameters of the prior on the
* date of this node. 
*/
	tethers[ntips].win_width = atof( tokens[4] ) ;
			
	tethers[tip_num].tau = atof( tokens[5] ) ;		
	tethers[tip_num].lambda = (1.0 / tethers[tip_num].tau) ;

		/* Set tip date to middle of window, but	*/
		/* keep track of max possible date too		*/
	tip_date[tip_num] += 0.5 * tethers[ntips].win_width ;
	date += tethers[ntips].win_width ;

# endif

/************************************************
*	Finally, keep track of the largest date seen
* so far, then increment the counts and return.
*/
	if( date > max_tip_date ) 
		max_tip_date = date ;
		
	ntips++, nnodes++ ;	
	return( 0 ) ;	/* success */
}
/* __________________________________________________________________ */

int record_internal( char *tokens[] )
/******************************************************
*	Record data about an internal node.
*/
{
	int		child_a, child_b, parent ;
	double	constr ;
	
/************************************************
*	Check for overflow
*/
	if( ntips == MAX_TIPS ) {
		fprintf( stderr, "Too many leaf nodes.\n" ) ;
		return( 1 ) ;	/* failure */
	}

/******************************************************
*	Get all the data . . .
*/	
	child_a = atoi( tokens[1] ) ;
	child_b = atoi( tokens[2] ) ;
	parent = atoi( tokens[3] ) ;
	constr = atof( tokens[4] ) ;
	
/***********************************************
*	Check it for sense and record the topology,
* names and constraints.
*/ 
	if( (child_a < 0) || (child_a >= MAX_NODES) ||	
		(child_b < 0) || (child_b >= MAX_NODES) ||
		(parent < 0) || (parent >= MAX_NODES) ) {
	/*********************************************
	*	Something looks bad . . .
	*/
		fprintf( stderr, "Bad node number for an internal node.\n" ) ;
		return( 1 ) ;
	}
	
	tonode[parent][N_KIDS] = 2 ;
	tonode[parent][1] = child_a ;
	tonode[parent][2] = child_b ;
	tonode[parent][PARENT] = -1 ;	/* will be replaced for all save root */
	
	tonode[child_a][PARENT] = parent ;
	tonode[child_b][PARENT] = parent ;
	
	constraint[parent] = constr ;
	sprintf( node_name[parent], "%d", parent ) ;
	nnodes++, nbranches += 2 ;
	
	return( 0 ) ; /* success */
}
/* _____________________________________________ */

void str_tolower( char *str ) 
/**********************************************
*	Convert a whole string to lower case.
*/
{
	if( str != NULL ) {
		while( *str != '\0' ) {
			*str = tolower( *str ) ;
			str++ ;
		}
	}
}
