/******************************************************************************
*	The functions below are intended as replacements for J. Thorne's
* logcondmvn() and friends - the ones that handle the multivariate
* normal approximation to the likelihood function for trees.
*
* The idea here is to make a quadratic approximation to the log-likelihood
* function in a neighborhood of its maximum.  As this may be a constrained
* maximum we must entertain the possibility that there are non-vanishing
* linear terms, so we end up saying (TeX notation):
*
*	Log-likelihood( x)  \approx 
*
*			\sum_{n_branches} (x_j - m_j) * grad_j 
*				+ 0.5 * \sum_{j,k} (x_j - m_j) Hess_{j,k} (x_k - m_k)
*
* Here grad_j are the components of the gradient, Hess_{j,k} are the
* components of the Hessian (hence Hess_{j,k} = Hess_{k,j}) and 
* m_j are the components of the point at which the log likelihood
* assumes its maximum.
*
* 	mrm:	29 July, UMIST
*			31 October, Whalley Range
*/
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <float.h>
# include <math.h>

/*******************************************************************
*	One of the routines below, tll_Hessian_spec(), requires some
* functions from the Numerical Recipes suite. These are not in the
* public domain, so we can't distribute them, and so the function
* doesn't do anything unless the macro HAVE_NUMERICAL_RECIPES is
* defined. If you *do* have Numerical Recipes (it's cheap and very
* useful) then uncomment the line below and modify the makefile
* or CodeWarrior project to add 
*
*	tred2.c, pythag.c tqli.c and nrutil.c.
*/
/* # define HAVE_NUMERICAL_RECIPES */

# define DEBUG_LEVEL 0
# include "tune_debug.h"

# include "arrayutl.h"
# include "mrm_utils.h" 
# include "tll.h"

/*******************************************************************
*	Prototypes for functions private to this file.
*/
static int covar_to_tll_Hessian(double *covar);
static int set_tll_pos_o_max(double *new_pos_o_max);
static void set_tll_gradient(double *new_gradient);
static int setup_tll_mem(int new_dim);

static void tll_cleanup(void);
static void tll_check_max(void);
static void tll_Hessian_neg_def_Q(void);
static void tll_Hessian_spec(void);
static void print_big_entries(double *vect, int dim, double tol);
/*******************************************************************
*	Variables private to this file. The names are chosen to 
* emphasize that all we are doing is a second-order approximation.
*/
static int		dim = 0 ;
static double	*pos_o_max = NULL, *gradient = NULL, *temp_x = NULL ;
static double	**Hessian = NULL, *Hess_dbls = NULL ;
/* ______________________________________________________________________ */

double delta_tree_loglike(	double	*crnt_x, 
							double	new_x,
							int 	pos )
/*******************************************************************
*	Compute the change in log likelihood of the tree when a
* single one of the branch-lengths changes. The result is
*
* 	delta = tll( new_lengths ) - tll( old_lengths )
*
* where tll( lengths ) gives the log-likelihood.
*/
{
	int		j ;
	double	delta_tll, delta_x ;
	
/********************************************************
*	If debugging, check whether the initialization 
* appears to have been done correctly.
*/
	assert( dim > 0 ) ;
	assert( gradient != NULL ) ;
	assert( Hessian != NULL) ;
	assert( (Hessian[0] != NULL) && (Hessian[dim-1] != NULL) ) ;
	
/*********************************************************
*	Add up the all the contributions and return. There's
* a bit of faffing 'cause only the lower-triangular part
* of the Hessian, which is symmetric, is available.
*/		
	/* linear term */
	delta_x = new_x - crnt_x[pos] ;
	delta_tll = delta_x * gradient[pos] ;

	/* quadratic terms */
	for( j=0 ; j < pos ; j++ ) {
		delta_tll +=	delta_x * Hessian[pos][j]
						* ( crnt_x[j] - pos_o_max[j] ) ;
	}

	delta_tll +=	0.5 * (new_x - pos_o_max[pos])
						* Hessian[pos][pos]
						* (new_x - pos_o_max[pos]) ;
						
	delta_tll -=	0.5 * (crnt_x[pos] - pos_o_max[pos])
						* Hessian[pos][pos]				
						* (crnt_x[pos] - pos_o_max[pos]) ;
						
	for( j = (pos + 1) ; j < dim ; j++ ) {
		delta_tll += 	delta_x * Hessian[j][pos]
						* ( crnt_x[j] - pos_o_max[j] ) ;
	}

	return( delta_tll ) ;
}
/* ______________________________________________________________________ */

double vect_delta_tree_loglike(	double	*crnt_x, 
								double	*new_x,
								int 	*changes,
								int		N_changes )
/*******************************************************************
*	Compute the change in log-likelihood of the tree when 
* many of the branch-lengths change. The result is
*
* 	delta = tll( new_lengths ) - tll( old_lengths )
*
* where tll( lengths ) gives the log-likelihood.
*/
{
	int		j, k, pos ;
	double	delta_tll, delta_x ;
	
/********************************************************
*	If debugging, check whether the initialization 
* appears to have been done correctly.
*/
	assert( dim > 0 ) ;
	assert( temp_x != NULL ) ;
	assert( gradient != NULL ) ;
	assert( Hessian != NULL) ;
	assert( (Hessian[0] != NULL) && (Hessian[dim-1] != NULL) ) ;
	
	/* Also, check the args */
	assert( crnt_x != NULL ) ;
	assert( new_x != NULL ) ;
	assert( changes != NULL ) ;
	assert( N_changes > 0 ) ;
	
/*********************************************************
*	Copy crnt_x into this file's private workspace.
*/
	memcpy( temp_x, crnt_x, dim * sizeof(double) ) ;
	
/*********************************************************
*	Add up the all the contributions, moving from
* crnt_x[] to new_x[] one step at a time. Each step
* has the same form as delta_tree_loglike() above.
*/	
	delta_tll = 0.0 ;
		
	/* Handle one position at a time */
	for( k = 0 ; k < N_changes ; k++ ) {
	/********************************************
	*	Gradient terms first  . . . 
	*/
		pos = changes[k] ;
		delta_x = new_x[pos] - temp_x[pos] ;
		delta_tll += delta_x * gradient[pos] ;

	/**************************************
	* 	. . . then the quadratic terms. 
	*/
		for( j=0 ; j < pos ; j++ ) {
			delta_tll +=	delta_x * Hessian[pos][j]
							* ( temp_x[j] - pos_o_max[j] ) ;
		}

		/* the diagonal terms */
		delta_tll +=	0.5 * (new_x[pos] - pos_o_max[pos])
							* Hessian[pos][pos]
							* (new_x[pos] - pos_o_max[pos]) ;
							
		delta_tll -=	0.5 * (temp_x[pos] - pos_o_max[pos])
							* Hessian[pos][pos]				
							* (temp_x[pos] - pos_o_max[pos]) ;
							
		for( j = (pos + 1) ; j < dim ; j++ ) {
			delta_tll += 	delta_x * Hessian[j][pos]
							* ( temp_x[j] - pos_o_max[j] ) ;
		}
	
	/***************************************************
	*	We've now got the change in log-likelihood
	* involved in making the change at position pos.
	*/
		temp_x[pos] = new_x[pos] ;	
	}

	return( delta_tll ) ;
}
/* ______________________________________________________________________ */

double tree_loglike( double *x )
/***********************************************************************
*	Up to normalization, compute the full approximation to the
* log-likelihood, linear terms and all.
*/
{
	int		j, k ;
	double	tll, *delta_x ;
	
/********************************************************
*	If debugging, check whether the initialization 
* appears to have been done correctly.
*/
	assert( dim > 0 ) ;
	assert( temp_x != NULL ) ;
	assert( gradient != NULL ) ;
	assert( Hessian != NULL) ;
	assert( (Hessian[0] != NULL) && (Hessian[dim-1] != NULL) ) ;
	
	/* Also, check the arg */
	assert( x != NULL ) ;

/*********************************************************
*	Load this file's private workspace, temp_x, with
* the displacements from pos_o_max[].
*/
	delta_x = temp_x ;	/* a nickname to make the code clearer */
	for( j=0 ; j < dim ; j++ ) {
		delta_x[j] = x[j] - pos_o_max[j] ;
	}
	
/**********************************************************
*	Now add everything up . . .
*/
	tll = 0.0 ;
	for( j=0 ; j < dim ; j++ ) {
		tll += delta_x[j] * gradient[j] ;
		for( k=0 ; k < j ; k++ )
			tll += delta_x[j] * Hessian[j][k] * delta_x[k] ;
			
		tll += 0.5 * delta_x[j] * Hessian[j][j] * delta_x[j] ;
	}
	
	return( tll ) ;
}
/* ______________________________________________________________________ */

int init_tree_loglike(	int		dim,		/* must be positive	*/
						double	*pos_o_max,	/* must ne non-NULL	*/
						double	*gradient, 	/* may be NULL 		*/
						double	*covar )	/* must be non-NULL	*/
/***********************************************************************
*	Initialize the machinery described in tree_tll.c
* All the checking of arguments is done by lower level functions.
*/
{
	int	verdict = 0 ;	/* Assume success */
	
	if( (verdict = setup_tll_mem( dim )) != 0 ) 
		return( verdict ) ;
	else if( (verdict = set_tll_pos_o_max( pos_o_max )) != 0 ) 
		return( verdict ) ;
	else if( (verdict = covar_to_tll_Hessian( covar )) != 0 ) 
		return( verdict ) ;
	else
		set_tll_gradient( gradient ) ;
		
	return( verdict ) ;
}
/* ______________________________________________________________________ */

int file_init_tree_loglike( int dim, FILE *fp )
/***********************************************
*	Read derivs from an already-open file.
*/
{
	int		j, verdict ;
	
/***********************************************
*	Check the args and set up memory . . .
*/	
	if( fp == NULL ) {
		fprintf( stderr, "NULL file passed to file_init_tree_loglike.\n" ) ;
		return( 1 ) ;
	}
	
	if( (verdict = setup_tll_mem( dim )) != 0 ) 
		return( verdict ) ;
		
/************************************************
*	. . . then fill all the arrays.
*/
			
	if( (verdict = read_val_list(fp, pos_o_max, dim)) != 0 ) 
		return( verdict ) ;
	 
	 if( (verdict = read_val_list(fp, gradient, dim)) != 0 )
	 	return ( verdict ) ;
	 	
	 for( j=0 ; j < dim ; j++ ) {
	 	if( (verdict = read_val_list(fp, Hessian[j], (j+1))) != 0 ) 
	 		return( verdict ) ;
	 }
	 
/**************************************************
*	Check that the position of the maximum and
* make sure that the Hessian is negative definite.
*/
	tll_check_max() ;
	tll_Hessian_neg_def_Q() ; 
	tll_Hessian_spec() ;	/* does nothing unless		*/
							/* HAVE_NUMERICAL_RECIPES	*/
							/* is defined above.		*/
		
	verdict = 0 ;		/* If we get here, all has gone well */
	return( verdict ) ;	
}
/* ___________________________________________________________________ */

int covar_to_tll_Hessian( double *covar )
/************************************************************
*	This function is intended for compatibility with 
* Jeff Thorne's original evolution-of-evolutionary rates code.
* It takes the covariance matrix, which is imagined to be
* stored row-after-row in a linear array of dim^2 entries,
* inverts it and stores the lower-triangular part in this
* file's private Hessian array. 
*/
{
	int		j, row ;
	size_t	bytes_to_copy, n_Hess_dbls ;
	double	**temp, *row_ptr ;
	
	void	invert( double **matrix, int dim ) ;
	
/*************************************************************
*	If debugging, check against some avoidable disasters.
*/
	assert( covar != NULL ) ;
	assert( dim > 0 ) ;
	assert( Hessian != NULL ) ;
	
/*************************************************************
*	Try to allocate a square matrix big enough to hold 
* the covariance matrix and, if successful, copy the 
* data into place.
*/
	temp = dmat( dim-1, dim-1 ) ;	/* args are max_row, max_col */
	if( temp == NULL ) {
		return( 1 ) ;
	}
	else {
	/**********************************************************
	*	All is well, copy the data into place
	*/
		bytes_to_copy = dim * sizeof( double ) ;
		
		row_ptr = covar ;
		for( row = 0 ; row < dim ; row++ ) {
			memcpy( temp[row], row_ptr, bytes_to_copy ) ;
			row_ptr += dim ;
		}
	}
	
/**************************************************************
*	Invert the covariance matrix and copy the supra-diagonal
* part of it into the Hessian.
*/
	invert( temp, dim-1 ) ;

	for( row = 0 ; row < dim ; row++ ) {
		row_ptr = temp[row] ;
		bytes_to_copy = (row + 1) * sizeof( double ) ;
		
		memcpy( Hessian[row], row_ptr, bytes_to_copy ) ;
	}
	
/***************************************************************
*	Change the sign of all the terms.
*/
	n_Hess_dbls = (((size_t) dim) * (dim + 1)) / 2 ;
	for( j=0 ; j < n_Hess_dbls ; j++ ) {
		Hess_dbls[j] *= -1 ;
	}
	
/***************************************************************
*	Free the temporary matrix and return.
*/
	free( temp ) ;
	return( 0 ) ;
}
/* __________________________________________________________________ */

int set_tll_pos_o_max( double *new_pos_o_max ) 
/************************************************************
*	The approximation embodied in this file depends on
* derivs and local coordinates calculated relative to 
* some pos_o_max.
*/ 
{
	size_t	bytes_to_copy ;
	
	assert( gradient != NULL ) ;	/* debugging paranoia	*/
	assert( dim >= 0 ) ;			/* more of the same		*/

	if( pos_o_max == NULL ) {
		fprintf( stderr, "Yikes! NULL pos_o_max set_tll_pos_o_max()!\n" ) ;
		return( 1 ) ;
	}
	
	bytes_to_copy = dim * sizeof( double ) ;
	memcpy( pos_o_max, new_pos_o_max, bytes_to_copy ) ;
	
	return( 0 ) ; /* success */
}
/* ___________________________________________________________________ */

void set_tll_gradient( double *new_gradient ) 
/************************************************************
*	Record the gradient of the log-likelihood function
* evaluated at the pos_o_max.
*/ 
{
	size_t	bytes_to_copy ;
		
	assert( gradient != NULL ) ;	/* debugging paranoia	*/
	assert( dim >= 0 ) ;			/* more of the same		*/
	
	bytes_to_copy = dim * sizeof( double ) ;
	if( new_gradient != NULL )
		memcpy( gradient, new_gradient, bytes_to_copy ) ;
	else
		memset( gradient, 0, bytes_to_copy ) ;
}
/* ______________________________________________________________________ */

int setup_tll_mem( int new_dim )	
/***********************************************************
*	Set up storage for the elements of the approximation.
* Return zero fo success, non-zero for failure.
*/
{
	int			row ;
	size_t		n_Hess_dbls, row_length ;
	double		*row_ptr ;
	
/********************************************************
*	Check the arg and croak if it's silly.
*/
	if( new_dim <= 0 ) {
		fprintf( stderr, "Yikes! dim <= 0 in init_tree_tll()!\n" ) ;
		return( 1 ) ;
	}
	else {
		dim = new_dim ; 
	}

/********************************************************
*	Allocate vectors to hold the gradient, the origin
* of the local coord system and a workspace. The routine 
* dvec() takes as its arg the highest index we'll use. 
* It uses calloc() at base, so the result is automatically 
* filled with zeroes.
*/
	pos_o_max = dvec( dim - 1 ) ;
	gradient = dvec( dim - 1 ) ;
	temp_x = dvec( dim - 1 ) ;
	if( (gradient == NULL) 	|| 
		(pos_o_max == NULL) || (temp_x == NULL) ) {
	/*****************************************************
	*	Clear up any partial results and return in failure.
	*/
		tll_cleanup() ;
		return( 2 ) ;
	}
	
/********************************************************
*	Now allocate the Hessian. It's symmetric, so we
* store only the lower-triangular part of the matrix. 
*/
	n_Hess_dbls = (((size_t) dim) * (dim + 1)) / 2 ;
	 
	Hessian = (double **) calloc( (size_t) dim, sizeof( double* ) ) ;
	Hess_dbls = (double *) calloc( n_Hess_dbls, sizeof( double ) ) ; 
	if( (Hessian == NULL) || (Hess_dbls == NULL) ) {
	/*****************************************************
	*	Clear up any partial results and return in failure.
	*/
		tll_cleanup() ;
		return( 3 ) ;
	}
	else {
		row_ptr = Hess_dbls ;
		row_length = 1 ;
		for( row = 0 ; row < dim ; row++, row_length++ ) {
			Hessian[row] = row_ptr ;
			row_ptr += row_length ;
		}
	}
			
/*********************************************************
*	If we reach these lines all has gone well - 
* return zero to indicate success.
*/
	return( 0 ) ;
}
/* ______________________________________________________________________ */

void tll_cleanup()
/**************************************************************
*	Safely free any pieces of memeory that we appear to have 
* have allocated, then return the globals to their initial
* values.
*/
{
	dim = 0 ;
	freedvec( pos_o_max ), pos_o_max = NULL ;
	freedvec( gradient ), gradient = NULL ;
	
	if( Hessian != NULL ) {
		free( Hessian ), Hessian = NULL ;
	}
	
	if( Hess_dbls != NULL ) {
		free( Hess_dbls ), Hess_dbls = NULL ;
	}
}
/* _______________________________________________________________________ */

void tll_copy_max_pos( double *pos )
/***********************************************************
*	Copy the pos_o_max into pos[].
*/
{
	size_t	bytes_to_copy ;
	
	assert( pos != NULL ) ;
	
	bytes_to_copy = dim * sizeof( double ) ;
	memcpy( pos, pos_o_max, bytes_to_copy ) ;
}
/* ________________________________________________________________________ */

void tll_check_max()
/**************************************************************
*	Check how far it is from the user-input maximum to the
* actual maximum. This involves doing a single Newton's
* method step to find a zero of the gradient, finding 
* 
*		dx = - H_inv * grad
*
* where H_inv is the inverse of the Hessian and grad is the
* gradient. The maximum ought then to be at 
* 
*		x = max_pos + dx
*
* allowing for the constraints that x_j >= 0.
*/
{
	int			j, k, all_well ;
	double		**H_inv, *dx, ratio, maxratio ;
	
	extern void	invert(double**a, int max_index);
	
/*******************************************
*	Allocate all the memory we'll need.
*/
	H_inv = dmat( (dim-1), (dim-1) ) ;	/* args are highest indices, not dims */
	dx = calloc( dim, sizeof(double) ) ;
	if( (dx == NULL) || (H_inv == NULL) ) {
		fprintf( stderr, "Out of memory in tll_check_max().\n" ) ;
		return ;
	}

	do {
	/******************************************
	*	Copy the Hessian and invert it.
	*/
		for( j=0 ; j < dim ; j++ ) {
		/**********************************************
		*	Recall that Hessian is symmetric, so only 
		* the lower-triangular part is stored in 
		* Hessain[][].
		*/
			for( k=0 ; k <= j ; k++ )
				H_inv[j][k] = Hessian[j][k] ;
				
			for( k = (j+1) ; k < dim ; k++ )
				H_inv[j][k] = Hessian[k][j] ;
				
		}
		
		invert( H_inv, (dim-1) ) ;

	/********************************************************
	*	Compute dx.
	*/
		for( j=0 ; j < dim ; j++ ) {
			dx[j] = 0.0 ;
			
			for( k=0 ; k < dim ; k++ ) {
				dx[j] += H_inv[j][k] * gradient[k] ;
			}
		}
		
	/*********************************************************
	*	Correct dx so that it doesn't send any coords 
	* negative.
	*/
		for( j=0 ; j < dim ; j++ ) {
			if( (pos_o_max[j] + dx[j]) < 0.0 ) {
				dx[j] = -pos_o_max[j] ;
			}
		}
		
	/**********************************************************
	*	Check that the ratio (dx / pos_o_max) is small, then
	* free all the memory and return.
	*/
		all_well = 1 ; maxratio = 0;
		for( j=0 ; j < dim ; j++ ) {
			if( pos_o_max[j] > 0.0 ) {
				ratio = fabs( dx[j] / pos_o_max[j] ) ;
				if (ratio > maxratio) maxratio = ratio;
				if( ratio > 1 ) {
					all_well = 0 ;
					printf( "tll_check_max(): (dx[%d] / max[%d]) = %.2f.",
								j, j, ratio ) ;
								
					printf( "  dx[%d] = %.4e,   pos_o_max[%d] = %.4e\n",
												j, dx[j], j, pos_o_max[j] ) ;
				}
				/* Before doing the following, check likelihood 
				   is made better */
				/* 
				pos_o_max[j] += dx[j];
				for (k = 0; k < dim; k++)
				  gradient[k] -= Hessian[k][j]*dx[j];
				*/
			}
		}
		/*
		printf ("tll_check_max(): adjusted pos_o_max because maxratio = %f\n",
			maxratio);
		fflush(NULL);
		*/
		maxratio = 0;
	
	} while (maxratio > 0.01);
	
	freedmat( H_inv, (dim-1), (dim-1) ) ;
	free( dx ) ;
	
	if( all_well )
	  printf( "tll_check_max(): all appears to be well. (maxratio = %f)\n",
		  maxratio ) ;
}
/* __________________________________________________________________ */

void tll_Hessian_neg_def_Q( void )
/********************************************************
*	Use the Gershgorin circle theorem to check whether
* the eigenvalues of the Hessain appear to be bounded 
* below zero.
*/
{
	int		row, col, all_well ;
	double	row_sum, tol ;
	
	all_well = 1 ;
	for( row = 0 ; row < dim ; row++ ) {
		row_sum = 0.0 ;
		for( col = 0 ; col < row ; col++ ) 
			row_sum += fabs( Hessian[row][col] ) ;
			
		row_sum += Hessian[row][row] ;	// should be huge and negative
		
		for( col = (row+1) ; col < dim ; col++ )
			row_sum += fabs( Hessian[col][row] ) ;
			
		tol = dim * DBL_EPSILON * fabs(Hessian[row][row]) ;
		if( row_sum > -(20 * tol) ) {
			all_well = 0 ;

			printf( "tll_Hessain_neg_def_Q(): " ) ;
			printf( "the Gershgorin circle for row %d\n\t", row ) ;
			if( row_sum > (20 * tol) )
				printf( "encloses zero (%.6e).\n", row_sum ) ;
			else
				printf( "may enclose zero (%.6e).\n", row_sum ) ;
		}
	}
	
	if( all_well )
		printf( "tll_Hessain_neg_def_Q(): all appears to be well.\n" ) ;
}
/* __________________________________________________________________ */

void tll_Hessian_spec( void )
/********************************************************
*	Use Numerical Recipes routines to find all the 
* eignevalues of the Hessian and check that they're 
* negative.
*/
{
# if defined( HAVE_NUMERICAL_RECIPES ) 

	int		row, col, N_positive ;
	double	**full_Hess, **evects, *diag, *off_diag ;
	double	max_eval ;
	
	void	my_tqli(double d[], double e[], int n, double **evects);
	void	my_tred2(double **a, int n, double d[], double e[]);
	
/********************************************************
*	Allocate all the memory we'll need.
*/
	full_Hess = dmat( dim, dim ) ;
	evects = dmat( dim, dim ) ;
	
	diag = dvec( dim ) ;
	off_diag = dvec( dim ) ;
	
	if( (diag == NULL) || (off_diag == NULL) || 
		(evects == NULL) || (full_Hess == NULL) ) {
	/****************************************************
	*	At least one allocation failed, quit now.
	*/
		fprintf( stderr, 
			"tll_Hessian_neg_def_Q(): Allocation failed.\n" ) ;
		return ;
	}
	
/**********************************************************
*	Copy the data into place, then use modified Numerical
* Recipes routines to find the eigenvals.
*/
	for( row = 0 ; row < dim ; row++ ) {
		for( col = 0 ; col <= row ; col++ ) 
			full_Hess[row+1][col+1] = Hessian[row][col] ;
			
		for( col = (row+1) ; col < dim ; col++ ) 
			full_Hess[row+1][col+1] = Hessian[col][row] ;
	}
	
	/* Initialize the array that will 		*/
	/* receive evects as the identity.		*/
	for( row = 1 ; row <= dim ; row++ ) 
		evects[row][row] = 1.0 ;
		
	/* Reduce the matrix to tri-diagonal form */
	my_tred2( full_Hess, dim, diag, off_diag ) ;
	
	/* Get evals and evects of the tri-diagonal matrix */
	my_tqli( diag, off_diag, dim, evects ) ;

/***********************************************************
*	Count the number of positive eigenvalues. Print each
* one and also print the most significant entries of 
* their evects.
*/
	N_positive = 0 ;
	max_eval = diag[1] ;
	for( row = 1 ; row <= dim ; row++ ) {
		if( diag[row] > 0.0 )		N_positive++ ;	
		if( diag[row] > max_eval )	max_eval = diag[row] ;
	}
	
	if( N_positive ) {
		fprintf( stderr, 
			"tll_Hessian_spec(): %d positive evals?!?\n",
			N_positive ) ;
			
		for( row = 1 ; row <= dim ; row++ ) {
			if( diag[row] > 0.0 ) {
				fprintf( stderr, "\t%.6e: ", diag[row] ) ;	
				print_big_entries( evects[row], dim, 2.0 ) ;
				fprintf( stderr, "\n" ) ;
			}
		}
	}
	else {
		fprintf( stderr, 
			"tll_Hessian_spec(): max eval is %.6e\n", 
			max_eval ) ;
	}
	
	fprintf( stderr, "\n" ) ;

/************************************************************
*	Free everything, then return.
*/
	freedvec( diag ) ;
	freedvec( off_diag ) ;
	
	freedmat( evects, dim, dim ) ;
	freedmat( full_Hess, dim, dim ) ;
	
# endif /* HAVE_NUMERICAL_RECIPES */
}
/* __________________________________________________________ */

void print_big_entries( double *vect, int dim, double tol )
/***************************************************************
*	Find and display those components of an evect that have 
* absolute value bigger than tol / sqrt( dim ). Note that
* this function uses Numerical Recipes array conventions.
*/
{
	int		j ;
	double	full_tol ; 
	
	full_tol = fabs( tol / sqrt( dim ) ) ;
	for( j=1 ; j <= dim ; j++ ) {
		if( fabs( vect[j] ) >= tol ) {
			fprintf( stderr, "( %d, %.4f ) ", 
					(j-1), vect[j] / sqrt( dim ) ) ;
		}
	}
}
